;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Block comment edit functions
;;  Copyright (C) 1989, 1991 Hewlett-Packard
;;
;;  These routines allow sane editing of block-style comments, such as:
;;
;;  /*****************************************************************
;;   *  This is a                                                    *
;;   *  block style comment                                 Boo!     *
;;   *  such as might be used in                                     *
;;   *  the C programming language.                                  *
;;   *****************************************************************/
;;
;;   or:
;;
;;  /*****************************************************************
;;   ||
;;   ||
;;   *****************************************************************/
;;
;;   or even:
;;
;;  /*=================================================================
;;    ||                                                             ||
;;    ||                                                             ||
;;    =================================================================*/
;;
;;  They will extract the text from the block, allow you to edit it in
;;  another buffer, and then re-create the block from the text in the
;;  buffer.  It will create a block of any desired width (default, 76
;;  characters), and correctly put the "*" characters back on the right
;;  hand side of the comment block.
;;
;;  These routines were designed for use wtih C comment blocks, but they
;;  can be easily used with other languages or comment styles by tweaking
;;  the appropriate variables.  See the documentation for
;;  block-comment-edit for details.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; From: (Steve Koren)
;; Date: Thu, 10 Aug 1989 01:43:55 GMT
;; Subject: Better way to edit comment blocks
;; 
;; Ever wanted to edit source code comment blocks?  There is a function
;; in emacs to do this; it's called c-com-edit, I think.  However, there
;; are a few problems with it.  First, its only good for C comments
;; (this doesn't bother me too much, but it might bother some people).
;; Its limitations bother me, though.  It can only edit comments like
;; this:
;; 
;; /*
;;  * comment
;;  */
;; 
;; or
;; 
;; /*
;; ** comment
;; **/
;; 
;; I don't like either one of those.  I like this:
;; 
;; /********************************************************************
;;  *                                                                  *
;;  * Big hairy comment above a function or procedure - stands out     *
;;  * well!                                                            *
;;  *                                                                  *
;;  ********************************************************************/
;; 
;; Previously, I had to manually adjust the right margin '*' characters
;; every time I did anything. Also, since c-com-mode couldn't cope with
;; these comments, I couldn't use auto-fill or paragraph reformatting.
;; 
;; So, I spent some of last saturday writing a better one.  This one can
;; deal with most any type of comment you can imagine - C comments with
;; a flush right '*', C comments open on the right, lisp comments, etc.
;; The main function of interest is called "block-edit-comment".  See the
;; documentation on that for more details on how to change its behaviour.
;; 
;; The default mappings are as follows:
;; 
;;     \C-c i       insert a new (empty) comment block
;;     \C-c e       edit the comment around or before point
;;     \C-c \C-c    finish editing the comment and put it back
;; 
;; Disclaimer:  before last weekend, I didn't even know lisp.  This is
;;              probably not the most elegant piece of lisp code in the
;;              world.  However, it does seem to work.  I'd appriciate
;;              any feedback, suggestions, comments, etc.  If this
;;              proves to be useful, perhaps Darryl would consider folding
;;              it in to the Grand Unified Emacs?
;; 
;; Oh yeah, its best to auto-load the "comedit" libary in your .emacs.
;; 
;; So, without further babbling, here it is. 
;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 1, or (at your option)
;; any later version.
;; 
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar comedit-edit-mode 'text-mode
  "*Mode in which comments will be edited.  Defaults to 'text-mode.")

(defvar comedit-comment-prefix  " * "  "*Prefix for filled comment block")
(defvar comedit-comment-suffix  "*"    "*Suffix for filled comment block")
(defvar comedit-comment-fill-length 76 "*Length of filled comment block")

(defvar comedit-first-line-prefix "/*" "*Prefix string for 1st comment line")
(defvar comedit-first-line-suffix "*"  "*Suffix string for 1st comment line")
(defvar comedit-first-line-fill   ?*   "*Fill char for 1st comment line")
(defvar comedit-first-line-length 76   "*Length of 1st line of comment block")

(defvar comedit-last-line-prefix  " "  "*Prefix string for last comment line")
(defvar comedit-last-line-suffix  "*/" "*Suffix string for last comment line")
(defvar comedit-last-line-fill    ?*   "*Fill char for last comment line")
(defvar comedit-last-line-length  77   "*Length of last line of comment block")

(if c-mode-map (define-key c-mode-map    "\C-ce" 'block-comment-edit))
(if c-mode-map (define-key lisp-mode-map "\C-ce" 'block-comment-edit))

(if c-mode-map (define-key c-mode-map    "\C-ci" 'comedit-insert-new-comment))
(if c-mode-map (define-key lisp-mode-map "\C-ci" 'comedit-insert-new-comment))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Add enuf spaces to lines in the region to make them all the same
;;  length.  If a line is already long enough or too long, don't add
;;  any spaces.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun comedit-add-spaces-to-lines (length)
  "Add spaces to current line until it reaches a given length"
  (save-excursion
    (while (< (point) (mark))
      (end-of-line)
      (if (> length (current-column))
        (insert (make-string (- length (current-column)) ? ))
      )
      (next-line 1)
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Surround lines in a region with appropriate comment prefix and suffix
;;  characters.  Also, add beginning of comment and end of comment lines.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun comedit-surround-lines ()
"Adds comedit-comment-prefix before lines, makes lines
comedit-comment-fill-length characters long, and adds comedit-comment-suffix
at end of lines."

  ;; Figure out how much stuff to add once we subtract the length of the
  ;; comment prefix and suffix.

  (let ((slength    (- comedit-comment-fill-length
                       (length comedit-comment-prefix)
                       (length comedit-comment-suffix)
                    ))
        (fl_length  (- comedit-first-line-length
                       (length comedit-first-line-prefix)
                       (length comedit-first-line-suffix)
                       -1
                    ))
        (ll_length  (- comedit-last-line-length
                       (length comedit-last-line-prefix)
                       (length comedit-last-line-suffix)
                       -1
                    ))
       )

    ;; add spaces to lines so they are all the same length, but only if
    ;; the suffix exists.

    (if (> (length comedit-comment-suffix) 0)
        (comedit-add-spaces-to-lines slength)
    )

    ;; add prefix before and suffix after lines

    (save-excursion
      (while (< (point) (mark))
        (beginning-of-line)
        (insert comedit-comment-prefix)
        (end-of-line)
        (insert comedit-comment-suffix)
        (next-line 1)
      )
    )

    ;; add first and last lines of comment

    (save-excursion

      ;; Build the first line of the comment

      (insert (concat comedit-first-line-prefix
                      (make-string (- fl_length 1) comedit-first-line-fill)
                      comedit-first-line-suffix
                      "\n"))

      (goto-char (mark))

      ;; Build the last line of the comment

      (insert (concat comedit-last-line-prefix
                      (make-string (- ll_length 1) comedit-last-line-fill)
                      comedit-last-line-suffix))
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Extract meat from comment block lines - removes prefix and suffix.
;;  This leaves the cursor at the first position in the buffer (the desired
;;  location).
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun comedit-extract-lines ()
"Extracts lines from comedit-comment-prefix and comedit-comment-postfix"
  (goto-char 0)
  (replace-string comedit-comment-prefix "" nil)
  (goto-char 0)
  (replace-string comedit-comment-suffix "" nil)
  (goto-char 0)
  (replace-regexp "[ ]*$" "" nil)
  (goto-char 0)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Main code to edit a block comment
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun block-comment-edit ()
"
Copies the comment surrounding point into a temporary buffer for editing
in a more suitable major mode.  The comment will be extracted from a block
of the form:

/***********************************************************************
 *  Your comment goes                                                  *
 *  inside this block                                        Boo!      *
 *  of asterisks                                                       *
 ***********************************************************************/

or even:

 /*=================================================================
   ||                                                             ||
   ||    This is a comment block, too!                            ||
   ||                                                             ||
   =================================================================*/

By the appropriate tweaking of variables, block-edit-comment may be used
to edit nearly any other style of comment, including those in other
(non-C) languages.

The following variables control the behavior of block-comment-edit:

     comedit-comment-prefix

         - A string that is used to prefix each line of the comment
           (should be \" *  \" for the above example)

     comedit-comment-suffix

         - A string that is used to suffix each line of the comment
           (should be \"*\" for the above example).  If this variable
           is \"\", no spaces will be added to the end of the lines.

     comedit-comment-fill-length

         - An integer that represents the length of the comment block text
           lines.  The default is 76 charactesr.  This only affects how
           the block is put back - it will still be dismantled correctly
           for editing.

     comedit-first-line-prefix

         - A string that represents the prefix character or characters for
           the first line of the comment.  In the above example, this
           would be \"/\"

     comedit-first-line-suffix

         - A string that represents the suffix character or characters for
           the first line of the comment.  In the above example, this
           would be \"*\"

     comedit-first-line-fill

         - A character representing the fill character for the comment
           line (the character inserted between the prefix and suffix to
           make the comment line longer if necessary).  In the above
           example, this would be ?*

     comedit-first-line-length

         - An integer representing the length of the first line of the
           comment block.  (This is not necessarily the same as the length
           of the text lines or last line).  The default is 76.

     comedit-last-line-prefix

         - Just like comedit-first-line-prefix, but for the last line.

     comedit-last-line-suffix

         - Just like comedit-first-line-suffix, but for the last line.

     comedit-last-line-fill

         - Just like comedit-first-line-fill, but for the last line.

     comedit-last-line-length

         - Just like comedit-first-line-length, but for the last line.
           Note that in the above example, this is 77, not 76, since the
           last line is one character longer.

     comedit-edit-mode

         - The mode used to edit the comments.  Defaults to text mode.

You should make sure that when you are done editing the comment text,
the resulting lines will fit back in the block.  If they do not, the
right margin will not be properly justified.

In C mode, use \C-ce (control-c e) to edit a comment.  Use \C-c\C-c
to return to normal editing.  block-comment-edit uses recursive editing
for ~editing comments.  \C-c\C-c is bound to exit-recursive-edit.  After
the edit, undo works as expected.

If block-comment-edit is invoked when point is not in a comment block,
the previous comment block (above point) will be used.  This means that
you can invoke block-comment-edit from within a body of code and
generally get the right comment block.

The following functions are provided to easily change the above
parameters to common comment block styles:

        comedit-set-c-style-1
        comedit-set-c-style-2
        comedit-set-c-style-3
        comedit-set-c-style-4
        comedit-set-lisp-style-1

See the documentation on those functions for details.

Hint:  to change the style of a comment block, first use \C-ce to edit
the comment.  Then, change the appropriate variables to the desired
value, and use \C-c\C-c to put the comment back with the new style.
"
  (interactive)
  (save-excursion
    (let ((comedit-buffer-name (get-buffer-create "*Comment Edit*"))
          (comedit-this-buffer-name (buffer-name))
          (comedit-start       0)
          (comedit-end         0)
          (comedit-fline       0)
          (comedit-lline       0)
          (comedit-begin-comment-regexp
             (concat "^"
                     (regexp-quote comedit-first-line-prefix)
                     (regexp-quote (char-to-string comedit-first-line-fill))
                     "*"
                     (regexp-quote comedit-first-line-suffix)
             )
          )
          (comedit-end-comment-regexp
             (concat (regexp-quote comedit-last-line-prefix)
                     (regexp-quote (char-to-string comedit-last-line-fill))
                     "*"
                     (regexp-quote comedit-last-line-suffix)
                     "$"
             )
          )
        )

      (if (not (looking-at comedit-begin-comment-regexp))
          (re-search-backward comedit-begin-comment-regexp)
      )
      (beginning-of-line)
      (setq comedit-start (point))
      (next-line 1)
      (setq comedit-fline (point))
      (re-search-forward comedit-end-comment-regexp)
      (setq comedit-end (point))
      (beginning-of-line)
      (setq comedit-lline (point))
      (copy-to-buffer comedit-buffer-name comedit-fline comedit-lline)
      (switch-to-buffer comedit-buffer-name t)
      (comedit-extract-lines)
      (funcall comedit-edit-mode)
      (local-set-key "\C-c\C-c" 'exit-recursive-edit)
      (message "Type \C-c\C-c to return to normal editing.")
      (recursive-edit)

      ;; glue newline to end of buffer if it doesn't exist

      (end-of-buffer)
      (if (not (char-equal (preceding-char) ?\n))
          (insert "\n")
      )

      (switch-to-buffer comedit-this-buffer-name)
      (goto-char comedit-start)
      (delete-char (- comedit-end comedit-start))
      (insert-buffer comedit-buffer-name)
      (kill-buffer comedit-buffer-name)
      (comedit-surround-lines)
    )
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Insert a new comment block in the current style.  The comment block
;;  will have three blank lines.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun comedit-insert-new-comment ()
"*Insert a new (empty) comment block at point.  This function is initially
 mapped to \C-ci in lisp and c mode."

  (interactive)
  (save-excursion
    (insert "\n\n\n\n\n")
    (backward-char 1)
    (set-mark (point))
    (backward-char 3)
    (comedit-surround-lines)
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Change the comment style
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun comedit-set-c-style-1 ()
"*Change comment style to:

/***************************************************************************
 *                                                                         *
 * Your Comment Here                                                       *
 *                                                                         *
 ***************************************************************************/
"
  (interactive)

  (setq comedit-comment-prefix           " * ")
  (setq comedit-comment-suffix           "*"  )
  (setq comedit-comment-fill-length      76   )

  (setq comedit-first-line-prefix        "/*" )
  (setq comedit-first-line-suffix        "*"  )
  (setq comedit-first-line-fill          ?*   )
  (setq comedit-first-line-length        76   )

  (setq comedit-last-line-prefix         " "  )
  (setq comedit-last-line-suffix         "*/" )
  (setq comedit-last-line-fill           ?*   )
  (setq comedit-last-line-length         77   )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Change the comment style
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun comedit-set-c-style-2 ()
"*Change comment style to:

/*-------------------------------------------------------------------------*
 |                                                                         |
 | Your Comment Here                                                       |
 |                                                                         |
 *-------------------------------------------------------------------------*/
"
  (interactive)

  (setq comedit-comment-prefix           " | ")
  (setq comedit-comment-suffix           "|"  )
  (setq comedit-comment-fill-length      76   )

  (setq comedit-first-line-prefix        "/*" )
  (setq comedit-first-line-suffix        "*"  )
  (setq comedit-first-line-fill          ?-   )
  (setq comedit-first-line-length        76   )

  (setq comedit-last-line-prefix         " *" )
  (setq comedit-last-line-suffix         "*/" )
  (setq comedit-last-line-fill           ?-   )
  (setq comedit-last-line-length         77   )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Change the comment style
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun comedit-set-c-style-3 ()
"*Change comment style to:

/*
 *
 * Your Comment Here
 *
 */
"
  (interactive)

  (setq comedit-comment-prefix           " * ")
  (setq comedit-comment-suffix           ""   )
  (setq comedit-comment-fill-length      76   )

  (setq comedit-first-line-prefix        "/*" )
  (setq comedit-first-line-suffix        ""   )
  (setq comedit-first-line-fill          ?*   )
  (setq comedit-first-line-length        2    )

  (setq comedit-last-line-prefix         " "  )
  (setq comedit-last-line-suffix         "*/" )
  (setq comedit-last-line-fill           ?*   )
  (setq comedit-last-line-length         3    )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Change the comment style
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun comedit-set-c-style-4 ()
"*Change comment style to:

/***************************************************************************
   
  Your Comment Here
   
 ***************************************************************************/
"
  (interactive)

  (setq comedit-comment-prefix           "  " )
  (setq comedit-comment-suffix           ""   )
  (setq comedit-comment-fill-length      76   )

  (setq comedit-first-line-prefix        "/*" )
  (setq comedit-first-line-suffix        "*"  )
  (setq comedit-first-line-fill          ?*   )
  (setq comedit-first-line-length        76   )

  (setq comedit-last-line-prefix         " "  )
  (setq comedit-last-line-suffix         "*/" )
  (setq comedit-last-line-fill           ?*   )
  (setq comedit-last-line-length         77   )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Change the comment style
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun comedit-set-lisp-style-1 ()
"*Change comment style to:

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 
;; Your comment here
;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
"
  (interactive)

  (setq comedit-comment-prefix           ";; ")
  (setq comedit-comment-suffix           ""   )
  (setq comedit-comment-fill-length      76   )

  (setq comedit-first-line-prefix        ";;" )
  (setq comedit-first-line-suffix        ";"  )
  (setq comedit-first-line-fill          ?;   )
  (setq comedit-first-line-length        76   )

  (setq comedit-last-line-prefix         ";;" )
  (setq comedit-last-line-suffix         ";"  )
  (setq comedit-last-line-fill           ?;   )
  (setq comedit-last-line-length         76   )
)
