;;; -*- Mode:Emacs-Lisp -*-

;;; Created 12-feb-91 by Jamie Zawinski <jwz@lucid.com>
;;;
;;; This file makes it possible to define buffer-local variables in the -*-
;;; line as is possible in several other versions of emacs.  For example,
;;;
;;;	-*- mode: emacs-lisp; version-control: never; frob:("a" "b" "c") -*-
;;;
;;; will make the variables `version-control' and `frob' be local to the 
;;; current buffer, and have the values 'never (a symbol) and '("a" "b" "c")
;;; (a list of strings) respectively.
;;; 
;;; The variable named "mode" is a special case, as usual.  "eval" is not
;;; special-cased as it is with end-of-file variables.
;;;
;;; variable-names are case-insensitive (that is, they are downcased before
;;; being interned) but variable values grabbed with `read', so they are case-
;;; sensitive.  That is, the following are equivalent:
;;;
;;;	-*- mode: emacs-lisp; case-fold-search: t -*-
;;;	-*- Mode: emacs-lisp; Case-Fold-Search: t -*-
;;;
;;; but this is not:
;;;
;;;	-*- Mode: emacs-lisp; Case-Fold-Search: T -*-
;;;
;;; However, for backward-compatibility, the value of mode is case-insensitive,
;;; so this is legal:
;;;
;;;	-*- Mode: Emacs-Lisp; Case-Fold-Search: t -*-
;;;
;;; I would have made everything case-insensitive, but there's not an easy
;;; way to make symbols case-insensitive without destroying the case in strings
;;; as well.  (That is, I want ``-*- Tags-File-Name: "~/lib/emacs/TAGS"'' to 
;;; not downcase the string.)


;; originally defined in files.el
;; added a call to `hack-buffer-properties-line' - all else is unchanged.
(defun hack-local-variables (&optional force)
  "Parse, and bind or evaluate as appropriate, any local variables
for current buffer."
  ;; Look for variables in the -*- line.
  (hack-buffer-properties-line)
  ;; Look for "Local variables:" line in last page.
  (save-excursion
    (goto-char (point-max))
    (search-backward "\n\^L" (max (- (point-max) 3000) (point-min)) 'move)
    (if (let ((case-fold-search t))
	  (and (search-forward "Local Variables:" nil t)
	       (or (not inhibit-local-variables)
		   force
		   (y-or-n-p (format"Set local variables as specified at end of %s? "
				    (file-name-nondirectory buffer-file-name))))))
	(let ((continue t)
	      prefix prefixlen suffix beg)
	  ;; The prefix is what comes before "local variables:" in its line.
	  ;; The suffix is what comes after "local variables:" in its line.
	  (skip-chars-forward " \t")
	  (or (eolp)
	      (setq suffix (buffer-substring (point)
					     (progn (end-of-line) (point)))))
	  (goto-char (match-beginning 0))
	  (or (bolp)
	      (setq prefix
		    (buffer-substring (point)
				      (progn (beginning-of-line) (point)))))
	  (if prefix (setq prefixlen (length prefix)
			   prefix (regexp-quote prefix)))
	  (if suffix (setq suffix (regexp-quote suffix)))
	  (while continue
	    ;; Look at next local variable spec.
	    (if selective-display (re-search-forward "[\n\C-m]")
	      (forward-line 1))
	    ;; Skip the prefix, if any.
	    (if prefix
		(if (looking-at prefix)
		    (forward-char prefixlen)
		  (error "Local variables entry is missing the prefix")))
	    ;; Find the variable name; strip whitespace.
	    (skip-chars-forward " \t")
	    (setq beg (point))
	    (skip-chars-forward "^:\n")
	    (if (eolp) (error "Missing colon in local variables entry"))
	    (skip-chars-backward " \t")
	    (let* ((str (buffer-substring beg (point)))
		   (var (read str))
		  val)
	      ;; Setting variable named "end" means end of list.
	      (if (string-equal (downcase str) "end")
		  (setq continue nil)
		;; Otherwise read the variable value.
		(skip-chars-forward "^:")
		(forward-char 1)
		(setq val (read (current-buffer)))
		(skip-chars-backward "\n")
		(skip-chars-forward " \t")
		(or (if suffix (looking-at suffix) (eolp))
		    (error "Local variables entry is terminated incorrectly"))
		;; Set the variable.  "Variables" mode and eval are funny.
		(cond ((eq var 'mode)
		       (funcall (intern (concat (downcase (symbol-name val))
						"-mode"))))
		      ((eq var 'eval)
		       (if (string= (user-login-name) "root")
			   (message "Ignoring `eval:' in file's local variables")
			 (eval val)))
		      (t (make-local-variable var)
			 (set var val))))))))))


(defun hack-buffer-properties-line ()
  "Set local variables specified in the -*- line.  Ignores `mode'."
  (save-excursion
    (goto-char (point-min))
    (skip-chars-forward " \t\n")
    (let ((result '())
	  (end (save-excursion (end-of-line) (point))))
      (cond ((not (search-forward "-*-" end t))
	     ;; doesn't have one.
	     nil)
	    ((looking-at "[ \t]*\\([^ \t\n:;]\\)+[ \t]*-\\*-")
	     ;; Antiquated form: "-*- ModeName -*-".  Ignore it.
	     nil)
	    (t
	     ;; Usual form: '-*-' [ <variable> ':' <value> ';' ]* '-*-'
	     ;; (last ";" is optional).
	     (let ((end (save-excursion (end-of-line) (point))))
	       (save-excursion
		 (if (search-forward "-*-" end t)
		     (setq end (- (point) 3))
		   (error "-*- not terminated before end-of-line")))
	       (while (< (point) end)
		 (or (looking-at "[ \t]*\\([^ \t\n:]+\\)[ \t]*:[ \t]*")
		     (error "malformed -*- line"))
		 (goto-char (match-end 0))
		 (let ((key (intern (downcase (buffer-substring
						(match-beginning 1)
						(match-end 1)))))
		       (val (read (current-buffer))))
		   (setq result (cons (cons key val) result))
		   (skip-chars-forward " \t;")))
	       (setq result (nreverse result)))))
      (while result
	(let ((key (car (car result)))
	      (val (cdr (car result))))
	  (if (eq key 'mode)
	      ;; Mode is handled by `set-auto-mode'.
	      nil
	    ;; otherwise, make it local to this buffer and set it.
	    (make-local-variable key)
	    (set key val)))
	(setq result (cdr result))))))


