;; EDT emulation
;; Paul Davis <davis%m_scr@sdr.slb.com>, March 1988

;; IMPORTANT RESTRICTION 

;; this code is intended to be used on a DEC VT series
;; terminal (or one that uses DEC's SS3 escape sequence
;; "\eO" for its function keys) and with an Emacs that
;; already has an SS3 keymap installed.

;; rationale : who would want to emulate EDT on a non-DEC
;; type keypad anyway ? The keybinding in the original edt
;; emulator is just too slow, and although this version pays
;; a price in being somewhat non-portable (standard function
;; keymap slots are not used) it made more sense in the 
;; environment where it was developed, in particular, allowing
;; the easy switching between an EDT and other keypads. Using
;; function-keymap for that makes life difficult.

;; escape clause: I am working on a portable version that
;; uses the function keymap, with the idea that the code
;; would be loaded into Emacs at build time, thereby
;; avoiding users having to wait whilst the horrendous
;; function \\[define-keypad-key] does its job.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Date: Wed, 5 Sep 90 11:59:05 PDT
;; From: pauld@scenic.wa.com (Paul Davis)
;; To: galagun@sandia.gov
;; Subject: EDT emulation (improved)
;; Cc: info-gnu-emacs@prep.ai.mit.edu
;;
;; I never released this (I think !), but you might be interested in this
;; - a MUCH better emulation of EDT I wrote for people converting from
;; VMS to Unix.
;;
;; Let me know if you want any more information or help (there's not much
;; information unless you understand elisp), let me know.
;;
;; Unfortunately, it is SPECIFICALLY aimed at people using a VT100-style
;; keyboard, but has much better overall emulation of EDT. If you really
;; want help setting it up for a Sparc, just shout and I might be able to 
;; find the time.
;;
;; Paul Barton-Davis			<pauld@scenic.wa.com>
;; ScenicSoft, Inc.			
;; (206) 776-7760
;; 			"100 hours a week and ambivalent ... "
;;
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'edt-keypad)

;;; Some bits from edt.el, modified in some cases
;;; ---------------------------------------------------------------

(defvar edt-last-deleted-lines ""
  "Last text deleted by an EDT emulation line-delete command.")
(defvar edt-last-deleted-words ""
  "Last text deleted by an EDT emulation word-delete command.")
(defvar edt-last-deleted-chars ""
  "Last text deleted by an EDT emulation character-delete command.")

(defun edt-delete-current-line (num)
  "Delete one or specified number of lines after point.
This includes the newline character at the end of each line.
They are saved for the EDT undelete-lines command."
  (interactive "p")
  (let ((beg (point)))
    (forward-line num)
    (if (not (eq (preceding-char) ?\n))
	(insert "\n"))
    (setq edt-last-deleted-lines
	  (buffer-substring beg (point)))
    (delete-region beg (point))))

(defun edt-delete-current-word (num)
  "Delete one or specified number of words after point.
They are saved for the EDT undelete-words command."
  (interactive "p")
  (let ((beg (point)))
    (forward-word num)
    (setq edt-last-deleted-words
	  (buffer-substring beg (point)))
    (delete-region beg (point))))

(defun edt-delete-current-char (num)
  "Delete one or specified number of characters after point.
They are saved for the EDT undelete-chars command."
  (interactive "p")
  (setq edt-last-deleted-chars
	(buffer-substring (point) (min (point-max) (+ (point) num))))
  (delete-region (point) (min (point-max) (+ (point) num))))

(defun edt-undelete-lines ()
  "Yank lines deleted by last EDT line-deletion command.
Put point at start of inserted text."
  (interactive)
  (setq here (point))
  (insert edt-last-deleted-lines)
  (goto-char here))

(defun edt-undelete-words ()
  "Yank words deleted by last EDT word-deletion command.
Put point at start of inserted text."
  (interactive)
  (setq here (point))
  (insert edt-last-deleted-words)
  (goto-char here))

(defun edt-undelete-chars ()
  "Yank characters deleted by last EDT character-deletion command.
Put point at start of inserted text."
  (interactive)
  (setq here (point))
  (insert edt-last-deleted-chars)
  (goto-char here))

(defun move-to-beginning ()
  "Move cursor to the beginning of buffer, but don't set the mark."
  (interactive)
  (goto-char (point-min)))

(defun move-to-end ()
  "Move cursor to the end of buffer, but don't set the mark."
  (interactive)
  (goto-char (point-max)))

(defun update-mode-line ()
  "Make sure mode-line in the current buffer reflects all changes."
  (set-buffer-modified-p (buffer-modified-p))
  (sit-for 0))

;;;----------------------------------------------------------------------
;;; New EDT keypad emulation functions.

(defun edt-replace ()
  "Kill the region and replace it with the most recently
killed text"
  (interactive)
  (kill-region (region-beginning) (region-end))
  (yank 2))

(defun edt-append-kill ()
  "Emulation of EDT append function. Not very satisfcatory since there
this doesn't check whether you want to append words, lines or characters.
Appended text can be retrieved using edt-undelete-lines."
  (interactive)
  (setq edt-last-deleted-lines (list edt-last-deleted-lines
				     (buffer-substring (region-beginning) (region-end))))
  (append-next-kill)
  (kill-region (region-beginning) (region-end)))

(defun edt-section ()
  "Useless function to satisfy the dedicated. Moves 16 lines
in the current direction"
  (interactive)
  (if (equal edt-direction-string " ADVANCE")
      (beginning-of-line 16)
    (beginning-of-line -16)))

(defun edt-forward-word ()
  "Move forwards by words the way that EDT does,
although retaining some of GNU's word definition. This
needs expanding to skip characters other than a-zA-Z0-9
occuring at the start of a line."
  (interactive)
  (forward-word 1)
  (skip-chars-forward "^a-zA-Z0-9"))

(defun edt-change-case ()
  "Stupid function to flip the case of all characters in
the current region"
  (interactive)
  (setq here (region-beginning))
  (while (<= here (region-end))
    (if (<= ?a (char-after here))
	(upcase-region here (1+ here))
      (downcase-region here (1+ here)))
    (setq here (1+ here))))

(defun edt-next-bol ()
  "Move to the next beginning of line"
  (interactive)
  (if (equal edt-direction-string " ADVANCE")
      (beginning-of-line 2)
    (beginning-of-line 0)))

(defun edt-next-eol ()
  "Move to the next end-of-line"
  (interactive)
  (cond ((eolp)
	 (if (equal edt-direction-string " ADVANCE")
	     (end-of-line 2)
	   (end-of-line 0)))
	(t (end-of-line 1))))

;; Searching, EDT-style.

;; edt-set-search-buffer (FIND) puts a string into
;; the edt-search-buffer, and uses it. Subsequent
;; calls to edt-search (FINDNXT) use whatever is in
;; the search buffer.

(defvar edt-search-buffer nil
  "*Holds string for the next edt-search")

(define-key minibuffer-local-map "\eOR" 'exit-minibuffer)

(defun edt-set-search-buffer (string)
  "Put a string into the search buffer, and use it"
  (interactive "sSearch for: ")
  (setq edt-search-buffer string)
  (edt-search))

(defun edt-search ()
  (interactive)
  (cond ((null edt-search-buffer) 
	 (message "nothing in the search buffer"))
	(t (if (equal edt-direction-string " ADVANCE")
	       (search-forward edt-search-buffer)
	     (search-backward edt-search-buffer)))))

(defun edt-fill ()
  "Runs fill-region-as-paragraph, but only
if we're in text-mode."
  (interactive)
  (if (eq major-mode 'text-mode)
      (fill-region-as-paragraph (region-beginning) (region-end))
    (message "not in text mode")))

(defun edt-advance-direction ()
  "Set EDT Advance mode so that keypad commands move forward"
  (interactive)
  (define-key edt-SS3-map "q" 'edt-forward-word)
  (define-key edt-SS3-map "s" 'forward-char)
  (define-key edt-SS3-map "w" 'forward-paragraph)
  (setq edt-direction-string " ADVANCE")
  (update-mode-line))

(defun edt-backup-direction ()
  "Set EDT Backup mode so that keypad commands move backwards"
  (interactive)
  (define-key edt-SS3-map "q" 'backward-word)
  (define-key edt-SS3-map "s" 'backward-char)
  (define-key edt-SS3-map "w" 'backward-paragraph)
  (setq edt-direction-string " BACKUP")
  (update-mode-line))

(or (assq 'edt-direction-string minor-mode-alist)
    (setq minor-mode-alist (cons '(edt-direction-string edt-direction-string)
				 minor-mode-alist)))

(defvar esc-o-retained nil
  "Holder for any default binding of \eO")

(defun edt-emulation-on ()
  "Install the EDT emulator keymap and show the
current `direction' on the modeline."
  (interactive)
  (setq x (lookup-key esc-map "O"))
  (if (keymapp x)
      (setq esc-o-retained (copy-keymap x))
    (setq esc-o-retained x))
  (define-key esc-map "O" edt-SS3-map)
  (edt-advance-direction))

(defun edt-emulation-off ()
  "Remove the EDT `direction' indicator from the modeline
and reinstall the plain SS3-map."
  (interactive)
  (setq edt-direction-string nil)
  (update-mode-line)
  (define-key esc-map "O" esc-o-retained))

;; EDT emulator keypad for DEC
;; this keymap can be installed at any time by
;; (define-key esc-map "O" 'edt-SS3-map)
;; which is one of the tasks performed by
;; the function edt-emulation-on (SCR version).

;; preamble: define the keymaps, ensuring that
;; GOLD C-g still quits.

(defvar edt-SS3-map (make-sparse-keymap)
  "*Keymap for EDT keypad on VT100")
(defvar edt-gold-SS3-map (make-sparse-keymap)
  "*Keymap for EDT GOLD keypad on VT100")

(define-key edt-SS3-map "P"  edt-gold-SS3-map)
(define-key edt-gold-SS3-map "\C-g" 'keyboard-quit)

;; cursor arrows

(define-key edt-SS3-map "A" 'previous-line)
(define-key edt-SS3-map "B" 'next-line)
(define-key edt-SS3-map "C" 'forward-char)
(define-key edt-SS3-map "D" 'backward-char)

;; plain keypad (no PF1 - its the GOLD key)

(define-key edt-SS3-map "Q" 'help-for-help)
(define-key edt-SS3-map "R" 'edt-search)
(define-key edt-SS3-map "S" 'edt-delete-current-line)
(define-key edt-SS3-map "e" 'exit-minbuffer)
(define-key edt-SS3-map "n" 'set-mark-command)
(define-key edt-SS3-map "m" 'edt-delete-current-word)
(define-key edt-SS3-map "l" 'edt-delete-current-char)
(define-key edt-SS3-map "p" 'edt-next-bol)
(define-key edt-SS3-map "q" 'edt-forward-word)
(define-key edt-SS3-map "r" 'edt-next-eol)
(define-key edt-SS3-map "s" 'forward-char)
(define-key edt-SS3-map "t" 'edt-advance-direction)
(define-key edt-SS3-map "u" 'edt-backup-direction)
(define-key edt-SS3-map "v" 'kill-region)
(define-key edt-SS3-map "w" 'forward-page)
(define-key edt-SS3-map "x" 'edt-section)
(define-key edt-SS3-map "y" 'edt-append-kill)

;; GOLD keypad

(define-key edt-gold-SS3-map "\eOQ" 'help-for-help)
(define-key edt-gold-SS3-map "\eOR" 'edt-set-search-buffer)
(define-key edt-gold-SS3-map "\eOS" 'edt-undelete-lines)
(define-key edt-gold-SS3-map "\eOe" 'replace-string)
(define-key edt-gold-SS3-map "\eOn" 'redraw-display)
(define-key edt-gold-SS3-map "\eOm" 'edt-undelete-words)
(define-key edt-gold-SS3-map "\eOl" 'edt-undelete-chars)
(define-key edt-gold-SS3-map "\eOp" 'open-line)
(define-key edt-gold-SS3-map "\eOq" 'edt-change-case)
(define-key edt-gold-SS3-map "\eOr" 'kill-line)
(define-key edt-gold-SS3-map "\eOs" 'quoted-insert)
(define-key edt-gold-SS3-map "\eOt" 'move-to-end)
(define-key edt-gold-SS3-map "\eOu" 'move-to-beginning)
(define-key edt-gold-SS3-map "\eOv" 'yank)
(define-key edt-gold-SS3-map "\eOw" 'execute-extended-command)
(define-key edt-gold-SS3-map "\eOx" 'edt-fill)
(define-key edt-gold-SS3-map "\eOy" 'edt-replace)
