;;;;;
;; LCD Archive Entry:
;; gnus-speedups|Felix Lee|flee@cs.psu.edu|
;; Simple performance enhancements for GNUS|
;; 1992-13-11|1.6: stable|~/misc/gnus-speedups.Z|

;; This package contains two plug-compatible performance enhancements
;; for GNUS 3.13 and 3.14:

;; 1. Faster subject and date sorting in the *Subject* buffer.

;; 2. Functions that implement faster set operations on group lists,
;; which make startup and saving faster.

;; This file must be loaded after "gnus.el".  Loading this instead of
;; "gnus.el" will also work fine.  A GNUS maintainer will want to add
;; this at the end of the local gnus wrapper.

(require 'gnus)
(provide 'gnus-speedups)

;;;;;
;; 1. Faster subject and date sorting in the *Subject* buffer.

;; The basic idea is explained in 'gnus-keyed-sort.

" Sample hook usage:

To automatically sort a newsgroup by subject:
(setq gnus-select-group-hook 'gnus-sort-headers-by-subject)

To automatically sort by subject and date:
(setq gnus-select-group-hook
      (function
       (lambda ()
	 (gnus-sort-headers-by-date)
	 (gnus-sort-headers-by-subject)
	 )))
"
;; XXX These examples should be put in the doc for gnus-select-group-hook

(defun gnus-keyed-sort (list compare extract)
  "Sort LIST stably and return the sorted list.  Does not modify LIST.
Arguments are (LIST COMPARE EXTRACT).  Elements in the list are
compared as if the predicate were:
	(COMPARE (EXTRACT a) (EXTRACT b))
but EXTRACT is run over each element of the list in a preprocessing
stage for efficiency.  This reduces the number of EXTRACT calls from
O(N log N) to O(N).

Example: (gnus-keyed-sort load-path 'string< 'downcase)
"
  (let ( (keyed-list
	  (mapcar
	   (function (lambda (it) (cons (funcall extract it) it)))
	   list)) )
    (setq keyed-list
	  (sort keyed-list
		(function
		 (lambda (a b) (funcall compare (car a) (car b))))))
    (mapcar (function (lambda (it) (cdr it)))
	    keyed-list)
    ))

(defun gnus-keyed-sort-headers (compare extract &optional reverse)
  "Sort current group's headers by COMPARE and EXTRACT.  Sorting is
done as if the predicate were
	(COMPARE (EXTRACT a) (EXTRACT b))
See 'gnus-keyed-sort for more detailed description.
Optional argument REVERSE means reverse the sense of the sort.
Note: interrupting the sort leaves the headers unsorted.
"
  (setq gnus-newsgroup-headers
	(gnus-keyed-sort gnus-newsgroup-headers compare extract))
  (if reverse
      (setq gnus-newsgroup-headers (nreverse gnus-newsgroup-headers)))
  )

;; For backward compatibility with GNUS 3.13
;; XXX (eval-and-compile ...)
(if (not (fboundp 'gnus-sortable-date))
    (fset 'gnus-sortable-date 'gnus-comparable-date))

(defun gnus-sort-headers-by-date (&optional reverse)
  "Sort the current group by date.  Optional argument REVERSE means
reverse order.  Does not redisplay the *Subject* buffer.  See also
'gnus-summary-sort-by-date.
"
  (gnus-keyed-sort-headers
   (function string<)
   (function
    (lambda (it)
      (gnus-sortable-date (nntp-header-date it))))
   reverse))

;; XXX It should be 'gnus-sort-fold-case, not 'case-fold-search
(defun gnus-sort-headers-by-subject (&optional reverse)
  "Sort the current group by subject.  \"Re:\"s are ignored.  If
'case-fold-search is non-nil, then capitalization is ignored.
Optional argument REVERSE means reverse order.  Does not redisplay the
*Subject* buffer.  See also 'gnus-summary-sort-by-subject.
"
  ;; The main complication here is we try to speed up the sort process
  ;; by hoisting conditions outside the sort.
  (gnus-keyed-sort-headers
   (function string<)
   (if case-fold-search
       (function
	(lambda (it)
	  (downcase (gnus-simplify-subject (nntp-header-subject it) 're-only))))
     (function
      (lambda (it)
	(gnus-simplify-subject (nntp-header-subject it) 're-only)))
     )
   reverse)
  )

(defun gnus-summary-sort-by-subject (reverse)
  "Sort the *Subject* buffer by subject.  With a prefix argument,
reverse the order of the sort.  Uses 'gnus-sort-headers-by-subject.
"
  (interactive "P")
  (let ( (current (gnus-summary-article-number)) )
    (gnus-sort-headers-by-subject reverse)
    (gnus-summary-prepare)
    (gnus-summary-goto-subject current)
    ))

(defun gnus-summary-sort-by-date (&optional reverse)
  "Sort the *Subject* buffer by date.  With a prefix argument, reverse
the order of the sort.  Uses 'gnus-sort-headers-by-date.
"
  (interactive "P")
  (let ( (current (gnus-summary-article-number)) )
    (gnus-sort-headers-by-date reverse)
    (gnus-summary-prepare)
    (gnus-summary-goto-subject current)
    ))

;;; Path: princeton!newsserver.jvnc.net!howland.reston.ans.net!zaphod.mps.ohio-state.edu!uwm.edu!lll-winken.llnl.gov!fnnews.fnal.gov!fnnews!SNYDER
;;; From: SNYDER@d0sb10.fnal.gov
;;; Newsgroups: gnu.emacs.gnus
;;; Subject: Faster article threading for GNUS
;;; Date: 07 May 1993 04:20:37 GMT
;;; Organization: SUNY Stony Brook High Energy Physics
;;; Lines: 89
;;; Message-ID: <SNYDER.93May6222039@d0sb10.fnal.gov>
;;; NNTP-Posting-Host: d0sb10.fnal.gov
;;; 
;;; 
;;; hi -
;;; 
;;; Like many people here, i rather like the user interface which GNUS presents,
;;; but i am sometimes frustrated by its speed (or lack thereof).
;;; 
;;; So this is my contribution.  It is a  drop-in replacement for the GNUS
;;; threading functions which runs significantly faster than the original.
;;; I've been using it for quite a while with 3.14.1 with no problems;
;;; a couple days ago, i installed 3.14.4, and it seems to work fine with
;;; that version as well.
;;; 
;;; In a quick test on a heavily-threaded newsgroup containing 846 articles,
;;; the setup time (measured from when GNUS said it was done parsing headers
;;; to when the subject listing appeared) was reduced from 92 seconds with the
;;; original version to about 27 seconds with this new version.
;;; (On a VAXstation 3100M76; your mileage may vary...)
;;; 
;;; I don't know how this compares to the other GNUS speedups which have been
;;; announced; it does have the advantage of being self-contained.
;;; 
;;; scott snyder
;;; snyder@fnald0.fnal.gov
;;; 

(defun gnus-make-threads (newsgroup-headers)
  "Make conversation threads tree from NEWSGROUP-HEADERS."
  (let ((tab (gnus-make-hashtable))
	(refer nil)
	(d nil)
	(dlist nil)
	(roots nil)
	(headers newsgroup-headers))

    (mapcar (function (lambda (h)
			(gnus-sethash (nntp-header-id h) (list h) tab)))
	    newsgroup-headers)

    (while headers
      (setq h (car headers))
      (setq headers (cdr headers))
      ;; Ignore invalid headers.
      (if (vectorp h)			;Depends on nntp.el.
	  (progn
	    ;; Ignore broken references, e.g "<123@a.b.c".
	    (setq refer (nntp-header-references h))
	    (setq dlist (and refer
			     (string-match "\\(<[^<>]+>\\)[^>]*$" refer)
			     (gnus-gethash
			      (substring refer
					 (match-beginning 1) (match-end 1))
			      tab)))
	    (setq d (car dlist))
	    ;; Check subject equality.
	    (or gnus-thread-ignore-subject
		(null d)
		(string-equal (gnus-simplify-subject
			       (nntp-header-subject h) 're)
			      (gnus-simplify-subject
			       (nntp-header-subject d) 're))
		;; H should be a thread root.
		(setq d nil))
	    (if d
		;; H depends on D.
		(setcdr dlist (cons h (cdr dlist)))
	      ;; H is a thread root.
	      (setq roots (cons h roots)))
	    ))
      )
    ;; Make complete threads from the roots.
    ;; Note: dependencies are in reverse order, but
    ;; gnus-make-threads-1 processes it in reverse order again.  So,
    ;; we don't have to worry about it.
    (mapcar
     (function
      (lambda (root)
	(gnus-make-threads-1 root tab))) (nreverse roots))
    ))


(defun gnus-make-threads-1 (parent tab)
  (cons parent
	(mapcar
	 (function
	  (lambda (child)
	    (gnus-make-threads-1 child tab)))
	 (nreverse (cdr (gnus-gethash (nntp-header-id parent) tab)))
	 ))
  )
