;;
;; hideshow.el
;;
;; author: <ttn@netcom.com>
;;
;; Copyright (C) 1993,1994 Thien-Thi Nguyen
;;
;; lcd archive entry goes here !!!todo!!!
;;
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 2 of the License, or (at your
;; option) any later version.
;; 
;; hideshow is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;; 
;; You should have received a copy of the GNU General Public License along
;; with this program; if not, write to the Free Software Foundation, Inc.,
;; 675 Mass Ave, Cambridge, MA 02139, USA.
;;

;; version 3.1

;; commentary

;; this file provides 'hs minor-mode'.  when active, five commands:
;;   hs-{hide,show}-{all,block} and hs-minor-mode
;; are available.  they implement block hiding and showing.  blocks are
;; defined in mode-specific way.  in c-mode or c++-mode, they are simply
;; curly braces, while in lisp-ish modes they are parens.  multi-line
;; comments (c-mode) can also be hidden.  the command M-x hs-minor-mode
;; toggles the minor mode or sets it (similar to outline minor mode).

;; !!!todo!!!
;; suggested usage

;;   (load-library "hideshow")
;;   (add-hook 'c-mode-hook 'hs-minor-mode t)  ; other modes similarly
;;   (add-hook 'c-mode-hook                    ; other modes similarly
;;      '(lambda ()
;;         (define-key hs-minor-mode-map "\C-ch" 'hs-hide-block)
;;         (define-key hs-minor-mode-map "\C-cs" 'hs-show-block)))
;;

;; etc

;; bug reports and fixes welcome.  documentation is ongoing.
;; thanks go to ada@unison.com and preston.f.crow@dartmouth.edu for valuable
;; feedback and bug reports.


;;-----------------------------------------------------------------------------
;; dependencies

(require 'cl)				; common lisp package


;;-----------------------------------------------------------------------------
;; variables

(defvar hs-hide-hooks nil
  "*if non-nil, these hooks will be called by run-hooks at the end
of the following functions:\n
\ths-hide-all
\ths-hide-block")

(defvar hs-show-hooks nil
  "*if non-nil, these hooks will be called by run-hooks at the end
of the following functions:\n
\ths-show-all
\ths-show-block")


;;-----------------------------------------------------------------------------
;; support funcs

;; snarfed from outline.el, but added buffer-read-only
(defun hs-flag-region (from to flag)
  "Hides or shows lines from FROM to TO, according to FLAG.  If FLAG
is \\n (newline character) then text is shown, while if FLAG is \\^M
\(control-M) the text is hidden."
  (let ((modp (buffer-modified-p))
	buffer-read-only)		; nothing is immune
    (unwind-protect (progn
		      (subst-char-in-region
		       from to
		       (if (= flag ?\n) ?\C-m ?\n)
		       flag t) )
      (set-buffer-modified-p modp))))

(defun hs-hide-block-at-point (&optional end)
  "hide block iff on block beginning.  if optional arg END is non-nil,
position point at end of block after hiding."
  (when (looking-at "\\s\(")
    (let* ((p (point))
	   (q (progn (forward-sexp) (point))))
      (forward-line -1) (end-of-line)
      (when (and (< p (point)) (> (count-lines p q) 1))
	(hs-flag-region p (point) ?\C-m))
      (goto-char (if end q p)))))

(defun hs-show-block-at-point (&optional end)
  "show block iff on block beginning.  if optional arg END is non-nil,
position point at end of block after showing."
  (when (looking-at "\\s\(")
    (let ((p (point)))
      (hs-flag-region (point) (progn (forward-sexp) (point)) ?\n)
      (goto-char (if end (1+ (point)) p)))))

(defun hs-safety-is-job-n ()
  "warns if selective-display or selective-display-ellipses is nil"
  (let ((str ""))
    (unless selective-display
      (setq str "selective-display nil "))
    (unless selective-display-ellipses
      (setq str (concat str "selective-display-ellipses nil")))
    (when (/= (length str) 0)
      (message "warning: %s" str)
      (sit-for 2))))

(defun hs-inside-comment-p ()
  "returns t if point is inside a comment, otherwise nil.
actually, for multi-line-able comments, returns a list containing
the buffer position of the start and the end of the comment."
  ;; is it single-line-only or multi-line-able?
  (save-excursion
    (let ((p (point))
	  q)
      (if (string= comment-end "")	; single line
	  (let (found)
	    (beginning-of-line)
	    (setq found (re-search-forward hs-c-start-regexp p t))
	    (and found (not (search-forward "\"" p t))))
	(re-search-forward hs-c-end-regexp (point-max) 1)
	(setq q (point))
	(forward-comment -1)
	(re-search-forward hs-c-start-regexp (point-max) 1)
	(when (< (- (point) (length comment-start)) p)
	  (list (match-beginning 0) q)))))) 

(defun hs-grok-mode-type ()
  "setup variables for new buffers if applicable."
  (when (and (boundp 'comment-start)
	     (boundp 'comment-end))
    (setq hs-c-start-regexp (regexp-quote comment-start))
    (if (string-match " +$" hs-c-start-regexp)
	(setq hs-c-start-regexp
	      (substring hs-c-start-regexp 0 (1- (match-end 0)))))
    (setq hs-c-end-regexp (if (string= "" comment-end) "\n"
			    (regexp-quote comment-end)))
    (if (string-match "^ +" hs-c-end-regexp)
	(setq hs-c-end-regexp
	      (substring hs-c-end-regexp (match-end 0))))
    (setq block-start-regexp
	  (if (or (eq major-mode 'c-mode)
		  (eq major-mode 'c++-mode)) "{" "\\s\("))
    (setq block-end-regexp
	  (if (or (eq major-mode 'c-mode)
		  (eq major-mode 'c++-mode)) "}" "\\s\)"))))

(defun hs-find-block-beginning ()
  "finds the block beginning, placing point at block-start.  return point
as a buffer position.  if no block beginning (top-level), return nil."
  (let (done
	(here (point)))
    (while
	(and (not done)
	     (re-search-backward (concat "\\(" block-start-regexp
					 "\\)\\|\\(" block-end-regexp "\\)")
				 1 t))
      (if (match-beginning 1)		; start-regexp
	  (setq done (match-beginning 1))
	(forward-char 1)
	(forward-list -1)))
    (goto-char (or done here))
    done))

(defmacro hs-life-goes-on (&rest body)
  "executes optional BODY iff variable hs-minor-mode is non-nil"
  (list 'if 'hs-minor-mode (cons 'progn body)))


;;-----------------------------------------------------------------------------
;; commands

;;;###autoload
(defun hs-hide-all ()
  "hides all top-level blocks.  lastly, the hooks set in
hs-hide-hooks are called."
  (interactive)
  (hs-life-goes-on
   (message "hiding all blocks ...")
   (goto-char 1)
   (while (progn
	    (forward-comment (buffer-size))
	    (re-search-forward "\\s\(" (point-max) t))
     (forward-char -1)
     (hs-hide-block-at-point t))
   (hs-safety-is-job-n)
   (message "hiding all blocks ... done")
   (run-hooks 'hs-hide-hooks)))

(defun hs-show-all ()
  "shows all top-level blocks.  lastly, the hooks set in
hs-show-hooks are called."
  (interactive)
  (hs-life-goes-on
   (message "showing all blocks ...")
   (hs-flag-region 1 (point-max) ?\n)
   (message "showing all blocks ... done")
   (run-hooks 'hs-show-hooks)))
  
;;;###autoload
(defun hs-hide-block (&optional end)
  "selects a block and hides it.  block is defined as a sexp for
lispish modes, mode-specific otherwise.  comments are blocks, too.
if optional arg END is non-nil, positions point at end of block after
hiding.  lastly, the hooks set in hs-hide-hooks are called."
  (interactive "P")
  (hs-life-goes-on
   (let ((c-reg (hs-inside-comment-p)))
     (if c-reg
	 (cond ((string= comment-end "")
		(message "can't hide a single-line comment"))
	       ((< (count-lines (car c-reg) (cadr c-reg)) 2)
		(message "not enougn comment lines to hide"))
	       (t
		(goto-char (cadr c-reg))
		(forward-line -1)
		(hs-flag-region (car c-reg) (point) ?\C-m)
		(goto-char (if end (cadr c-reg) (car c-reg)))
		(hs-safety-is-job-n)
		(run-hooks 'hs-hide-hooks)))
       (when (or (looking-at block-start-regexp)
		 (hs-find-block-beginning))
	 (hs-hide-block-at-point end)
	 (hs-safety-is-job-n)
	 (run-hooks 'hs-hide-hooks))))))
  
(defun hs-show-block (&optional end)
  "shows you what was hidden.  lastly, the hooks set in
hs-show-hooks are called."
  (interactive "P")
  (hs-life-goes-on
   (let ((c-reg (hs-inside-comment-p)))
     (if c-reg
	 (cond ((string= comment-end "")
		(message "already looking at the entire comment"))
	       (t
		(hs-flag-region (car c-reg) (cadr c-reg) ?\n)
		(goto-char (if end (cadr c-reg) (car c-reg)))))
       (when (or (looking-at block-start-regexp)
		 (hs-find-block-beginning))
	 (hs-show-block-at-point end)
	 (hs-safety-is-job-n)
	 (run-hooks 'hs-show-hooks))))))
  
;;-----------------------------------------------------------------------------
;; Create hideshow major and minor modes
;; Add menu-bar hide/show options
;;
;; This Section adapted from outline.el by Preston Crow (crow@cs.dartmouth.edu)
;; w/ some mods by ttn.

(defvar hs-mode-map nil "")

(unless hs-mode-map
  (setq hs-mode-map (nconc (make-sparse-keymap) text-mode-map))

  (define-key hs-mode-map [menu-bar hideshow]
    (cons "hideshow" (make-sparse-keymap "hideshow")))

  (define-key hs-mode-map [menu-bar hideshow hs-hide-block]
    '("hide block" . hs-hide-block))
  (define-key hs-mode-map [menu-bar hideshow hs-show-block]
    '("show block" . hs-show-block))
  (define-key hs-mode-map [menu-bar hideshow hs-hide-all]
    '("hide all" . hs-hide-all))
  (define-key hs-mode-map [menu-bar hideshow hs-show-all]
    '("show all" . hs-show-all))
  )

(defvar hs-minor-mode nil
  "Non-nil if using hideshow mode as a minor mode of some other mode.\n
See the function hs-minor-mode for more information.")

(make-variable-buffer-local 'hs-minor-mode)
(make-variable-buffer-local 'hs-c-start-regexp)
(make-variable-buffer-local 'hs-c-end-regexp)
(make-variable-buffer-local 'block-start-regexp)
(make-variable-buffer-local 'block-end-regexp)
(put 'hs-minor-mode 'permanent-local t)
(put 'hs-c-start-regexp 'permanent-local t)
(put 'hs-c-end-regexp 'permanent-local t)
(put 'block-start-regexp 'permanent-local t)
(put 'block-end-regexp 'permanent-local t)

(pushnew '(hs-minor-mode " hs") minor-mode-alist :test 'equal)

(defun hs-mode ()
"Hide Show Mode:

no info available"
  (interactive)
  (kill-all-local-variables)
  (setq selective-display t)
  (use-local-map hs-mode-map)
  (setq mode-name "hs")
  (setq major-mode 'hs-mode)
  (define-abbrev-table 'text-mode-abbrev-table ())
  (setq local-abbrev-table text-mode-abbrev-table)
  (set-syntax-table text-mode-syntax-table)
  (run-hooks 'text-mode-hook 'hs-mode-hook)
)

(defvar hs-minor-mode-prefix "\C-c"
  "*Prefix key to use for hideshow commands in hideshow minor mode.")

(defvar hs-minor-mode-map nil)
(unless hs-minor-mode-map
  (setq hs-minor-mode-map (make-sparse-keymap))
  (define-key hs-minor-mode-map [menu-bar]
    (lookup-key hs-mode-map [menu-bar]))
  (define-key hs-minor-mode-map hs-minor-mode-prefix
    (lookup-key hs-mode-map "\C-c")))

(pushnew (cons 'hs-minor-mode hs-minor-mode-map)
	 minor-mode-map-alist
	 :test 'equal)

;;;###autoload
(defun hs-minor-mode (&optional arg)
  "Toggle hideshow minor mode.\n
With arg, turn hideshow minor mode on if arg is positive, off otherwise.
When hideshow minor mode is on, the menu bar is augmented with hideshow
commands and the hideshow commands are enabled.  The variables\n
\tselective-display\n\tselective-display-ellipses\n
are set to t.  Lastly, the hooks set in hs-minor-mode-hook are called.\n
Turning hideshow minor mode off reverts the menu bar and the
variables to default values and disables the hideshow commands."
  (interactive "P")
  (setq hs-minor-mode
        (if (null arg)
	    (not hs-minor-mode)
          (> (prefix-numeric-value arg) 0)))
  (if hs-minor-mode
      (progn
	(setq selective-display t
	      selective-display-ellipses t)
	(hs-grok-mode-type)
	(run-hooks 'hs-minor-mode-hook))
    (kill-local-variable 'selective-display)
    (kill-local-variable 'selective-display-ellipses)))


;;-----------------------------------------------------------------------------
;; that's it

(provide 'hideshow)

;; end of hideshow.el
