;;; ispell-message - Spell check a message ignoring headers and citations
;;;   Version 2.0
;;;
;;; Copyright (C) 1993 Free Software Foundation, Inc.
;;;
;;; Authors: Daniel Quinlan  <quinlan@bucknell.edu>
;;;          Ethan Bradford  <ethanb@u.washington.edu>
;;;          Alon Albert     <al%imercury@uunet.uu.net>

(defun non-empty-string (string)
  "Return regexp-quote of string if non-empty, else unmatchable regexp."
  (if (or (not string) (string-equal string ""))
      "\\'\\`" ; An unmatchable string if string is null.
    (regexp-quote string)))

(defun ispell-message ()
  "Check the spelling of a mail message or news post.
   Don't check spelling of message headers or included messages.

   To spell-check whenever a message is sent, include this line in .emacs:
   (setq news-inews-hook (setq mail-send-hook 'ispell-message))

   Or you can set bind the function to C-c i in gnus or mail with:
   (setq mail-mode-hook (setq news-mode-hook
     (function (lambda() (local-set-key \"\\C-ci\" 'ispell-message)))))"
 (interactive)
 (let ((case-fold-search nil)
       (non-internal-message nil))	; t if message was passed to emacs.
   (save-excursion
     (goto-char (point-min))

     ;; Don't spell-check the headers.
     (if (search-forward mail-header-separator nil t)
	 (forward-line 1)		; Be at start of first msg. line
       (while (and (looking-at "[A-Z][a-zA-Z-]*:\\|\t\\| ")
		   (not (eobp)))
	   (forward-line 1))
       (setq non-internal-message t)
       )

     (let ((cite-regexp			;Prefix of inserted text
	    (cond
	     ((featurep 'supercite)	; sc 3.0
	      (concat "\\(" (sc-cite-regexp) "\\)" "\\|"
		      (non-empty-string sc-reference-tag-string)))
	     ((featurep 'sc)		; sc 2.3
	      (concat "\\(" sc-cite-regexp "\\)" "\\|"
		      (non-empty-string sc-reference-tag-string)))
	     (non-internal-message	; Assume nn sent us this message.
	      (concat "In [a-zA-Z.]+ you write:" "\\|"
		      "In <[^,;&+=]+> [^,;&+=]+ writes:" "\\|"
		      " *> *"))
	     ((equal major-mode 'news-reply-mode) ;Gnus
	      (concat "In article <" "\\|"
		      (non-empty-string mail-yank-prefix)
		      ))
	     ((boundp 'vm-included-text-prefix) ; VM mail message
	      (concat "[^,;&+=]+ writes:" "\\|"
		      (non-empty-string vm-included-text-prefix)
		      ))
	     (t				; vanilla mail message.
	      (non-empty-string mail-yank-prefix))
	   )))

       (while (not (eobp))
	 ;; Skip across text cited from other messages.
	 (while (and (looking-at (concat "^[ \t]*$\\|" cite-regexp))
		     (not (eobp)))
	   (forward-line 1))
	 (if (not (eobp))
	     ;; Check the next batch of lines that *aren't* cited.
	     (let ((start (point)))
	       (if (re-search-forward
		    (concat "^\\(" cite-regexp "\\)") nil 'end)
		   (beginning-of-line))
	       (save-excursion (ispell-region (- start 1) (point))))))))))

