; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         jbm-shell.el
; RCS:          $Header: jbm-shell.el,v 1.12 89/11/13 16:51:27 jbm Exp $
; Locker:       $Locker: jbm $
;
; Description:  Shell interface.  Based on Dave Worley's shell
;               enhancements and Wolfgang Rupprecht's shellext.el
;               extensions, but with lots of changes and re-writes.
;
; File Author:  Jim Moy, CICD-DMS
; Created:      Thu Oct 12 15:47:03 1989
; Modified:     Sat Feb 23 17:12:55 1991 (Darryl Okahata)
;
; (c) Copyright 1989, 1990, 1991, Hewlett-Packard Company, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;******************************************************************************
;; Disclaimer
;;******************************************************************************

;; Copyright (C) 1985, 1986, 1987, 1988 Free Software Foundation, Inc.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the GNU Emacs General Public
;; License for full details.
;
;; Everyone is granted permission to copy, modify and redistribute
;; GNU Emacs, but only under the conditions described in the
;; GNU Emacs General Public License.   A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(require 'electric)
(provide 'jbm-shell)

;;******************************************************************************
;; User Variables
;;******************************************************************************

(defvar shell-track-pwd nil
  "*If non-nil, use shell-pwd-regexp to match the current directory in
the shell output in order to set the current directory for the shell
buffer.  For this to work, your prompt must contain $PWD somewhere in it.
See the documentation for shell-pwd-regexp for more information.")

(defvar shell-pwd-regexp "\\[.*\\] \\(/.*\\): "
  "*This regexp is used to match the current directory in your prompt
so that the buffer's directory may be set.
 In order for this to work, your prompt must always reflect your
current directory.  This regexp should be set such that \1 will match
the current directory.
 See the section in the on-line help on regular expressions.  You can
get there by evaluating this expression: (Info-find-node "emacs"
"regexps")")

(defvar shell-max-history 100 "*Max shell history retained")

(defvar shell-electric-max-height nil
  "*Maximum height of any shell electric mode windows.")

(defvar shell-history-list-order nil
  "*If t, the electric history list is show in order (most recent command
last) otherwise it will be shown in reverse order.")

;;******************************************************************************
;; Internal Variables
;;******************************************************************************

(defvar last-input-start nil
  "In a shell-mode buffer, marker for start of last unit of input.")
(defvar last-input-end nil
  "In a shell-mode buffer, marker for end of last unit of input.")

(defvar jbm-shell-mode-map nil)

(defvar shell-whitespace-regexp "[ ?\t]+"
  "Regexp matching whitespace between tokens.")

(defvar shell-not-whitespace-regexp "[^ ?\t]"
  "Regexp matching what is not whitespace.")

(defvar shell-history-flagged-line-regexp "[>D]"
  "The regexp used to see if I'm 'looking-at' the beginning of a
electric shell history buffer line.")

(defvar shell-history-next-line-regexp ".*\n[ >D]"
  "The regexp used to see if there is another line after this one
in an electric shell history buffer.")

(defvar shell-history-marked-line-regexp "^>"
  "The regexp used to see if this line is marked.")

(defvar shell-history-any-line-regexp "^[ >D][ *] *\\(%s\\) \\(.+$\\)"
  "The regexp used to match a valid electric shell history buffer line.
Note that the regexp needs to be expanded with format before use.")

(defvar shell-history-delete-line-regexp "^D"
  "The regexp used to check if a line has been marked for deletion.")

(defvar shell-control-character-lookup (make-vector 27 nil)
  " A lookup array to see what control characters should be replaced
with when an attempt was to put them into the shell buffer.  The
indexes into the array are simply the numeric values of the control
character.
 If the lookup returns a character, simple replacement is done.  If
the lookup returns nil, the control character is removed from the
string being inserted.  If the lookup returns a symbol other than nil,
the control character is removed from the string being inserted, and
the symbol is invoked with apply and one argument, the control character.
 This vector is initialized in shell-init-control-lookup.")

;;
;; Now initialize it.
;;
(let ((i (string-to-char "\C-a"))
      (end (string-to-char "\C-z")))
  (while (<= i end)
    (aset shell-control-character-lookup i i)
    (setq i (1+ i)))
  ;;
  ;; ^M gets replaced by a newline.
  ;;
  (aset shell-control-character-lookup ?\C-m ?\n)
  (aset shell-control-character-lookup ?\C-g 'shell-handle-control-character)
  )

(defun jbm-shell-mode ()
  "Major mode for interacting with an inferior shell.

 This mode provides a command history and an electric mode to access it,
as well as an electric mode for completions.  

 \\[jbm-shell-send-input] at the end of the buffer sends the string
between the process mark and the end of the buffer to the shell.  The
process mark may be found by \\[shell-beginning-of-line].

 \\[shell-send-input-history-noreset] is similar, but doesn't reset
the history counter to the top of the history list.  With
\\[jbm-shell-send-input], the command is added to the history list, and
the counter is set to it for recall by \\[shell-previous-command].  By
not resetting the counter, several previous commands may be sequenced
by sending the input with \\[shell-send-input-history-noreset] and
recalling the next command with \\[shell-next-command].

 \\[shell-push-input] does the same thing as \\[jbm-shell-send-input], but
doesn't append a newline, which is useful for programs which normally
read \"raw\" input.

 \\[shell-previous-token] recalls the prefix ARGth whitespace
separated token from the previous command to where point is.  It
defaults to the last token from the previous command.

 \\[shell-history-search-forward] and
\\[shell-history-search-backward] both search through the command
history list for a regexp which is prompted for.  The search begins at
the current index in the list which can be viewed in the shell history
listing.  If the command is found it is placed on the command line.
The current history index is set if the pattern is found.

The electric command history is invoked by \\[shell-list-history].  The
electric completion mode is invoked with \\[shell-filename-expand].
These commands invoke electric buffers with their own major modes for
selecting items from their menus; see the documentation on those modes
for more info.

\\{jbm-shell-mode-map}

Entry to this mode calls the value of jbm-shell-mode-hook with no args,
if that value is non-nil.

If shell-track-pwd is non-nil, the regexp in shell-pwd-regexp will
be used to match the current working directory and set the buffer's
default directory.  See documentation on those variables.
"
  (interactive)
  (if (eq major-mode 'jbm-shell-mode)
      nil
    (kill-all-local-variables)
    (setq major-mode 'jbm-shell-mode)
    (setq mode-name "Shell")
    (setq mode-line-process '(": %s"))
    (use-local-map jbm-shell-mode-map)
    (make-local-variable 'last-input-start)
    (make-local-variable 'last-input-end)
    ;; Last shell search string.
    (make-local-variable 'shell-last-search)
    ;; Max shell history retained.
    (make-local-variable 'shell-history-list)
    ;; Where we are on the history list.  -1 unless currently walking up/down.
    (make-local-variable 'shell-history-index)
    (make-local-variable 'shell-process-object)
    (make-local-variable 'shell-started-up)
    (make-local-variable 'shell-history-real-reverse)
    (setq last-input-start (make-marker))
    (setq last-input-end (make-marker))
    (setq shell-last-search "")
    (setq shell-history-list nil)
    (setq shell-history-index -1)
    (setq shell-process-object nil)
    (setq shell-started-up nil)
    (setq shell-history-real-reverse nil)
    (run-hooks 'jbm-shell-mode-hook)))

(if jbm-shell-mode-map
    (setq jbm-shell-mode-map nil))

(let ((map (make-keymap)))
  (define-key map "\M-\t" 'shell-filename-expand)
  (define-key map "\M-p" 'shell-previous-command)
  (define-key map "\M-." 'shell-previous-token)
  (define-key map "\M-n" 'shell-next-command)
  (define-key map "\M-r" 'shell-history-search-backward)
  (define-key map "\M-s" 'shell-history-search-forward)
  (define-key map "\M-h" 'shell-list-history)
  (define-key map "\M-a" 'shell-beginning-of-line)
  (define-key map "\M-o" 'shell-send-input-history-noreset)
  (define-key map "\M-\r" 'shell-push-input)
  (define-key map "\C-c" nil)
  (define-key map "\r" 'jbm-shell-send-input)
  (define-key map "\C-c\C-r" 'move-to-window-line)
  (define-key map "\C-c\C-d" 'shell-send-eof)
  (define-key map "\C-c\C-c" 'interrupt-shell-subjob)
  (define-key map "\C-c\C-z" 'stop-shell-subjob)
  (define-key map "\C-c\C-\\" 'quit-shell-subjob)
  (define-key map "\C-c\C-o" 'kill-output-from-shell)
  (define-key map "\C-c\C-l" 'show-output-from-shell)
  (define-key map "\C-ch" 'shell-list-history)
  (setq jbm-shell-mode-map map))


(defun jbm-shell (&optional shell-number)
  "
 Run an inferior shell, with I/O through buffer *shell*.

 If buffer exists but shell process is not running, make new shell.
Shell program used comes from variable SHELL, or defaults to /bin/ksh.
The buffer is put in jbm-shell-mode, which gives you a number of
facilities for interacting with the shell.

 A negative prefix argument causes further shell windows (*shell<2>*,
*shell<3>*, etc.) to be created.  A positive argument causes the shell
window with that number to be selected, instead of *shell* (which is
number 1).  Shells with numbers > 1 must be created explicitly; shell
1 is created automatically if it does not already exist.

 See jbm-shell-mode.

 The current directory of the shell is the same as the default-directory
from which the command was given."
  (interactive "p")
  (let* ((prog (or (getenv "SHELL") "/bin/ksh"))
	 (name (file-name-nondirectory prog))
	 (dont nil)
	 shell-name)
    ; process the user argument
    (cond
     ; a null argument is the same as 1
     ((null shell-number)
      (setq shell-number 1))
     ; a negative argument means find a new shell number
     ((< shell-number 1)
      (progn
	(setq shell-number 2)
	(while
	    (get-buffer (concat "*shell<" (int-to-string shell-number) ">*"))
	  (setq shell-number (1+ shell-number)))))
     ; if he's trying to select a shell > 1, check to see if it exists first
     ((> shell-number 1)
      (if (null (get-buffer (concat "*shell<" 
				    (int-to-string shell-number)
				    ">*")))
	  (progn
	    (ding)
	    (setq dont t)))))
    ; if we haven't found an error, do it
    (if (not dont)
	(progn
	  (setq shell-name (if (eq shell-number 1)
			       "shell"
			       (concat "shell<" 
				       (int-to-string shell-number)
				       ">")))
	  (switch-to-buffer (apply 'jbm-make-shell shell-name prog '("-i")))
	  (if (not shell-started-up)
	      (run-shell-startup-sequence))))
    ))

(defun jbm-make-shell (name program &rest switches)
  (let ((buffer (get-buffer-create (concat "*" name "*")))
	proc status size)
    (setq proc (get-buffer-process buffer))
    (if proc (setq status (process-status proc)))
    (save-excursion
      (set-buffer buffer)
      ;;    (setq size (buffer-size))
      (if (memq status '(run stop))
	  nil
	(if proc (delete-process proc))
	(setq proc (apply 'start-process name buffer
			  (concat exec-directory "env")
			  (format "TERMCAP=emacs:co#%d:tc=unknown:"
				  (screen-width))
			  "TERM=emacs"
			  "EMACS=t"
			  "PAGER=cat"
			  "COLUMNS=80"
			  "-"
			  program
			  switches))
	(setq name (process-name proc)))
      (goto-char (point-max))
      (set-marker (process-mark proc) (point))
      (jbm-shell-mode))
    buffer))

;; Change directories when we've just created a buffer, but
;; let other functions track the pwd.

(defun run-shell-startup-sequence ()
  "Sends some initial startup commands to the shell before the normal
process filter takes over.  The output of a newline by the shell indicates
the shell has been started up."
  (let ((start-str (concat
		    "cd "
		    default-directory
		    ";"
		    "set +m;"
		    "echo\n"
		    )))
    (setq shell-process-object
	  (get-buffer-process (current-buffer)))
    (set-process-filter shell-process-object 'shell-startup-filter)
    ;;
    ;; Change to the default directory.  Why do I need to
    ;; turn off job control to get interrupt through as
    ;; advertised?  The echo command gives me the newline
    ;; I'm detecting in shell-startup-filter.
    ;; 
    (process-send-string shell-process-object start-str)
    ))

;; I'll use this filter to get rid of the first line of output which
;; is caused by the extra cd command.  I'm using the cd command because
;; I want the current directory to be the same as the one which was
;; current when the shell command was given.  Also, I don't want to muck
;; with make-shell because other packages make use of it.

(defun shell-startup-filter (proc string)
  "Startup filter for jbm-shell-mode.
Ignores the first line, and then nulls out the process filter."
  (let ((buf (process-buffer proc)))
    (save-excursion
      (set-buffer buf)
      (if (string-match "\n" string)
	  (progn
	    (setq shell-started-up t)
	    (set-process-filter shell-process-object 'shell-normal-filter))))))

(defun shell-normal-filter (proc string)
  "Process filter for jbm-shell-mode.  Handle killed-buffer case, or
take normal path and process the shell output."
  (let ((procbuf (process-buffer proc)))
    ;;
    ;; A perusal of buffer.c shows that the "Selecting deleted buffer"
    ;; messages comes from code checking the buffer pointer and its
    ;; name, so I will check the buffer name here.
    ;;
    (if (buffer-name procbuf)
	(shell-filter-handle-output proc procbuf string)
      (set-process-filter proc nil))))

(defun shell-filter-handle-output (proc buf string)
  "Guts of the shell filter during normal operation."
  (let ((obuf (current-buffer)))
    (set-buffer buf)
    (unwind-protect
	(let ((pmark (process-mark proc)))
	  (save-excursion
	    (goto-char pmark)
	    ;;
	    ;; Handle various special cases of stuff coming out of shell.
	    ;;
	    (shell-filter-control-characters)
	    (if shell-track-pwd (shell-handle-prompt))
	    (if (string-match "^Password:" string)
		(shell-handle-password proc)
	      (insert string))
	    
	    (set-marker pmark (point)))
	  (goto-char pmark))
      (set-buffer obuf))))

(defun shell-handle-password (process)
  "Read password in response to the filter detecting the prompt.
Send to the process."
  (insert "Password: ")
  ;; Some code stolen out of telnet.el, but improved to handle backspaces.
  (let ((answ "") tem)
    (while (not(or  (= (setq tem (read-char)) ?\^m)
		    (= tem ?\n)))
      (if (equal tem ?\C-?)
	  (progn
	    (backward-delete-char 1)
	    (setq answ (substring answ 0 (1- (length answ)))))
	(insert ".")
	(setq answ (concat answ (char-to-string tem)))))
    (send-string process (concat answ "\n"))))

;; This function only works within the context of shell-normal-filter.
;; It's been extracted from that function for readability.

(defun shell-filter-control-characters ()
  (let ((index -1)
	(max (length string))
	look bump savechar)
    ;; Go through the whole string looking for control chars.
    (while (setq index (string-match "[\C-a-\C-m]" string (1+ index)))
      (setq savechar (aref string index))
      (setq look (aref shell-control-character-lookup savechar))
      (if (char-or-string-p look)
	  ;; Simple character substitution.
	  (aset string index look)
	;; Bump the following characters up one.
	(setq bump index)
	(while (< (1+ bump) max)
	  (aset string bump (aref string (setq bump (1+ bump)))))
	;; Invoke symbol if it wasn't a nil.
	(if (and look (symbolp look))
	    (funcall look savechar))))))

;; This function only works within the context of shell-normal-filter.
;; It's been extracted from that function for readability.

(defun shell-handle-prompt ()
  "Takes care of scanning any output from the shell for the PWD in order
to set the buffer's default-directory."
  (if (string-match shell-pwd-regexp string)
      (let* ((str (substring string (match-beginning 1)
			     (match-end 1)))
	     (end (1- (length str))))
	(setq default-directory
	      ;; Check for trailing slash needed by this variable.
	      (if (= (aref str end) ?/)
		  str
		(concat str "/"))))))

(defun shell-handle-control-character (char)
  "This gets called as a result of an entry in variable
shell-control-character-lookup."
  (cond ((equal char ?\C-g)
	 ;;
	 ;; Control-G means beep.
	 ;;
	 (beep))
	(t nil)))

(defun shell-send-input-history-noreset ()
  "Calls shell-send-input with an argument so that the command history
counter doesn't get reset.  No history entry is made."
  (interactive)
  (jbm-shell-send-input t))
  
(defun jbm-shell-send-input (&optional arg)
  "Send input to subshell.
 At the end of buffer, or on the last line in the buffer, sends all
text after last output as input to the subshell, including a newline
inserted at the end.
 Optional ARG non-nil means don't make a command history entry and
don't reset the command history pointer.  This allows a sequence of
commands to be executed by using shell-next-command."
  (interactive)
  (or (get-buffer-process (current-buffer))
      (error "Current buffer has no process"))
  (end-of-line)
  (if (not (eobp))
      (error "Not on last line.")
    (move-marker last-input-start
		 (process-mark (get-buffer-process (current-buffer))))
    (insert ?\n)
    (move-marker last-input-end (point))
    (let ((process (get-buffer-process (current-buffer))))
      (process-send-region process last-input-start last-input-end)
      ;; Make history entry.  Reset so prev and next commands start with
      ;; the most recent command entered, unless arg is given.
      (let ((input-string (buffer-substring last-input-start
					    (1- last-input-end))))
	(if (and (string-match shell-not-whitespace-regexp input-string)
		 (not arg))
	    (let ((len (length shell-history-list)))
	      (setq shell-history-list
		    (cons input-string
			  (if (>= len shell-max-history)
			      (shell-first-n (1- shell-max-history)
					     shell-history-list)
			    shell-history-list)))
	      (setq shell-history-index -1))))
      (set-marker (process-mark process) (point)))))
  
(defun shell-send-eof ()
  "Send eof to subshell (or to the program running under it)."
  (interactive)
  (process-send-eof))

(defun kill-output-from-shell ()
  "Kill all output from shell since last input."
  (interactive)
  (goto-char (point-max))
  (beginning-of-line)
  (kill-region last-input-end (point))
  (insert "*** output flushed ***\n")
  (goto-char (point-max)))

(defun show-output-from-shell ()
  "Display start of the last batch of shell output at top of window.
Also put cursor there."
  (interactive)
  (set-window-start (selected-window) last-input-end)
  (goto-char last-input-end))

(defun interrupt-shell-subjob ()
  "Interrupt this shell's current subjob."
  (interactive)
  (interrupt-process nil t))

(defun kill-shell-subjob ()
  "Send kill signal to this shell's current subjob."
  (interactive)
  (kill-process nil t))

(defun quit-shell-subjob ()
  "Send quit signal to this shell's current subjob."
  (interactive)
  (quit-process nil t))

(defun stop-shell-subjob ()
  "Stop this shell's current subjob."
  (interactive)
  (stop-process nil t))


(defun shell-filename-expand (arg)
  "Complete the filename before (point) as far as possible.  If there
are multiple completions, prompt in an electric buffer for selection.
Prefix ARG non-nil means to list completions sorted across rows
instead of down the columns.  Useful when there are many completions."
  (interactive "P")
  (let* ((end (point))
	 (beg (progn
		(re-search-backward "[ \t\n;]")
		(forward-char 1)
		(point)))
	 (new (shell-pathname-completion (buffer-substring beg end) arg)))
    (cond ((eq new t)
	   (goto-char end)
	   (message "File name is already complete"))
	  ((null new)
	   (goto-char end)
	   (message "No completion"))
	  ((listp new)
	   ;; Completed via electric completion buffer.
	   (goto-char end)
	   (re-search-backward "[ \t\n/;]")
	   (forward-char 1)
	   (delete-region (point) end)
	   (insert (car new)))
	  (t
	   (delete-region beg end)
	   (insert new)))
    ))

(defun shell-pathname-completion (path-name arg)
  "Complete PATHNAME as far as possible, return this string.  If
multiple completions are possible, list in an electric buffer for
selection. ARG non-nil means list completions sorted across rows
instead of columns."
  (let* ((dir (file-name-directory path-name))
	 (file (file-name-nondirectory path-name))
	 (completion
	   (file-name-completion file (or dir ""))))
    (if (string-equal file completion)	; we are at a branch point
	(let* ((list (sort (file-name-all-completions (or file "") (or dir ""))
			   'string-lessp))
	       (e-complete (shell-list-completions list arg)))
	  (if e-complete
	      e-complete
	    path-name))
      (if dir
	  (and completion (or (eq completion t)	; leave t or nil alone
			      (concat dir completion)))
	completion))))


(defun shell-previous-command ()
  "Insert the previous command into the shell buffer."
  (interactive)
  (let ((history (nthcdr (1+ shell-history-index) shell-history-list)))
    (if history
	(progn
	  (delete-region (process-mark (get-buffer-process (current-buffer)))
			 (point-max))
	  (goto-char (point-max))
	  (insert (car history))
	  (setq shell-history-index (1+ shell-history-index)))
;	  (message "history %d" shell-history-index)
      (error "End of history list, (history %d)" shell-history-index))))

(defun shell-next-command ()
  "Insert the next command into the shell buffer."
  (interactive)
  (delete-region (process-mark (get-buffer-process (current-buffer)))
	       (point-max))
  (goto-char (point-max))
  (if (<= 0 shell-history-index)  
      (setq shell-history-index (1- shell-history-index))
    (error "Top of history list (history %d)" shell-history-index))
  (if (<= 0 shell-history-index)
      (progn 
;	(message "history %d" shell-history-index)
	(insert (nth shell-history-index shell-history-list)))))

(defun shell-history-search-backward (string)
  "Search the history list for an occurance of STRING.
Null string argument means use the last search pattern."
  (interactive
   (list (let ((str nil)
	       (prompt (format "History search (%s): " shell-last-search)))
	   (if (string-equal "" (setq str (read-string prompt)))
	       shell-last-search
	     (setq shell-last-search str)))))
  (let* ((index (1+ shell-history-index)) ; start at next command
	 (history (nthcdr index shell-history-list)))
    (while (and history
		(null (string-match string (car history))))
      (setq index (1+ index)
	    history (cdr history)))
    (if history
	(progn
	  (setq shell-history-index index)
	  (delete-region (process-mark (get-buffer-process (current-buffer)))
			 (point-max))
	  (goto-char (point-max))
	  (insert (car history))
;	  (message "history %d" shell-history-index)
	  )
      (error "No match found, history %d" shell-history-index))))


(defun shell-history-search-forward (string)
  "Search the history list for an occurance of STRING.
Null string argument means use the last search pattern."
  (interactive
   (list (let ((str nil)
	       (prompt (format "History search (%s): " shell-last-search)))
	   (if (string-equal "" (setq str (read-string prompt)))
	       shell-last-search
	     (setq shell-last-search str)))))
  (let ((index shell-history-index))
    (while (and (<= 0 index)		; not as effecient as backwards hum...
		(null (string-match
		       string (nth (setq index (1- index))
				   shell-history-list)))))
    ;; index is bounded by: (-1 <= index <= shell-history-index) 
    (if (<= 0 index)
	(progn
	  (setq shell-history-index index)
	  (delete-region (process-mark (get-buffer-process (current-buffer)))
			 (point-max))
	  (goto-char (point-max))
	  (insert (or (nth index shell-history-list) ""))
;	  (message "history %d" shell-history-index)
	  )
      (error "No match found, history %d" shell-history-index))))


(defun shell-beginning-of-line ()
  "Goto the beginning of the command line. This is the process mark and,
hopefully, at the end of the prompt string."
  (interactive)
  (goto-char (process-mark (get-buffer-process (current-buffer)))))

(defun shell-push-input ()
  "Pushes all pending shell input to shell. Like \\[jbm-shell-send-input], only
it doesn't append a newline. Useful for programs that expect to talk 
to a tty in raw mode (eg. tip(1)). The pushed input doesn't get recorded
on the shell's history list."
  (interactive)
  (goto-char (point-max))
  (move-marker last-input-start
	       (process-mark (get-buffer-process (current-buffer))))
  (move-marker last-input-end (point))
  (let ((process (get-buffer-process (current-buffer))))
    (process-send-region process last-input-start last-input-end)
    (set-marker (process-mark process) (point))))

(defun shell-break-string (string pattern)
  "Break up STRING into a list of strings at regexp PATTERN delimiter."
  (let (match)
    (cond ((string-equal string "") nil)
	  (string (cons
		   (substring string 0 
			      (if (setq match (string-match pattern string))
				  (match-beginning 0)))
		   (shell-break-string (if match
				     (substring string (match-end 0)))
				 pattern)))
	  (t nil))))

(defun shell-previous-token (num)
  "Insert the NUMth whitespace separated token from the previous
command into the shell buffer.  Interactively, NUM is the prefix
argument, and if none is supplied the last token is inserted.  If NUM
is greater than the number of tokens, the last one is inserted."
  (interactive "P")
  (let* ((history (nthcdr (1+ shell-history-index) shell-history-list))
	 (command (if history (car history)))
	 (tokens (if command (shell-break-string command
						 shell-whitespace-regexp)))
	 (num-tokens (if tokens (length tokens))))
    (setq num (if (null num) num-tokens
		(let ((n (prefix-numeric-value num)))
		  (if (> n num-tokens)
		      num-tokens
		    n))))
    (if (null num)
	(error "No history entries yet."))
    (if (or (zerop num)
	    (<= num 0))
	(error "Token number too small: %d" num))
    (insert (nth (1- num) tokens))))


;;;****************************************************************************
;;; Electric command history.
;;;****************************************************************************

(defun shell-make-history-buffer (order)
  "List the history in the *History* buffer. A '*' indicates current
position on the history list.  Returns the *History* buffer.  Argument
ORDER non-nil means reverse the ordering of the list from that
specified in variable shell-history-list-order."
  (interactive)
  (let ((buf (get-buffer-create "*History*"))
	(rev (not shell-history-list-order))
	(history shell-history-list)
	(index 1)
	(hist-index shell-history-index)
	(dash-count (- (window-width) 8)))
    (save-excursion
      (set-buffer buf)
      (shell-history-menu-mode)
      (if (setq shell-history-real-reverse
		(or (and rev (not order))
		    (and (not rev) order)))
	  nil
	;; shell-history-list is constructed most-recent-first
	(setq history (reverse history)))
      (setq buffer-read-only nil)
      (erase-buffer)
      (insert " C  # Command\n - -- ")
      (while (> dash-count 0)
	(insert "-")
	(setq dash-count (1- dash-count)))
      (insert "\n")
      (while history
	(insert (format " %c%3d %s\n" 
			(if (= index (1+ hist-index)) ?* ?\ )
			index (car history)))
	(setq history (cdr history)
	      index (1+ index)))
      (setq buffer-read-only t))
    buf))

(defun shell-list-history (arg)
  "Electric shell history listing.  Optional prefix argument non-nil
means reverse the ordering of the list from that specified in variable
shell-history-list-order.  See shell-history-menu-mode."
  (interactive "P")
  (let (select buffer reverse)
    (save-window-excursion
      (Electric-pop-up-window (setq buffer (shell-make-history-buffer arg))
			      shell-electric-max-height)
      (unwind-protect
	  (progn
	    (set-buffer buffer)
	    (setq select
		  (catch 'shell-history-menu-select
		    (let ((first (progn (goto-char (point-min))
					(forward-line 2)
					(point)))
			  (last (progn (goto-char (point-max))
				       (forward-line -1)
				       (point)))
			  (goal-column 0))
		      (goto-char first)
		      ;;
		      ;; The last argument is what gets passed to the loop
		      ;; function as the first argument.
		      ;;
		      (Electric-command-loop 'shell-history-menu-select
					     nil
					     t
					     'shell-history-menu-looper
					     (cons first last))))))
	(set-buffer buffer)
	(setq reverse shell-history-real-reverse)
	(text-mode)
	(bury-buffer buffer)
	(message "")))
    (if (car select)
	(shell-history-execute buffer select (current-buffer) reverse))))

(defun shell-history-menu-looper (state condition)
  "Looper function for electric history commands."
  (cond ((and condition
	      (not (memq (car condition) '(buffer-read-only
					   end-of-buffer
					   beginning-of-buffer))))
	 (signal (car condition) (cdr condition)))
	((< (point) (car state))
	 (goto-char (point-min))
	 (forward-line 2))
	((> (point) (cdr state))
	 (goto-char (point-max))
	 (forward-line -1)
	 (if (pos-visible-in-window-p (point-max))
	     (recenter -1)))))

(defun shell-history-command-string (buf pt)
  "Takes Electric History List buffer BUF and point PT and returns
a pair containing the index into the shell-history-list of the command
string, and the command string on whatever line point is found."
  (save-excursion
    (set-buffer buf)
    (goto-char pt)
    (beginning-of-line)
    (let ((eol (save-excursion (end-of-line) (point))))
      (if (re-search-forward (format shell-history-any-line-regexp "[0-9]+")
			     eol
			     t)
	  (cons (1- (string-to-int (buffer-substring (match-beginning 1)
						     (match-end 1))))
		(buffer-substring (match-beginning 2) (match-end 2)))
	(cons nil nil)))))

(defun shell-history-execute (buf info shellbuf rv)
  "Processes whatever marks have been placed in the Electric History
List buffer, or any special history buffer commands.  Arguments are
the history BUF, INFO which is a list of three items, the point in the
buffer, either nil or token indicating command or token recall, a
possible the digit of the last key pressed, and a numeric prefix
argument if any was given.  SHELLBUF is the buffer where the shell is
waiting for commands.  RV is non-nil when the buffer ordering is in
reverse order (most-recent on top)."
  (let ((multiple nil)
	(simple nil))
    ;; Prefix argument was given, force the simple command.
    (if (nth 3 info)
	(setq simple t)
      (save-excursion
	(set-buffer buf)
	(setq opoint (point))
	(goto-char (point-min))
	(if (re-search-forward shell-history-marked-line-regexp nil t)
	    (if (= (progn (beginning-of-line) (point)) (nth 0 info))
		;; A mark, but point is also there, so flag simple.
		(setq simple t)
	      ;; Mark, and point is on a different line.
	      (setq multiple t))
	  ;; No marked lines.
	  (setq simple t))))
    ;; Get rid of delete history items.
    (setq shell-history-list
	  (shell-history-deletions shell-history-list buf rv))
    ;; Now do the right one outside the save-excursion.
    (cond
     ;; More than one mark.
     (multiple (shell-history-multiple-command buf (nth 0 info) shellbuf rv))
     ;; No marks.
     (simple (shell-history-simple-command (nth 1 info) buf (nth 0 info)
					   shellbuf (nth 2 info)))
     )))

(defun shell-history-multiple-command (buf point shellbuf rev)
  "Args: BUF is the history buffer, POINT is where point was when the
selection was made, SHELLBUF is the shell buffer where I want to
run the commands, and REV is whether BUF is ordered or not."
  (let ((cmdstr nil)
	(proc (get-buffer-process shellbuf)))
    (save-excursion
      (set-buffer buf)
      (let* ((start (save-excursion (if rev (goto-char (point-max))
				      (point-min))))
	     (limit (save-excursion (if rev (goto-char (point-min))
				      (point-max))))
	     (search-command (if rev 're-search-backward 're-search-forward)))
	(goto-char start)
	(while (apply search-command shell-history-marked-line-regexp nil t nil)
	  (let* ((ret (shell-history-command-string buf (point)))
		 (str (cdr ret)))
	    (if (and (not (string-equal str ""))
		     (not (null str)))
		(setq cmdstr (concat cmdstr ";" str)))))
	;; Get rid of first semi-colon.
	(setq cmdstr (concat (substring cmdstr 1) "\n"))
	))
    (set-buffer shellbuf)
    (goto-char (point-max))
    (insert ?\n)
    (set-marker (process-mark proc) (point))
    (process-send-string proc cmdstr)))

(defun shell-history-simple-command (type buf point shellbuf key)
  "Args: TYPE either nil or token, depending on the function to be executed.
BUF is the history buffer and the POINT therein.  SHELLBUF is the shell
buffer where I'm inserting stuff.  KEY, if non-nil is the digit representing
the last key pressed which invoked this command."
  (cond ((or (null type) (eq 'token type))
	 (let* ((ret (shell-history-command-string buf point))
		(begin (if (cdr ret) (string-match shell-not-whitespace-regexp
						   (cdr ret)))))
	   (if begin (setcdr ret (substring (cdr ret) begin)))
	   (if (or (null (cdr ret))
		   (string-equal (cdr ret) ""))
	       (error "Null string found!")
	     (set-buffer shellbuf)
	     (setq shell-history-index (car ret))
	     (cond ((null type)
		    (delete-region (process-mark
				    (get-buffer-process shellbuf))
				   (point-max))
		    (goto-char (point-max))
		    (insert (cdr ret)))
		   ((eq 'token type)
		    (let* ((digit key)
			   (tokens (shell-break-string (cdr ret)
						       shell-whitespace-regexp))
			   (num-tokens (if tokens (length tokens))))
		      (if (or (> digit num-tokens)
			      (null digit)
			      (<= digit 0))
			  (setq digit num-tokens))
		      (insert (concat " " (nth (1- digit) tokens))))
		   ))
	     )))))
  
(put 'shell-history-menu-mode 'mode-class 'special)


(defun shell-history-menu-mode ()
  "
 Major mode for editing an electric shell history buffer.

 Moving to a line and hitting \\[shell-history-menu-select] will
recall that command to the shell prompt.  If given a prefix argument,
the command corresponding to the number will be recalled.

 If a M-<digit> command is given, the digit'th token will be recalled
to the command line.  It will be inserted where point is found without
replacing the command which is already there.  This feature may be
used to build commands out of pieces of previous commands.
 Also, instead of moving to the line, this command may be given a
prefix argument to indicate which command the token is to be extracted
from.  For example, if the fourth token in the twenty-sixth command is
a long path that you want, you hit these keys to append it to your
existing command line: 2 6 M-4

 Multiple commands may be marked with \\[shell-history-menu-mark] for
sequential execution.  The commands, including the one point is on
when the select command is given, are concatenated together separated
by semicolons in the original order they were executed, and sent to
the shell.

 Hitting \\[shell-history-menu-delete] marks a history entry for
deletion from the history list.  The entry will no longer be listed
the next time the history list is created, nor will any of the search
commands find those entries.

 \\[shell-history-menu-select] will process any marked entries.  An
entry may not be marked for both execution and deletion.  The last
mark takes precedence.

\\{shell-history-menu-mode-map}
"
  (kill-all-local-variables)
  (use-local-map shell-history-menu-mode-map)
  (make-local-variable 'shell-history-real-reverse)
  (setq shell-history-real-reverse nil)
  (setq mode-name "Electric History Menu")
  (setq mode-line-buffer-identification "Electric History List")
  (if (memq 'mode-name mode-line-format)
      (progn (setq mode-line-format (copy-sequence mode-line-format))
	     (setcar (memq 'mode-name mode-line-format) "Buffers")))
  (setq truncate-lines t)
  (setq buffer-read-only t)
  (setq major-mode 'shell-history-menu-mode)
  (goto-char (point-min))
  (forward-line 2)
  (if (search-forward "\n." nil t) (forward-char -1))
  (run-hooks 'shell-history-menu-mode-hook))

(defvar shell-history-menu-mode-map nil)
(put 'shell-history-menu-undefined 'suppress-keymap t)

(if shell-history-menu-mode-map
    (setq shell-history-menu-mode-map nil))

(let ((map (make-keymap))
      (meta-map (make-keymap)))
  (fillarray map 'shell-history-menu-undefined)
  (define-key map "\e" meta-map)
  (fillarray (lookup-key map "\e") 'shell-history-menu-undefined)
  (define-key map "\C-h" 'Helper-help)
  (define-key map "?" 'Helper-describe-bindings)
  (define-key map "q" 'shell-history-menu-quit)
  (define-key map "\C-c" nil)
  (define-key map "\C-c\C-c" 'shell-history-menu-quit)
  (define-key map "." 'shell-history-menu-select-token)
  (define-key map "\r" 'shell-history-menu-select)  
  (define-key map " " 'shell-history-menu-select)
  (define-key map "\C-l" 'recenter)
  (let ((i ?0))
    (while (<= i ?9)
      (define-key map (char-to-string i) 'digit-argument)
      (define-key meta-map (char-to-string i) 'shell-history-menu-select-token)
      (setq i (1+ i))))
  (define-key map "-" 'negative-argument)
  (define-key map "\e-" 'negative-argument)
  (define-key map "\C-u" 'universal-argument)
  (define-key map "\C-p" 'previous-line)
  (define-key map "\C-n" 'next-line)
  (define-key map "\M-r" 'move-to-window-line)

  (define-key map "p" 'previous-line)
  (define-key map "n" 'next-line)
  (define-key map "m" 'shell-history-menu-mark)
  (define-key map "u" 'shell-history-menu-unmark)
  (define-key map "d" 'shell-history-menu-delete)
  (define-key map "<" 'scroll-left)
  (define-key map ">" 'scroll-right)
  (define-key map "\C-?" 'shell-history-menu-backup-unmark)
  (define-key map "\C-v" 'scroll-up)
  (define-key map "\M-v" 'scroll-down)
  (define-key map "\e>" 'end-of-buffer)
  (define-key map "\e<" 'beginning-of-buffer)
  (define-key map "\e\e" 'eval-expression)
  (setq shell-history-menu-mode-map map))

(defun shell-history-menu-quit ()
  "Leave Electric History Menu"
  (interactive)
  (throw 'shell-history-menu-select (list nil nil nil nil)))

(defun shell-history-menu-undefined ()
  (interactive)
  (ding)
  (message (if (and (eq (key-binding "\C-c\C-c") 'shell-history-menu-quit)
		    (eq (key-binding " ") 'shell-history-menu-select)
		    (eq (key-binding "\C-h") 'Helper-help)
		    (eq (key-binding "?") 'Helper-describe-bindings))
	       "Type C-c C-c to exit, Space to select, C-h for help, ? for commands"
	     (substitute-command-keys "\
Type \\[shell-history-menu-quit] to exit, \
\\[shell-history-menu-select] to select, \
\\[Helper-help] for help, \\[Helper-describe-bindings] for commands.")))
  (sit-for 4))


(defun shell-history-menu-delete ()
  "Marks the shell history entry for deletion from the history list."
  (interactive)
  (beginning-of-line)
  (let ((buffer-read-only nil))
    (delete-char 1)
    (insert ?D)
    (if (looking-at shell-history-next-line-regexp)
	(forward-line 1))))

(defun shell-history-menu-mark ()
  "Mark shell history entry the current line."
  (interactive)
  (beginning-of-line)
  (let ((buffer-read-only nil))
    (delete-char 1)
    (insert ?>)
    (if (looking-at shell-history-next-line-regexp)
	(forward-line 1))))

(defun shell-history-menu-unmark ()
  "Cancel shell history mark on the current line."
  (interactive)
  (beginning-of-line)
  (let ((buffer-read-only nil))
    (if (looking-at shell-history-flagged-line-regexp)
	(progn
	  (delete-char 1)
	  (insert ? )
	  (backward-char 1)))))

(defun shell-history-menu-backup-unmark ()
  "Back up one line and cancel shell history mark."
  (interactive)
  (let ((top (save-excursion (goto-char (point-min))
			     (forward-line 2)
			     (point))))
    (if (<= (point) top)
	nil
      (forward-line -1)
      (if (looking-at shell-history-flagged-line-regexp)
	  (shell-history-menu-unmark)))))

(defun shell-history-menu-throw (point arg throwtype key)
  "Does the actual throw out of the eletric mode buffer, handing back
control to the original history command function."
  (if (null arg)
      (throw 'shell-history-menu-select (list point throwtype key nil))
    (let* ((num (prefix-numeric-value arg))
	   (pattern (format shell-history-any-line-regexp num))
	   (found nil))
      (save-excursion
	(goto-char (point-min))
	(throw 'shell-history-menu-select
	       ;; Throw a list.
	       (list
		;; (1) Where point is.
		(if (not (re-search-forward pattern nil t))
		    (error "No command numbered %d" num)
		  (point))
		;; (2) The throw type, nil or token.
		throwtype
		;; (3) The value of the meta-digit, if throwtype is token.
		key
		;; (4) The numeric value of arg, if it existed.
		num
		))
	))))

(defun shell-history-menu-select (arg)
  "Recall the command on the current line to the shell buffer.  If
marks have been made in the buffer, they are processed.  If a prefix
argument is given, marks are ignored and the history command
corresponding to that number is recalled."
  (interactive "P")
  (save-excursion
    (beginning-of-line)
    (let ((buffer-read-only nil))
      (delete-char 1)
      (insert ?>)))
  (shell-history-menu-throw (point) arg nil nil))

(defun shell-history-menu-select-token (arg)
  "Select the whitespace separated token corresponding to the digit
key pressed and insert in the shell buffer.  If not invoked through
one of the digit keys, you are prompted for a digit.  If a prefix
argument is given, the history command corresponding to that number is
used to pick the token."
  (interactive "P")
  (let ((digit (- last-command-char (string-to-char "0")))
	(stupid-user nil))
    (while (or (<= digit 0) (> digit 9))
      (message "Digit:")
      (sit-for 0)
      (setq digit (- (read-char) (string-to-char "0")))
      (if (and stupid-user (or (<= digit 0) (> digit 9)))
	  (progn
	    (beep)
	    (message "<<< Need to press a digit key at the prompt. >>>")
	    (sleep-for 1))
	(setq stupid-user t)))
    (shell-history-menu-throw (point) arg 'token digit)))

(defun shell-first-n (n list)
    "Returns the first N items of LIST.  If N is greater than the length
of LIST then LIST is returned."
  (let ((len (length list)))
    (reverse (nthcdr (- len n) (reverse list)))))

(defun shell-remove-nils (list)
  "Returns a list which is the same as LIST with the nil elements removed.
Utilizes a vector to eliminate the need for extending the max eval depth."
  (let* ((newlist nil)
	 (len (length list))
	 (vector (make-vector len nil)))
    ;; fill vector
    (let ((i 0) str)
      (while list
	(setq str (car list) list (cdr list))
	(aset vector i str)
	(setq i (1+ i))))
    ;; make new list
    (let ((i 0) str)
      (while (< i len)
	(setq str (aref vector i))
	(if (null str)
	    nil
	  (setq newlist (append newlist (list str))))
	(setq i (1+ i))))
    ;; return
    newlist))

(defun shell-history-deletions (list buf rev)
  "Delete history items from the history LIST as marked in history BUF.
Argument REV non-nil means that the indexing of the buffer is reversed
from the ordering of LIST."
  (let ((pat (format shell-history-any-line-regexp "[0-9]+"))
	(delpat shell-history-delete-line-regexp)
	max end)
    (save-excursion
      (set-buffer buf)
      ;; Get max index and end point.
      (goto-char (point-max))
      (forward-line -1)
      (end-of-line)
      (setq end (point))
      (beginning-of-line)
      (re-search-forward pat (save-excursion (end-of-line) (point)) t)
      (setq max (string-to-int (buffer-substring (match-beginning 1)
						 (match-end 1))))
      ;; Go through the buffer looking for lines marked for deletion.
      (goto-char (point-min))
      (forward-line 2)
      (while (re-search-forward delpat end t)
	(beginning-of-line)
	(let ((str (buffer-substring (point) (save-excursion (end-of-line)
							     (point))))
	      index first)
	  (string-match pat str)
	  (setq index (string-to-int (substring str (match-beginning 1)
						(match-end 1))))
	  (if (not rev) (setq index (- max (1- index))))
	  (setq list (if (setq first (shell-first-n (1- index) list))
			 (append first (list nil) (nthcdr index list))
		       (nthcdr index list))))
		
	(end-of-line)))
    (shell-remove-nils list)))


;;;****************************************************************************
;;; Electric completions.
;;;****************************************************************************

(defun shell-make-completions-buffer (list name cfirst)
  "Return a buffer with the completions in LIST.  The buffer is named NAME.
If third argument CFIRST is non-nil, the buffer items are sorted down the
columns instead of across rows."
  (let* ((listlen (length list))
	 (fwidth (+ 2 (longest-string list)))
         (cols (/ (window-width) fwidth))
	 (extras (if (> listlen cols) (mod listlen cols) nil))
	 (rows (if extras (1+ (/ listlen cols))
		 (let ((div (/ listlen cols)))
		   (if (zerop div) 1 div))))
         (buf (get-buffer-create name))
	 array)
    (setq list (sort list 'string-lessp))
    (setq array (fill-2d-array rows cols list (if cfirst extras nil)))
    (save-excursion
      (set-buffer buf)
      (setq buffer-read-only nil)
      (erase-buffer)
      (let ((i 0))
	(while (< i rows)
	  (let ((j 0) cvec el)
	    (while (< j cols)
	      (setq cvec (aref array i))
	      (setq el (aref cvec j))
	      (if (stringp el)
		  (insert (concat el (make-string (- fwidth (length el)) ? ))))
	      (setq j (1+ j))))
	  (insert "\n")
	  (setq i (1+ i))))
      (setq buffer-read-only t)
      (shell-completion-menu-mode)
      (setq shell-completion-field-width fwidth))
    buf))


(defun fill-2d-array (rows cols list rfirst)
  "Returns a 2D array ROWS high, COLS wide, and fills it with the
items in LIST.  If fourth argument RFIRST is non-nil, fill the same
index of all the rows before going to the next column.  The actual
value of RFIRST is the number of columns having an extra row."
  (let ((i 0)
	(array (make-2d-array rows cols))
	str cvec)
    (if rfirst
	(while (and (< i cols) list)
	  (let ((j 0) (r rows))
	    ;; Take care of extra row on these columns.
	    (if (< i rfirst)
		nil
	      (setq r (1- r)))
	    (while (and (< j r) list)
	      (setq str (car list)
		    list (cdr list))
	      (setq cvec (aref array j))
	      (aset cvec i str)
	      (setq j (1+ j))))
	  (setq i (1+ i)))
      (while (and (< i rows) list)
	(setq cvec (aref array i))
	(let ((j 0))
	  (while (and (< j cols) list)
	    (setq str (car list)
		  list (cdr list))
	    (aset cvec j str)
	    (setq j (1+ j))))
	(setq i (1+ i))))
    array))

(defun make-2d-array (rows cols)
  "Creates a 2D array ROWS high and COLS wide.  It's actually not a 2D
array, but a vector of ROWS length where each element is a vector which
is COLS in length."
  (let ((i 0) array)
    (setq array (make-vector rows nil))
    (while (< i rows)
      (aset array i (make-vector cols nil))
      (setq i (1+ i)))
    array))

(defun longest-string (list)
  "Given a list of strings, return the length of the longest one."
  (let (len string (max 0))
    (while list
      (setq string (car list)
            list (cdr list))
      (if (> (setq len (length string)) max)
          (setq max len)))
    max))

(defun shell-list-completions (list arg)
  "Invokes the electric completions buffer on the LIST of completions.
ARG non-nil means the electric buffer has entries sorted across the columns
instead of down the rows."
  (let (select buffer)
    (save-window-excursion
      (Electric-pop-up-window
       (setq buffer
	     (shell-make-completions-buffer list " *Completions*" (not arg)))
       shell-electric-max-height)
      (unwind-protect
	  (progn
	    (set-buffer buffer)
	    (setq select
		  (catch 'shell-completion-menu-select
		    (let ((first (point-min))
			  (last (progn (goto-char (point-max))
				       (forward-line -1)
				       (end-of-line)
				       (point)))
			  (goal-column nil))
		      (goto-char first)
		      ;;
		      ;; The last argument is what gets passed to the loop
		      ;; function as the first argument.
		      ;;
		      (Electric-command-loop 'shell-completion-menu-select
					     nil
					     t
					     'shell-completion-menu-looper
					     (cons first last))))))
	(set-buffer buffer)
	(text-mode)
	(bury-buffer buffer)
	(message "")))
    (if select
	(save-excursion
	  (set-buffer buffer)
	  (let* ((bound (save-excursion (end-of-line) (point)))
		 (end (re-search-forward "[^ \t]+" bound)))
	    (if end
		(list (buffer-substring select (point)))
	      nil)))
      )))

(defun shell-completion-menu-looper (state condition)
  "Looper function for electric completion commands."
  (cond ((and condition
	      (not (memq (car condition) '(buffer-read-only
					   end-of-buffer
					   beginning-of-buffer))))
	 (signal (car condition) (cdr condition)))
	((< (point) (car state))
	 (goto-char (point-min)))
	((> (point) (cdr state))
	 (goto-char (point-max))
	 (forward-line -1)
	 (if (pos-visible-in-window-p (point-max))
	     (recenter -1)))))

(put 'shell-completion-menu-mode 'mode-class 'special)


(defun shell-completion-menu-mode ()
  "
 Major mode for editing an electric completion buffer.

 Movement in the buffer is done with \\[next-line], \\[previous-line],
\\[shell-completion-forward-field], and
\\[shell-completion-backward-field].

 Selection for append to the command line is performed with
\\[shell-completion-menu-select].

\\{shell-completion-menu-mode-map}
"
  (kill-all-local-variables)
  (use-local-map shell-completion-menu-mode-map)
  (make-local-variable 'shell-completion-field-width)
  (setq shell-completion-field-width nil)
  (setq mode-name "Electric Completion Menu")
  (setq mode-line-buffer-identification "Electric Completion List")
  (if (memq 'mode-name mode-line-format)
      (progn (setq mode-line-format (copy-sequence mode-line-format))
	     (setcar (memq 'mode-name mode-line-format) "Names")))
  (setq truncate-lines t)
  (setq buffer-read-only t)
  (setq major-mode 'shell-completion-menu-mode)
  (goto-char (point-min))
  (run-hooks 'shell-completion-menu-mode-hook))

(defvar shell-completion-menu-mode-map nil)
(put 'shell-completion-menu-undefined 'suppress-keymap t)

(if shell-completion-menu-mode-map
    (setq shell-completion-menu-mode-map nil))

(let ((map (make-keymap))
      (meta-map (make-keymap)))
  (fillarray map 'shell-completion-menu-undefined)
  (define-key map "\e" meta-map)
  (fillarray (lookup-key map "\e") 'shell-completion-menu-undefined)
  (define-key map "\C-h" 'Helper-help)
  (define-key map "?" 'Helper-describe-bindings)
  (define-key map "q" 'shell-completion-menu-quit)
  (define-key map "\C-c" nil)
  (define-key map "\C-c\C-c" 'shell-completion-menu-quit)
  (define-key map "\r" 'shell-completion-menu-select)  
  (define-key map " " 'shell-completion-menu-select)
  (define-key map "\C-l" 'recenter)
  (let ((i ?0))
    (while (<= i ?9)
      (define-key map (char-to-string i) 'digit-argument)
      (define-key meta-map (char-to-string i) 'digit-argument)
      (setq i (1+ i))))
  (define-key map "-" 'negative-argument)
  (define-key map "\e-" 'negative-argument)
  (define-key map "\C-u" 'universal-argument)
  (define-key map "\C-p" 'previous-line)
  (define-key map "\C-n" 'next-line)
  (define-key map "\C-f" 'shell-completion-forward-field)
  (define-key map "\C-b" 'shell-completion-backward-field)
  (define-key map "\C-a" 'beginning-of-line)
  (define-key map "\M-r" 'move-to-window-line)

  (define-key map "p" 'previous-line)
  (define-key map "n" 'next-line)
  (define-key map "b" 'shell-completion-backward-field)
  (define-key map "f" 'shell-completion-forward-field)
  (define-key map "\C-v" 'scroll-up)
  (define-key map "\M-v" 'scroll-down)
  (define-key map "\e>" 'end-of-buffer)
  (define-key map "\e<" 'beginning-of-buffer)
  (define-key map "\e\e" 'eval-expression)
  (setq shell-completion-menu-mode-map map))

(defun shell-completion-menu-undefined ()
  (interactive)
  (ding)
  (message (if (and (eq (key-binding "\C-c\C-c") 'shell-completion-menu-quit)
		    (eq (key-binding " ") 'shell-completion-menu-select)
		    (eq (key-binding "\C-h") 'Helper-help)
		    (eq (key-binding "?") 'Helper-describe-bindings))
	       "Type C-c C-c to exit, Space to select, C-h for help, ? for commands"
	     (substitute-command-keys "\
Type \\[shell-completion-menu-quit] to exit, \
\\[shell-completion-menu-select] to select, \
\\[Helper-help] for help, \\[Helper-describe-bindings] for commands.")))
  (sit-for 4))

(defun shell-completion-menu-quit ()
  "Leave Electric Completion Menu"
  (interactive)
  (throw 'shell-completion-menu-select nil))

(defun shell-completion-menu-select ()
  "Recall the command on the current line to the shell buffer.  If
marks have been made in the buffer, they are processed.  If a prefix
argument is given, marks are ignored and the history command
corresponding to that number is recalled."
  (interactive)
  (throw 'shell-completion-menu-select (point)))

(defun shell-completion-forward-field (arg)
  "Moves forward ARG fields in the shell completion buffer."
  (interactive "p")
  (let ((neg (< arg 0)))
    (setq arg (abs arg))
    (while (> arg 0)
      (shell-completion-field-move (if neg -1 1))
      (setq arg (1- arg)))))

(defun shell-completion-backward-field (arg)
  "Move backward ARG fields in the shell completion buffer."
  (interactive "p")
  (shell-completion-forward-field (- arg)))

(defun shell-completion-field-move (arg)
  "This is a lower level function called by the interactive functions which
perform horizontal field movement."
  (let ((goal-column nil)
	fwidth pt eol bol)
    (setq fwidth shell-completion-field-width)
    (setq pt (point))
    (cond ((= 0 arg)
	   nil)
	  ((> arg 0)
	   (setq eol (save-excursion (end-of-line) (point)))
	   (if (= eol (+ pt fwidth))
	       (forward-line 1)
	     (forward-char fwidth)))
	  (t
	   ;; arg < 0
	   (setq bol (save-excursion (beginning-of-line) (point)))
	   (if (= pt (point-min))
	       ;; Beginning of buffer case.
	       nil
	     (if (= bol pt)
		 (progn (forward-line -1)
			(end-of-line)
			(backward-char fwidth))
	       (backward-char fwidth))))
	  )))
