;;;
;;;				NO WARRANTY
;;; 
;;; This software is distributed free of charge and is in the public domain.
;;; Anyone may use, duplicate or modify this program.  Thinking Machines
;;; Corporation does not restrict in any way the use of this software by
;;; anyone.
;;; 
;;; Thinking Machines Corporation provides absolutely no warranty of any kind.
;;; The entire risk as to the quality and performance of this program is with
;;; you.  In no event will Thinking Machines Corporation be liable to you for
;;; damages, including any lost profits, lost monies, or other special,
;;; incidental or consequential damages arising out of the use of this program.
;;;
;;; Thu May  4 1989
;;;
;;; $Header: /tmp_mnt/am/p7/utility/gmacs/f2/RCS/login.el,v 1.3 89/02/06 09:14:34 kahle Exp $

;; A "login" function for maintaining multiple remote sessions in shell buffers.
;; Written June 1987 by John R. Rose (Rose@THINK.COM).

;;main functions:
;;  meta-x login    ;;asks for host and protocol   
;;  meta-x rlogin   ;;asks for host
;;  meta-x run-lispm ;;asks for host

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Created by Brewster Kahle on Mon Feb  6 09:10:45 1989
;;;
;;; ******************* CHANGE LOG *******************
;;; 
;;; Added m-x run-lispm on 10/2/88 JBS
;;; added rlogin brewster
;;;
;;; *************** END OF CHANGE LOG ***************



(provide 'login)
(require 'extras) ;; lazy-switch-to-buffer

(defvar login-buffer-name-format "/%s"
  "*A format string used to name buffers created by M-x login.
   The string should have a %s in it, which will be replaced by the host name.")

(defvar login-default-protocol 'rlogin
  "*Protocol used if a no host-specific protocol is specified by the hosts file.
   See variables login-host-files and login-protocol-alist.")

(defvar login-ask-protocol 'maybe
  "*If nil, always guess which protocol \\[login] will use for each host.
   If t, always ask the user.  If neither t nor nil, ask only if unsure.")

(defconst login-protocol-alist
  '((rlogin
     ((concat exec-directory "env")
      (format "TERMCAP=emacs:co#%d:tc=unknown:" (screen-width))
      "TERM=emacs" "EMACS=t"
      "-" "rlogin" host "-e\7"))
    (local . shell)
    (shell ;;same-host "protocol"
     ((concat exec-directory "env")
      (format "TERMCAP=emacs:co#%d:tc=unknown:" (screen-width))
      "TERM=emacs" "EMACS=t"
      "-" (or (getenv "ESHELL") (getenv "SHELL") "/bin/sh") "-i"))
    (lispm . telnet) ;; Alias.  But probably want to set lisp mode eventually.
    (telnet
     ((or (getenv "ESHELL") (getenv "SHELL") "/bin/sh")
      "-c" (concat "telnet " host))))
  "*An alist of protocol names and templates, used by \\[login].
   Each entry maps a symbolic name (car entry) to a login protocol specification.
   Example protocol names are rlogin and telnet.
   Each template (cadr entry) is used to construct a shell command.
   It is a list of strings or forms evaluating to strings.
   If (caddr entry) is present, it is a Lisp form to evaluate
   in the login buffer; perhaps it sets a major mode appropriate
   to the login protocol.  (There should be some convenient way
   to specify initial commands to the remote host, but I'm not
   sure how to do this.)

   (Special rule:  If the cdr of an entry of this list is a symbol,
   it is an alias for another entry.)
   The symbol 'host is bound to the target host name during evaluation
   of the template forms.")

(defvar login-host-files
	'("/etc/hosts"
	  ("/tm/finger/lmfingerdb" lispm)
	  "~/.login-hosts")
  "*Files where host names sought, for choosing login protocols and for completion.
   The file can either be in /etc/hosts format (only the second column is used),
   or it can contain host names one per line.  A host name may be followed by
   !<proto>, where <proto> is one of the known login protocols, like rlogin or telnet.
   A list element can also be a 2-list of (<file> <proto>), in which case
   <proto> is assigned to all hosts named in that file without the !<proto> syntax.
   This list is scanned in such a way that earlier files take precedence over later.
   See the variables login-protocol-alist and login-default-protocol.")

(defvar login-host-file-limit 300
  "*Limit to the number of lines that will be parsed from the top
   of a file from login-host-files.")
(defvar login-host-obarray nil
  "Holds login host name completions.  Setup by login-parse-hosts.")

(defvar login-buffer-host nil
  "Local variable which holds the host logged into by the login command.")

(defvar login-buffer-protocol nil
  "Local variable which holds the protocol being used by the login command.")



(defun login (host &optional noselect new-buffer-p)
  "Start a login session to the given HOST, using a shell-mode buffer.
   Possible names come from /etc/hosts.  A single dot means the local host.
   Optional flags:  If NOSELECT, don't select the buffer.
   If NEW-BUFFER-P (interactive prefix), force creation of a new buffer.
   Otherwise, we try to reuse an old buffer, preferring the current buffer."
  (interactive (list (login-read-host "Login to" login-buffer-host)
		     nil current-prefix-arg))
  (let* ((name (format login-buffer-name-format host))
	 (buf (cond
	       (new-buffer-p (generate-new-buffer
			      (cond
			       ((stringp new-buffer-p) new-buffer-p)
			       ((interactive-p)
				(read-string "Login buffer name: " name))
			       (t name))))
	       ((equal login-buffer-host host) (current-buffer))
	       (t (get-buffer-create name))))
	 (proto (login-host-protocol host login-ask-protocol))
	 (prog (login-instantiate-protocol proto host)))
    (if (and noselect (not (eq (current-buffer) buf)))
	(save-excursion
	  (switch-to-buffer buf)
	  (login-start-process prog host proto))
      (lazy-switch-to-buffer buf)
      (login-start-process prog host proto))
    buf))

(defun rlogin (host &optional noselect new-buffer-p)
  "Start an rlogin session to the given HOST, using a shell-mode buffer.
   Possible names come from /etc/hosts.  A single dot means the local host.
   Optional flags:  If NOSELECT, don't select the buffer.
   If NEW-BUFFER-P (interactive prefix), force creation of a new buffer.
   Otherwise, we try to reuse an old buffer, preferring the current buffer."
  (interactive (list (login-read-host "Login to" login-buffer-host)
		     nil current-prefix-arg))
  (let ((login-ask-protocol nil))
    (login host noselect new-buffer-p)))

(defun login-read-host (prompt &optional default)
  (setq prompt (concat prompt (and default (format " [%s]" default)) ": "))
  (or login-host-obarray
      (login-parse-hosts login-host-files
			 login-host-file-limit))
  (let ((res (completing-read prompt login-host-obarray)))
    (or (plusp (length res))
	(and (plusp (length default))
	     (setq res default))
	(setq res "."))
    (intern res login-host-obarray)
    (if (string-equal res ".")
	(setq res (system-name)))
    res))



(defun login-start-process (prog &optional host proto)
  "Start up the login program PROG in the current buffer,
   putting it into shell mode, and recording the login host as HOST.

   Note:  If PROG is a list whose first element is a non-string,
   evaluate that non-string for effect in the current buffer.
   Such a PROG is produced sometimes by login-instantiate-protocol."
  (let ((proc nil)
	(setup-vars
	  (function
	    (lambda ()
	      (make-local-variable 'login-buffer-host)
	      (setq login-buffer-host host)
	      (make-local-variable 'login-buffer-protocol)
	      (setq login-buffer-protocol proto)
	      ))))
    (goto-char (point-max))
    (setq proc (get-buffer-process (current-buffer)))
    (if (and proc
	     (or (y-or-n-p "Kill previous process in this buffer? ")
		 (error "Can't have two login processes in one buffer!")))
	(progn
	  (insert ?\n "Killing previous login process...")
	  (set-marker (process-mark proc) (point-max))
	  (delete-process proc)
	  (let ((count 5))
	    (while (and (plusp (setq count (1- count)))
			(memq (process-status proc) '(run exit))) ;~signal
	      (accept-process-output)
	      (sleep-for 1))
	    (or (plusp count)
		(insert (format "  Process status: %s\n" (process-status proc)))))
	  (goto-char (point-max))
	  (insert ?\n)
	  (setq proc nil)))
    (if (null proc)
	(let ((setup-form nil))
	  ;; Decode hook for major mode:
	  (if (and (listp prog) (not (stringp (car prog))))
	      (setq setup-form (car prog) prog (cdr prog)))
	  ;; Handle simple single-string case:
	  (if (stringp prog) (setq prog (list "sh" "-c" prog)))
	  ;; Otherwise, it's a list of strings to pass to start-process.
	  (setq proc (apply 'start-process name (current-buffer) prog))
	  (process-kill-without-query proc)
	  (set-marker (process-mark proc) (point-max))
	  ;; This call resets a lot of buffer state, including local variables:
	  (shell-mode)
	  (funcall setup-vars)
	  ;; Now, from this consistent state, call the major mode hook:
	  (eval setup-form))
      (funcall setup-vars)
      )))



(defun login-host-protocol (host &optional ask-protocol)
  "Find the protocol for the given HOST.
   If ASK-PROTOCOL is true (as with login-ask-protocol) either
   prompt the user for a protocol."
  (and (symbolp host) (setq host (symbol-name host)))
  (setq host (intern host login-host-obarray))
  (let ((proto (get host 'login-protocol))
	(default (or (and (equal login-buffer-host
				 (symbol-name host))
			  login-buffer-protocol)
		     login-default-protocol))
	(asked nil))
    (and proto (setq default proto))
    (if (and ask-protocol
	     (or (null proto)
		 (eq ask-protocol t)))
	(let ((alist (mapcar
		       (function
			 (lambda (elem)
			   (setq elem (car elem))
			   (cons (symbol-name elem) elem)))
		       login-protocol-alist)))
	  (setq asked (completing-read
			(format "Protocol for %s [%s]: " host default) alist))
	  (if (zerop (length asked))
	      (setq asked default)
	    (setq asked (cdr (assoc asked alist))))))
    (if asked (put host 'login-protocol asked))
    (or asked proto default
	(error "No login protocol for %s" host))))

(defun login-instantiate-protocol (proto host)
  "Decode a protocol specification PROTO (a symbol like 'rlogin)
   relative to a particular HOST.  Can return a list which
   login-start-process expects:  (setup-form . strings)."
  (and (symbolp host) (setq host (symbol-name host)))
  (while (symbolp proto)
    (setq proto (or (cdr (assq proto login-protocol-alist))
		    (error "Unknown protocol %s for %s" proto host))))
  (if (or (atom proto) (nthcdr 2 proto))
      (error "Invalid protocol specifier:  %s" proto))
  (let ((strings (nth 0 proto)))
    (if (stringp strings)
	(setq strings (format strings host))
      (let ((host host));; for template
	(setq strings (mapcar 'eval strings))))
    (let ((setup-form (nth 1 proto)))
      (and (stringp setup-form) (setq setup-form nil))
      (if setup-form
	  (cons setup-form strings)
	strings))))

;; This stuff is used to parse the /etc/hosts file for possible host names:
(defvar hosts-mode-syntax-table nil)
(defun hosts-mode ()
  "Major mode for /etc/hosts and similar files."
  (interactive)
  (text-mode)
  (if hosts-mode-syntax-table
      (set-syntax-table hosts-mode-syntax-table)
    (set-syntax-table
     (setq hosts-mode-syntax-table (copy-syntax-table)))
    (modify-syntax-entry ?\# "<")
    (mapcar
     (function
      (lambda (ch) (modify-syntax-entry ch "w")))
     '(?\. ?\- ?\_)))
  (setq major-mode 'hosts-mode)
  (setq mode-name "Hosts"))

(defun login-parse-hosts-loop (nhosts limit default-proto)
  (hosts-mode)
  (let ((n (mod nhosts 100))
	host proto
	(max (if (integerp limit)
		 (save-excursion
		   (forward-line limit)
		   (point))
	       (point-max))))
    (while (< (point) max)
      (setq host nil ch (following-char))
      (cond
       ((null ch)) ;;End of buffer.
       ((= ch ?\#) (forward-line 1))
       ;; A bang with no host specifies a new protocol for all following hosts:
       ((= ch ?\!)
	(setq proto
	      (buffer-substring
		(progn (skip-chars-forward "! \t") (point))
		(progn (forward-word 1) (point))))
	(setq default-proto (and (plusp (length proto)) (intern proto))))
       ((and (<= ?0 ch) (<= ch ?9)) (forward-word 1))
       ((eq (char-syntax ch) ?w)
	(setq host (buffer-substring
		    (point) (progn (forward-word 1) (point))))
	;; only take the first one in each line:
	(forward-line 1))
       (t (forward-char 1) (forward-word 1) (forward-word -1))
       ;; Note:  Use just one regexp, so its compilation can be cached.
       ((not (looking-at "\\([ \t\n]+\\|#.*\\|[0-9][0-9.]*\\|[a-zA-Z][^ \t\n]*\\)"))
	(error "Bad host file!"))
       ((and (<= ?0 ch) (<= ch ?9)))
       ((memq (char-syntax ch) '(?_ ?w))
	(setq host (buffer-substring (match-beginning 0) (match-end 0)))))
      ;;(goto-char (match-end 0))
      (if host
	  (progn
	    (setq host (intern host login-host-obarray))
	    ;; Check for a protocol:
	    (setq proto nil)
	    (skip-chars-forward " \t")
	    (if (eql ?\! (following-char))
		(setq proto
		      (buffer-substring
			(progn (skip-chars-forward "! \t") (point))
			(progn (forward-word 1) (point)))))
	    (setq proto (and (plusp (length proto)) (intern proto)))
	    (or proto (setq proto default-proto))
	    (and proto (put host 'login-protocol proto))
	    (setq nhosts (1+ nhosts)
		  n (1+ n))
	    (if (eq n 100)
		(progn
		  (message "Parsing host names...%d" nhosts)
		  (setq n 0))))))
    nhosts))

(defun login-parse-hosts (file &optional limit default-proto)
  "Arrange to read host names for \\[login] from the given FILE.
   Non-interactively, LIMIT gives the maximum number of lines to parse,
   and DEFAULT-PROTO is the protocol to assign by default.
   FILE can also be a file list, in the format of login-host-files."
  (interactive "fParse host names from what file: ")
  (if (null login-host-obarray)
      (progn
        (setq login-host-obarray (make-vector 2011 0)))
	;; Always include the local name:
	(put (intern (system-name) login-host-obarray)
	     'login-protocol 'local)
	;; And "." is an alias for the current machine:
	(intern "." login-host-obarray))
  (let ((nhosts 0) host ch proto)
    (if (listp file)
	(let ((ptr (reverse file))) ;; So 1st file gets final say.
	  (while ptr
	    (setq file (car ptr) ptr (cdr ptr))
	    (if (listp file)
		(setq proto (car (cdr file)) file (car file))
	      (setq proto nil))
	    (if (file-exists-p file)
		(setq nhosts
		      (+ nhosts
			 (login-parse-hosts
			   file limit (or proto default-proto)))))))
      ;; This is the work done on each file:
      (if (equal file (buffer-file-name))
	  (progn
	    (message "Continue parsing host names...")
	    (setq nhosts (login-parse-hosts-loop 0 limit default-proto)))
	(save-excursion
	  (message "Parsing host names from %s..." file)
	  (find-file file)
	  (goto-char (point-min))
	  (setq nhosts (login-parse-hosts-loop 0 limit default-proto))
	  (bury-buffer) ;; Make it go far away.
	  )))
    (message "Parsing host names...done")
    nhosts))

(defun add-protocol-for-host (host protocol &optional do-not-add-if-already-there)
  (and (symbolp host) (setq host (symbol-name host)))
  (setq host (intern host login-host-obarray))
  (unless (and do-not-add-if-already-there (get host 'login-protocol))
    (put host 'login-protocol protocol)))

(defun run-lispm (host)
  "Run an inferior Lisp process on a LISPM."
  (interactive (list (login-read-host "Login to" login-buffer-host)))
  (require 'lisp-hacks)
  (add-protocol-for-host host 'telnet)
  (login host)
  (inferior-lisp-mode)
  (setq mode-name (concat "Inferior Lisp (" host ")"))
  (let ((buffer (get-buffer "*lisp*"))
	)
    (when buffer
      (let ((proc (get-buffer-process buffer)))
	(when (or (null proc)
		  (not (memq (process-status proc) '(run stop)))
		  )
	  (if proc (delete-process proc))
	  (kill-buffer buffer)
	  (setq buffer nil)
	  )))
    (unless buffer
      (rename-buffer "*lisp*")
      ))
  )
