;;; -*- Emacs-Lisp -*- 
;;; modify.el - simple code for maintaining modification tags in files

(defconst modify-version (substring "$Revision: 1.21 $" 11 -2) 
  "The revision number of modify.el - simple code for maintaining a
modification tag in files.  Please mention this when sending reports to
dodd@roebling.poly.edu. Complete RCS identity is

   $Id: modify.el,v 1.21 1993/02/01 13:17:06 dodd Exp $

This file is available via anonymous ftp in:

   /roebling.poly.edu:/pub/modify.el")

;;; Copyright (C) 1993 Lawrence R. Dodd
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Based on code posted by merlyn@iWarp.intel.com (Randal L. Schwartz)

;;; $Modified: Mon Feb  1 08:07:48 1993 by dodd $
;;; $Id: modify.el,v 1.21 1993/02/01 13:17:06 dodd Exp $
;;; $Revision: 1.21 $ 

;;;   Lawrence R. Dodd <dodd@roebling.poly.edu>
;;;   Department of Chemical Engineering
;;;   Polymer Research Institute
;;;   Polytechnic University 
;;;   Brooklyn, New York, 11201 USA

;;; INSTALLATION AND USAGE: 
;;;  
;;; Save as `modify.el' in a directory where emacs can find it. Stick 
;;; 
;;;                   (require 'modify) 
;;;                   
;;; in your ~/.emacs. Set `update-last-edit-date' to a value of `t,' `nil,' or 
;;; `ask' for each major mode, including the default-major-mode, in their 
;;; respective mode-hooks or make it a default by editing the value in this 
;;; file or sticking _one_ of the following in your ~/.emacs file.
;;;  
;;;        (setq-default update-last-edit-date t)
;;;        (setq-default update-last-edit-date nil)
;;;        (setq-default update-last-edit-date 'ask)
;;;  
;;; This variable has a unique value for each buffer it is visiting.

;;; Once installed, this code will update a time stamp string, like the one at
;;; the top of this file, in a manner similar to the way RCS updates its tags.

;;; If you stick the string "$Modified$" anywhere near the top of a file
;;; modify.el will replace it with "$Modified: time_saved by user_name $" when
;;; the filed is saved.  It will then update the time and user name each
;;; subsequent saving of the file.  Finally, if the variable
;;; `modify-insert-tag' is `t', `update-last-edit-date' is also `t', and the
;;; $Modified$ tag does not exist, then this code will make the top line of
;;; the file $Modified: current_time by user_name $ which, if necessary, is 
;;; prefixed by a comment character.

;;; first, prepend to write-file-hooks - checking first whether or not it is
;;; already there
(or (memq 'update-last-edit-date write-file-hooks)
    (setq write-file-hooks (append '(update-last-edit-date) write-file-hooks)))

;;; user defined variables 
(defvar modify-tag "Modified"
  "*A string representing part of the label of the tag to be replaced
replaced.  The modified string will be \"$modify-tag: time by user $\".
Formally was the string \"LastEditDate\"")

(defvar modify-insert-tag nil
  "*If true then a modification tag is inserted at the top of the file if one
does not already exist and `update-last-edit-date' is also true. This is a
buffer local variable.")
(make-variable-buffer-local 'modify-insert-tag)

(defvar update-last-edit-date nil
  "*If `t' then this buffer should have its modify-tag automatically
updated. If not `t' or `nil' (for example 'ask), then ask user. This is a
buffer local variable.")
(make-variable-buffer-local 'update-last-edit-date)

;;; the work horse 
(defun update-last-edit-date ()
  "Update `LastEditDate' tag while writing out a modified file.
Intended to be used as one of the `write-file-hooks' values. Does not write
file; return value if `nil' unconditionally."
  
  ;; if the variable is `nil' then we will exit
  (if update-last-edit-date
	 
      (let (limit separator originally_locked case-fold-search)

	;; IF WE CAN LOCATE IT...
	(if (and
	     
	     ;; can we locate the tag? - meat of package 
	     ;; search is case-sensitive
	     
	     (save-excursion
	       (setq limit (min 2000 (point-max)))
	       (goto-char (point-min))
	       (or
		;; does it have date already inserted?
		(and (re-search-forward
		      (concat "\\(\\$" modify-tag ":\\)"
			      "[^\n\\$]*" "\\(\\$\\)") limit t)
		     (setq separator " "))
		;; perhaps it is in its raw form of $modify-tag$?
		(and (re-search-forward
		      (concat "\\(\\$" modify-tag "\\)\\(\\$\\)") limit t)
		     (setq separator ": "))))
	     
	     ;; we found that tag, now if not explicitly equal to `t' then we 
	     ;; must ask
	     (or (eq t update-last-edit-date)
		 (y-or-n-p "Replace Modify Tag? ")))
	    
	    ;; ...THEN REPLACE IT
	    
	    (save-excursion

	      ;; make buffer writable if necessary
  	      (setq originally_locked (if buffer-read-only (toggle-read-only)))

  	      (setq case-fold-search t)

	      (replace-match
	       (concat "\\1" separator (current-time-string) " by "
		       (user-login-name) " \\2") t)
		
	      ;; if the buffer was originally read-only, then make it so again
  	      (if originally_locked (toggle-read-only)))

	;; ...ELSE SHOULD WE INSERT IT?
	(if modify-insert-tag 
	    ;; insert tag
	    (save-excursion
	      (goto-char (point-min))
	      (insert-string
	       (concat comment-start "$" modify-tag ": "
		       (current-time-string) " by " (user-login-name)
		       " $\n")))))))

  ;; force nil return value:
  nil)

;;; interactive defun - I have never used this - lrd
(defun update-last-edit-date-save-buffer (&rest args)
  "Invoke `save-buffer' setting a nil `update-last-edit-date' to the value of
'ask. All arguments (including prefix if interactive) are passed onto
`save-buffer'."
  (interactive (list 'INTERACTIVE current-prefix-arg))
  (if (eq 'INTERACTIVE (elt args 0))
      (let ((prefix-arg (elt args 1)))
	(or update-last-edit-date
	    (setq update-last-edit-date 'ask))
	(call-interactively 'save-buffer))
    (apply 'save-buffer args)))

;;; provide package
(provide 'modify)
