;;; -*-Emacs-Lisp-*-
;; This file is part of GNU Emacs.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the GNU Emacs General Public
;; License for full details.

;; Everyone is granted permission to copy, modify and redistribute
;; GNU Emacs, but only under the conditions described in the
;; GNU Emacs General Public License.   A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; From: (Len Weisberg)
;; Date: Mon, 30 Apr 1990 19:12:45 GMT
;; Subject: next-line-noextend (Yet another)
;; Organization: Hewlett-Packard
;;
;; A few months back, there was a discussion of versions of next-line which
;; do not extend the buffer.  Here is my version, which I think improves 
;; somewhat on others I have seen:
;;  1) It beeps when it cannot fulfill the entire move request, thereby 
;;       stopping keyboard macros when they run out of file to work on.
;;  2) This package includes a new version of previous-line, which also beeps.
;;  3) It correctly remembers the goal column.
;;  4) It works by modifying a lower level routine to take advantage of some
;;       useful information provided by the underlying C routine.
;;
;; Please read the NOTE: and BUG: sections for a little more information.
;; There is indeed a tiny bug, but it is in an area which has not been dealt
;; with at all in previous versions I have seen.
;;
;; I offer this version for inclusion in HUGE. 
;; Please let me know (here or by email) of any problems you may find
;; or suggestions for improvement.
;;
;; - Len Weisberg
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;  ----------------------------------------
(defun next-line-noextend (arg)
    ;;  NOTE: If this routine is renamed, then the test near the start of
    ;;  line-move-with-residue needs to be modified.
    ;;
    ;;  NOTE:  This version of next-line improves on the version which comes
    ;;	 with 18.55 in two ways:
    ;;	 1: It never modifies the file.
    ;;	 2: It beeps when it cannot fulfill the entire move request,
    ;;	       thereby stopping keyboard macros when they run out of
    ;;	       file to work on.
    ;;
    ;;  BUG:   This routine fails to beep in the single case in which the
    ;;       file does NOT end in line-feed AND the argument is EXACTLY one
    ;;       more than the number of lines left.  To fix this would require
    ;;       changing the C-routine forward-line, or quite a bit of extra
    ;;       checking for every line move.
  "Move cursor vertically down ARG lines.
If there is no character in the target line exactly under the current column,
the cursor is positioned after the character in that line which spans this
column, or at the end of the line if it is not long enough.
If there are not enough lines in the buffer after this one,
point is moved to the last line, there is a beep and a message.
This function never modifies the buffer.

The command \\[set-goal-column] can be used to create
a semipermanent goal column to which this command always moves.
Then it does not try to move vertically.

If you are thinking of using this in a Lisp program, consider
using `forward-line' instead.  It is usually easier to use
and more reliable (no dependence on goal column, etc.)."

  (interactive "p")
  (if  (line-move-with-residue arg)
      (progn
	(beep)
	(message "End of buffer")))
  nil)

;;;  ----------------------------------------
(defun previous-line-with-beep (arg)
    ;;  NOTE: If this routine is renamed, then the test near the start of
    ;;  line-move-with-residue needs to be modified.
  "Move cursor vertically up ARG lines.
If there is no character in the target line exactly over the current column,
the cursor is positioned after the character in that line which spans this
column, or at the end of the line if it is not long enough.
If there are not enough lines in the buffer before this one,
point is moved to the first line, there is a beep and a message.
This function never modifies the buffer.

The command \\[set-goal-column] can be used to create
a semipermanent goal column to which this command always moves.
Then it does not try to move vertically.

If you are thinking of using this in a Lisp program, consider using
`forward-line' with negative argument instead..  It is usually easier
to use and more reliable (no dependence on goal column, etc.)."
  (interactive "p")
  (if  (line-move-with-residue (- arg))
      (progn
	(beep)
	(message "Start of buffer")))
  nil)

;;;  ----------------------------------------
(defun line-move-with-residue (arg)
  "Internal routine used by interactive commands for moving by lines.
This version is a modification of line-move, making use of the value
returned by the C-routine forward-char."

  (let ((residue 0)
	(oldarg arg)
	(opoint (point)))       
    ;;  NOTE: this next test needs to be modified if the calling routines
    ;;      are renamed.
    (if (not (or (eq last-command 'next-line)
		 (eq last-command 'next-line-noextend)
		 (eq last-command 'previous-line-with-beep)
		 (eq last-command 'previous-line)))
	(setq temporary-goal-column
	      (if (and track-eol (eolp))
		  9999
		(current-column))))
    (if (not (integerp selective-display))
	(setq residue (forward-line arg))
      ;; Move by arg lines, but ignore invisible ones.
      (while (and (> arg 0)
		  (eq residue 0))		
	(vertical-motion 1)
	(forward-char -1)
	(setq residue (forward-line 1))  ;; arg will be adjusted below
	(setq arg (1- arg)))
      (while (< arg 0)
	(vertical-motion -1)
	(beginning-of-line)
	(setq arg (1+ arg)))
      (setq residue (+ residue arg)))	;; incr in case of residue above
    (move-to-column (or goal-column temporary-goal-column))
    (if (or (= oldarg 0)
	    (and (= residue 0)
		 (/= opoint (point))))
	(setq residue nil))	     
    residue))
