;;; Polynomial routines by Ove Ewerlid (ewerlid@mizar.DoCS.UU.SE).
;;;
;;; See below for the supported keybindings.
;;;
;;; Currently the below routines does not force all divisions
;;; to yield fractions. 
;;;
;;; Coefficients of the polynomials must integers or fractions.
;;; Proper operation is not guaranteed with floats due to rounding.
;;;


;;; Primitives from plain calc

(defmacro calc-slow-wrapper (&rest body)
  (list 'calc-do (list 'function (append (list 'lambda ()) body)) (point)))

(defun calc-clear-command-flag (f)
  (setq calc-command-flags (delq f calc-command-flags)))

;;; Key bindings

(define-key calc-mode-map "aaq" 'calc-poly-div-q)	;; return quotient
(define-key calc-mode-map "aar" 'calc-poly-div-r)	;; return remainder
(define-key calc-mode-map "aad" 'calc-poly-div-tot)	;; return q and r
(define-key calc-mode-map "aag" 'calc-poly-gcd)		;; return gcd
(define-key calc-mode-map "aam" 'calc-merge-exp-simp)	;; --> num/den form

;;; Interactive interface routines.

(defun calc-merge-exp-simp ()
  "Merge/expand/simplify numerator and denominatior separately.
Return result to top of stack."
  (interactive)
  (calc-slow-wrapper
   (calc-enter-result 1 "mes" (list 'math-simp-merge-quotes
				    (calc-top-n 1)))))

(defun calc-poly-gcd ()
  "Find GCD of polynomial in level 2 and 1.
Return result to top of stack."
  (interactive)
  (calc-slow-wrapper
   (calc-enter-result 2 "pgcd" (list 'math-poly-gcd1
				     (calc-top-n 2)
				     (calc-top-n 1)))
   ))

(defun calc-poly-div-q ()
  "Divide polynomial in level 2 with ditto in level 1.
Return quotient to top of stack."
  (interactive)
  (calc-slow-wrapper
   (calc-enter-result 2 "pqoutient" (list 'math-poly-div-q1
					  (calc-top-n 2)
					  (calc-top-n 1)))
   ))

(defun calc-poly-div-r ()
  "Divide polynomial in level 2 with ditto in level 1.
Return remainder to top of stack."
  (interactive)
  (calc-slow-wrapper
   (calc-enter-result 2 "premainder" (list 'math-poly-div-r1
					   (calc-top-n 2)
					   (calc-top-n 1)))
   ))

(defun calc-poly-div-tot ()
  "Divide polynomial in level 2 with ditto in level 1.
Return quotient plus remainder to top of stack."
  (interactive)
  (calc-slow-wrapper
   (calc-enter-result 2 "pdiv" (list 'math-poly-div-tot1
				     (calc-top-n 2)
				     (calc-top-n 1)))
   ))
    
;;; Interface to the raw calculation routines.

;;; Return gcd of two polynomials
(defun math-poly-gcd1 (pn pd)
  (math-poly-gcd pn pd))

;;; Return only quotient to top of stack (nil if zero)
(defun math-poly-div-q1 (pn pd)
  (let ((res (math-poly-div pn pd)))
    (cond ((null (car (car res))) '0)
	  (t (car res)))))

;;; Return only remainder to top of stack
(defun math-poly-div-r1 (pn pd)
  (let ((res (math-poly-div pn pd)))
    (cond ((null (car (cdr res))) '0)
	  (t (car (cdr res))))))

(defun math-poly-div-tot1 (pn pd)
  (let ((res (math-poly-div pn pd)))
    res))


;;; Return t if all coefficients in POLY are zero.
(defun math-poly-zerop (poly)
  (cond ((null poly) t)
	(t (and (math-zerop (car poly))
		(math-poly-zerop (cdr poly))))))

;;; Return sum of two polynomial coeff lists in a new list.	(p1 + p2)
(defun math-poly-add (p1 p2)
  (and (or p1 p2)
       (cons (math-add (or (car p1) 0) (or (car p2) 0))
	     (math-poly-add (cdr p1) (cdr p2)))))

;;; Return difference of two polynomial coeff. lists.  (p1 - p2)
(defun math-poly-sub (p1 p2)
  (and (or p1 p2)
       (cons (math-sub (or (car p1) 0) (or (car p2) 0))
	     (math-poly-sub (cdr p1) (cdr p2)))))

;;;  Recursively merge all quotients until only one remains.
;;;  Note that the recursion is designed to merge completely in one traversal.
;;;  Multiplication is carried thru in numerators and denominators.
;;;  It may sometimes not be necessary to do this in denominators.
;;;  The following identities are supported:
;;;
;;;    a   c   d a + b c	a       a + b c         c   d a + c       a   b   a + b
;;;    - + - = ---------        - + c = -------     a + - = -------       - + - = -----
;;;    b   d      b d           b          b            d      d          f   f     f
;;;    
;;;
;;;                    
;;;    a/b   a d                a/b    a             a    a d
;;;    --- = ---                --- = ---           --- = ---
;;;    c/d   c b                 c    c b           c/d    c
;;;                    
;;;
;;;    a c   a c                  b   a b           b     b a
;;;    - - = ---                a - = ---           - a = --- 
;;;    b d   b d                  c    c            c      c  
;;;


(defun math-simple-merge (expr)
;  (mpr 'MSM expr)
  (cond
    ((not (consp expr)) expr)
    ((math-primp expr) expr)
    (t
     (let ((lhs (nth 1 expr))
	   (rhs (nth 2 expr))
	   (op (car expr)))
       (cond ((memq op '(+ -))
	      (let ((lhs1 (math-simple-merge lhs))
		    (rhs1 (math-simple-merge rhs)))
		(cond ((and (eq '/ (car-safe lhs1)) (eq '/ (car-safe rhs1)))
		       (let ((a (nth 1 lhs1)) (b (nth 2 lhs1))
			     (c (nth 1 rhs1)) (d (nth 2 rhs1)))
			 (if (equal b d)
			     (cons '/ (list (cons op (list a c)) b))
			     (cons '/ (list (cons op (list
						      (math-mul-thru a d)
						      (cons '* (list b c))))
					    (cons '* (list b d)))))))
		      ((eq '/ (car-safe lhs1))
		       (let ((a (nth 1 lhs1)) (b (nth 2 lhs1)) (c rhs))
			 (cons '/ (list (cons op (list a (cons '* (list b c)))) b))))
		      ((eq '/ (car-safe rhs1))
		       (let ((a lhs1) (c (nth 1 rhs1)) (d (nth 2 rhs1)))
			 (cons '/ (list (cons op (list
						  (math-mul-thru d a) c)) d))))
		      (t (cons op (cons lhs1 (list rhs1))))
		      )))
	     ((eq '/ op)
	      (let ((lhs1 (math-simple-merge lhs))
		    (rhs1 (math-simple-merge rhs)))
		(cond ((and (eq '/ (car-safe lhs)) (eq '/ (car-safe rhs)))  
		       (let ((a (nth 1 lhs1)) (b (nth 2 lhs1))
			     (c (nth 1 rhs1)) (d (nth 2 rhs1)))
			 (cons '/ (cons (math-mul-thru a d)
					(list (list '* c b))))))
		      ((eq '/ (car-safe lhs1))
		       (let ((a (nth 1 lhs1)) (b (nth 2 lhs1)) (c rhs1))
			 (cons '/ (cons a (list (list '* c b))))))
		      ((eq '/ (car-safe rhs1))
		       (let ((a lhs1) (c (nth 1 rhs1)) (d (nth 2 rhs1)))
			 (cons '/ (cons (math-mul-thru a d) (list c)))))
		      (t (cons op (cons lhs1 (list rhs1)))))))
	     ((eq '* op)
	      (let ((lhs1 (math-simple-merge lhs))
		    (rhs1 (math-simple-merge rhs)))
		(cond ((and (memq (car-safe lhs) '(+ -))
			    (memq (car-safe rhs) '(+ -)))
		       (let ((a (nth 1 lhs1)) (b (nth 2 lhs1))
			     (c (nth 1 rhs1)) (d (nth 2 rhs1)))
			 (math-mul-thru lhs1 rhs1)))
		      ((memq (car-safe lhs1) '(+ -))
		       (let ((a (nth 1 lhs1))
			     (b (nth 2 lhs1))
			     (c rhs1))
			 (math-mul-thru lhs1 rhs1)))
		      ((memq (car-safe rhs1) '(+ -))
		       (let ((a lhs1)
			     (c (nth 1 rhs1))
			     (d (nth 2 rhs1)))
			 (math-mul-thru lhs1 rhs1)))
		      ((and (eq '/ (car-safe lhs)) (eq '/ (car-safe rhs)))  
		       (let ((a (nth 1 lhs1))
			     (b (nth 2 lhs1))
			     (c (nth 1 rhs1))
			     (d (nth 2 rhs1)))
			 (cons '/ (cons (math-mul-thru a c)
					(list (list '* c d))))))
		      ((eq '/ (car-safe lhs1))
		       (let ((a (nth 1 lhs1)) 
			     (b (nth 2 lhs1))
			     (c rhs1))
			 (cons '/ (cons (math-mul-thru a c)
					(list c)))))
		      ((eq '/ (car-safe rhs1))
		       (let ((a lhs1)
			     (c (nth 1 rhs1))
			     (d (nth 2 rhs1)))
			 (cons '/ (cons (math-mul-thru a c)
					(list d)))))
		      (t (cons op (cons lhs1 (list rhs1)))))))
	     (t expr)
	     )))))


;;; Merge all quotients and expand/simplify the numerator

(defun math-simp-merge-quotes (expr)
  (let ((quotient (math-simple-merge (math-simplify expr))))
    (cond ((and (consp quotient) (eq '/ (car quotient)))
	   (cons '/ (cons (math-sort-terms (math-simplify (nth 1 quotient)))
			  (list (nth 2 quotient)))))
	  (t (math-sort-terms (math-simplify quotient))))))


;;; Multiply two trees of terms yielding a product.
;;; The product is in the canonical form where the height is maximized to
;;; the right.  We do it by flattening each tree and then using `math-list-mul'
;;; to do the multiplication. 
;;; The plain lists are then converted to the canonical form.

(defun math-mul-thru (lhs rhs)
  (math-simplify
   (math-list-to-lnt '+ (math-list-mul (flatten-terms lhs)
				       (flatten-terms rhs)))))

(defun math-sort-terms (expr)
  (math-simplify
   (math-list-to-lnt '+ (sort (flatten-terms expr)
			      'math-abs-nbeforep))))

;;; Sort predicate used to sort list into equal symbolic order.

(defun sort-listp (arg1 arg2)
  (string< (car arg1) (car arg2)))

;;; Transform a list with negated or unnegated terms into a sort-list.
;;; '(1 2 (neg 3) 4)  --->  '( ("1" + 1) ("2" + 2) ("3" - 3) ("4" + 4) )

(defun do-sort-list (lst)
  (cond ((consp lst)
	 (cond ((eq 'neg (car-safe (car lst)))
		(cons (cons (flatten-to-string (cdr (car lst))) (cons '- (cdr (car lst))))
		      (do-sort-list (cdr lst))))
	       (t 
		(cons (cons (flatten-to-string (car lst)) (cons '+ (list (car lst))))
		      (do-sort-list (cdr lst))))))
	(t nil)))


;;; Transform a list into a Left Normalized Tree with one uniform operator.
;;; (math-list-to-lnt '+ '(1 2 3 4 5 6)) ---> (+ (+ (+ (+ (+ 1 2) 3) 4) 5) 6)
;;; Note: A list of length is returned as the car of the list.

(defun math-list-to-lnt (op lst)
  (cond ((nth 1 lst) 
	 (cons op (cons (math-list-to-lnt op (cdr lst)) (list (car lst)))))
	((null lst) nil)
	(t (car lst))
	))

;;; Transform a list into a Right Normalized Tree with one uniform operator.
;;; (math-list-to-rnt '* '(1 2 3 4 5 6)) ---> (* 1 (* 2 (* 3 (* 4 (* 5 6)))))
;;; Note: A list of length is returned as the car of the list.

(defun math-list-to-rnt (op lst)
  (cond ((nth 1 lst) 
	 (cons op (cons (car lst) (list (math-list-to-rnt op (cdr lst))))))
	((null lst) nil)
	(t (car lst))
	))

;;; Flatten a tree of terms (+ -) into a list suitable for sorting and summation.
;;; This means that differences are transformed into a sum with the
;;; rightmost part negated.

(defun flatten-terms (tree)
  (cond ((eq '+ (car-safe tree))
	 (append (flatten-terms (nth 1 tree)) (flatten-terms (nth 2 tree))))
	((eq '- (car-safe tree))
	 (append (flatten-terms (nth 1 tree))
		 (flatten-terms (cons 'neg (list (nth 2 tree))))))
	(t (list tree))))

;;; Flatten a tree of multipliers (*) into a list suitable for sorting.
;;; This means that differences are transformed into a sum with the
;;; right part negated.

(defun flatten-multipliers (tree)
  (cond ((eq '* (car-safe tree))
	 (append (flatten-multipliers (nth 1 tree))
		 (flatten-multipliers (nth 2 tree))))
	(t (list tree))))

;;; Flatten a tree into a string
;;; Symbols are converted to strings, numbers are convered to strings

(defun flatten-to-string (tree)
  (cond ((listp tree) (mapconcat 'flatten-to-string tree nil))
	((stringp tree) tree)
	((integerp tree) (int-to-string tree))
	((symbolp tree) (symbol-name tree))
	(t (print 'LOOSE))))

;;; Straight multiply of two flat lists.

(defun math-list-mul (poly1 poly2)
  (cond ((not (and poly1 poly2)) nil)
	(t (math-list-mul-rec poly1 poly2))))

(defun math-list-mul-rec (poly1 poly2)
  (cond ((null poly1) nil)
	(t
	 (append
	  (mapcar (function (lambda (x)
		    (math-simplify (math-mul x (car poly1)))
		    )) poly2)
	  (math-list-mul-rec (cdr poly1) poly2))
	 )))

;;; Before predicate ignoring signs.
(defun math-abs-beforep (a b)
  (math-beforep (math-abs a) (math-abs b)))

(defun math-abs-nbeforep (a b)
  (not (math-beforep (math-abs a) (math-abs b))))

;;; Equal predicate ignoring signs. 
(defun math-abs-equal (a b)
  (math-equal (math-abs a) (math-abs b)))

(defun math-abs-nequal (a b)
  (not (math-equal (math-abs a) (math-abs b))))


;;; Divide two polynomial lists.  Return (qoutient remainder).  
;;; If the remainder is zero return (quotient nil).
;;; If the quotient is zero return (nil remainder).
;;; The polynomials returned does not have any leading zeros.
;;; This routine is not especially efficient.

(defun math-poly-div (u v)
  (math-poly-div-rec (math-expand-tree u)
		     (math-expand-tree v)
		     0))

(defun math-poly-div-rec (u v &optional level)
  (if (null level) (setq level 0))
  (cond ((math-constp u)
	 (if (= 0 level)
	     (if (math-constp v)
		 (list (math-div u v) 0)
		 (list 0 (math-div u v)))
	     (math-div u v)))
	((math-constp v) (if (= 0 level)
			     (list (math-div u v) 0)
			     (math-div u v)))
	((math-equal u v) (if (= 0 level)
			      (list 1 0)
			      (or 1)))
	((or (math-atomic-factorp u) (math-atomic-factorp v))
	 (if (= 0 level)
	     (list (math-simplify (math-div u v)) 0)
	     (math-simplify (math-div u v))))
	(t
	 (let ((base (car (math-poly-div-base u v))))
	   (if (null base)
	       (if (= 0 level) (list (math-div u v) 0)
		   (math-div u v))
	       (let ((vp (math-is-polynomial v base)))
		 (cond
		   ((or (null base) (null vp)) (if (= 0 level)
						   (list (math-div u v) 0)
						   (math-div u v)))
		   (t
		    (let ((up (math-is-polynomial u base)))
		      (let ((res (math-poly-idiv up vp level)))
			(cond
			  ((null (nth 1 res))
			   (if (null (nth 0 res)) nil)
			   (if (= level 0)
			       (list
				(math-build-polynomial-expr (nth 0 res) base)
				0)
			       (math-build-polynomial-expr (nth 0 res) base)))
			  ((null (nth 0 res))
			   (if (null (nth 1 res)) nil)
			   (if (= level 0)
			       (list 0
				(math-build-polynomial-expr (nth 1 res) base))
			       (math-div
				(math-build-polynomial-expr (nth 1 res)
							    base) v)))
			  (t
			   (if (= level 0)
			       (list
				(math-build-polynomial-expr (nth 0 res) base)
				(math-build-polynomial-expr (nth 1 res) base))
			       (math-add
				(math-build-polynomial-expr (nth 0 res) base)
				(math-div
				 (math-build-polynomial-expr (nth 1 res) base)
				 v))))
			  )))))))))))

;;; Divide two polynomials in polynomial form.

(defun math-poly-idiv (u v &optional level)
  (if (null level) (setq level 0))
  (setq u (vconcat u))
  (setq v (vconcat v))
  (let ((m (- (length u) 1))
	(n (- (length v) 1))
	(k (- (length u) (length v)))
	(q (make-vector (1+ (- (length u) (length v))) nil))
	(j 0) (ql nil) (r nil))
    (cond ((> 0 m) nil)
	  ((> 0 n) nil)
	  ((> n m) (list nil (append u nil)))
	  ((and (= 0 m) (= 0 n))
	   (list (list (math-simplify (math-poly-div-rec (aref u 0)
						     (aref v 0)
						     (1+ level))) nil)))
	  (t
	   (while (> k -1)
	     (setq j (- (+ n k) 1))
	     (aset q k (math-simplify
			(math-poly-div-rec (aref u (+ n k)) 
				       (aref v n) (1+ level))))
	     (while (>= j k)
	       (aset u j (math-simplify
			  (math-sub (aref u j)
				    (math-mul-thru (aref q k)
						   (aref v (- j k))))))
	       (setq j (1- j)))
	     (setq k (1- k)))
	   (setq j (- m n))
	   (while (and (> j -1) (math-zerop (aref q j)))
	     (setq j (1- j)))
	   (while (> j -1)
	     (setq ql (cons (aref q j) ql))
	     (setq j (1- j)))
	   (setq j (1- n))
	   (while (and (> j -1) (math-zerop (aref u j)))
	     (setq j (1- j)))
	   (while (> j -1)
	     (setq r (cons (aref u j) r))
	     (setq j (1- j)))
;	   (mpr u v ql r)
	   (list ql r)
	   )))
  )

;;; Perform a psuedo division of polynomials

(defun math-poly-pseudo-div (u v)
  (setq u (math-exp-simp-list u))
  (setq v (math-exp-simp-list v))
;  (mpr 'PSEUDO (outp u) 'V (outp v))
  (setq u (vconcat u))
  (setq v (vconcat v))
  (let ((m (- (length u) 1))
	(n (- (length v) 1))
	(k (- (length u) (length v)))
	(j 0) (q nil) (r nil))
    (cond ((> 0 m) nil)
	  ((> 0 n) nil)
	  ((> n m) (list nil (append u nil)))
	  ((and (= 0 m) (= 0 n)) (list (append u nil) nil))
	  (t
	   (while (> k -1)
	     (setq j (- (+ n k) 1))
	     (while (> j -1)
	       (aset u j (if (>= j k)
			     (math-sub (math-mul-thru (aref v n) (aref u j))
				       (math-mul-thru (aref u (+ n k))
						      (aref v (- j k))))
			     (if (< j 0) 0
				 (math-mul-thru (aref v n) (aref u j)))))
	       (aset u j (math-simplify (aref u j)))
	       (setq j (1- j)))
	     (setq k (1- k)))
	   (let ((vn 1))
	     (setq k (- m n))
	     (setq j (- m k))
	     (while (> k -1)
	       (aset u j (math-mul-thru vn (aref u j)))
	       (setq vn (math-mul-thru vn (aref v n)))
	       (setq k (1- k))
	       (setq j (1+ j)))
	     (setq k (- m n))
	     (setq j m)
	     (while (and (> k -1) (math-zerop (aref u j))); skip leading zeroes
	       (setq k (1- k))
	       (setq j (1- j)))
	     (while (> k -1)
	       (setq q (cons (aref u j) q))
	       (setq k (1- k))
	       (setq j (1- j)))
	     (setq j (1- n))
	     (while (and (> j -1) (math-zerop (aref u j))); skip leading zeroes
	       (setq j (1- j)))
	     (while (> j -1)
	       (setq r (cons (aref u j) r))
	       (setq j (1- j)))
;	     (list u v q r)
	     (list q r)
	     ))))
  )
  
;;; Take a multivariate GCD
;;; Not a very efficient implementation.

(defun math-poly-gcd (u v)
;  (mpr 'GCD (out u) (out v))
  (cond ((math-equal u v) u)
	((math-equal u 0) v)
	((math-equal v 0) u)
	(t 
	 (let* ((base (math-poly-gcd-base u v))
		(u (math-simplify (math-expand-tree u)))
		(v (math-simplify (math-expand-tree v)))
		(ccf (math-com-con-factor-list (list u v))))
	   (cond
	     ((null base) ccf)
	     (t
	      (if (> (nth 1 (nth 0 base))
		     (nth 1 (nth 1 base)))
		  (progn (setq u1 u) (setq v1 v))
		  (progn (setq u1 v) (setq v1 u)))
	      (setq base (car (car base)))
	      (let ((gcd (math-poly-gcd-rec (math-simplify (math-div u ccf))
					    (math-simplify (math-div v ccf))
					    base)))
		(math-mul-thru ccf gcd))
	      ))))))

(defun math-poly-gcd-rec (u v base)
; (mpr 'GCDR (out u) (out v) (out base))
  (let* ((up (math-is-polynomial u base))
	 (vp (math-is-polynomial v base)))
    (cond ((and (math-constp-list up) (math-constp-list vp))
	   (math-build-polynomial-expr (math-poly-igcd up vp) base))
	  (t
	   (let ((cont (math-poly-gcd-list (append up vp))))
	     (setq u (car (math-poly-div u cont)))
	     (setq v (car (math-poly-div v cont)))
	     (setq u (math-is-polynomial u base))
	     (setq v (math-is-polynomial v base))
	     (let ((gcd (math-poly-igcd u v))
		   (cont1 nil))
	       (setq gcd (math-exp-simp-list gcd))
	       (setq cont1 (math-poly-gcd-list gcd))
	       (setq gcd (math-can-com-factor-list1 gcd cont1))
	       (setq gcd (math-cancel-neg (math-simplify
					   (math-expand-tree
					    (math-build-polynomial-expr
					     gcd base)))))
	       (math-mul-thru cont gcd)
	       )))
	  )))

;;; Run GCD on all elements in a list.
(defun math-poly-gcd-list (lst)
  (cond ((null lst) nil)
	(t
	 (let ((len (length lst)))
	   (cond ((= len 1) (car lst))
		 ((= len 2)
		  (math-poly-gcd (nth 0 lst) (nth 1 lst)))
		 (t (math-poly-gcd (car lst)
				   (math-poly-gcd-list (cdr lst))))
		 )))))



;;; Compute GCD of two monovariate polynomial lists with integer coeffs.

(defun math-poly-igcd (u v)
  (if (> (length u) (length v))
      (math-poly-igcd1 u v)
      (math-poly-igcd1 v u)))

(defun math-poly-igcd1 (u v)
;  (mpr 'IGCD1 (outp u) (outp v))
  (let ((d (math-frac-gcd (math-com-con-factor-list u)
			  (math-com-con-factor-list v)))
	(r nil)
	(delta 0) (g 1) (h 1) (gh 1))
    (setq u (math-can-com-factor-list u d))
    (setq v (math-can-com-factor-list v d))
    (let
	((gcd
	  (catch 'return
	    (while (or t)
	      (setq delta (- (length u) (length v)))
	      (setq r (car (cdr (math-poly-pseudo-div u v))))
	      (if (eq r nil) (throw 'return v))
	      (if (= 1 (length r)) (throw 'return '(1)))
	      (setq u v)
	      (setq gh (math-mul g (math-pow h delta)))
	      (setq v (mapcar (function
			       (lambda (x) (car (math-poly-div-rec x gh)))) r))
	      (setq g (car (reverse u)))
	      (setq h (if (>= (- 1 delta) 0)
			  (math-mul (math-pow g delta)
				    (math-pow h (- 1 delta)))
			  (math-poly-div-rec (math-pow g delta)
					     (math-pow h (- delta 1)))))
	      ))))
      (math-poly-mul 
       (list d)
       (math-can-com-factor-list gcd (math-com-con-factor-list gcd)))
      )))



;;; Check if a list of expression is all constant
(defun math-constp-list (lst)
  (cond ((consp lst)
	 (and (math-constp (car lst))
	      (math-constp-list (cdr lst))))
	(t t)))
			   

;;; Return the common constant factor of a list of expressions

(defun math-com-con-factor-list (u)
  (if (consp u)
      (let ((cf (or (math-common-constant-factor (car u)) 1))
	    (len (1- (length u))))
	(setq u (cdr u))
	(while (> len 0)
	  (setq cf (math-frac-gcd
		    cf (or (math-common-constant-factor (car u)) 1)))
	  (setq u (cdr u))
	  (setq len (1- len)))
	cf)
      nil))

;;; Cancel factor from list of expressions 

(defun math-can-com-factor-list (u factor)
  (cond ((consp u)
	 (append (list (math-cancel-common-factor (car u) factor))
		 (math-can-com-factor-list (cdr u) factor)))
	(t nil)))

;;; Variant of the above  
(defun math-can-com-factor-list1 (u factor)
  (cond ((consp u)
	 (append (list (car (math-poly-div (car u) factor)))
		 (math-can-com-factor-list1 (cdr u) factor)))
	(t nil)))
  

;;; Return true if is a factor containing no terms or quotes.
;;; Ex:   a b^3 c    is a factor
;;;       a (c + 1)  is not a factor

(defun math-atomic-factorp (expr)
  (cond ((not (consp expr)) t)
	((math-primp expr) t)
	(t 
	 (cond ((eq '* (car expr)) (and (math-atomic-factorp (nth 1 expr))
					(math-atomic-factorp (nth 2 expr))))
	       ((memq (car expr) '(+ - /)) nil)
	       (t t))))
  )
;;; Given expressions a and b, find a suitable base for dividing a with b.
;;; The base must exists in both expressions.
;;; The degree in the numerator must be higher or equal than the
;;; degree in the denominator.
;;; If the above conditions not is met the quotient is just a remainder.
;;; Return nil if this is the case.
;;; Otherwise return '(var exponent).


(defun math-poly-div-base (a b)
  (let ((a-base (math-sort-poly-base-list (math-total-polynomial-base a)))
	(b-base (math-sort-poly-base-list (math-total-polynomial-base b))))
    (catch 'return
      (while a-base
	(let ((maybe (math-in-list1 (car (car a-base)) b-base)))
	  (if maybe
	      (if (>= (nth 1 (car a-base)) (nth 1 maybe))
		  (throw 'return (car a-base))))
	  (setq a-base (cdr a-base))))
      )))

;;; Same as above but return the degree of the base in both a and b.
;;; Nor is there any requirement that degree(a) > degree(b)
;;; Take the base that has the highest degree considering both a and b.
;;; ("a^20+b+x^3+a+b", "a+b^2+x^5+a^22+b^10") --> (a 20) , (a 22)
(defun math-poly-gcd-base (a b)
  (let ((a-base1 (math-sort-poly-base-list (math-total-polynomial-base a)))
	(b-base1 (math-sort-poly-base-list (math-total-polynomial-base b)))
	(flag nil))
    (cond ((or (null a-base1) (null b-base1)) nil)
	  (t
	   (cond ((< (nth 1 (car a-base1))
		     (nth 1 (car b-base1)))
		  (setq a-base b-base1) (setq b-base a-base1) (setq flag t))
		 (t 
		  (setq a-base a-base1) (setq b-base b-base1)))
	   (catch 'return
	     (while a-base
	       (let ((maybe (math-in-list1 (car (car a-base)) b-base)))
		 (if maybe
		     (throw 'return 
		       (if flag 
			   (list maybe (car a-base))
			   (list (car a-base) maybe))))
		 (setq a-base (cdr a-base))))))
	  )))

;;; Sort a list of polynomial bases
(defun math-sort-poly-base-list (lst)
  (sort lst 'math-poly-div-base-p))

(defun math-poly-div-base-p (a b)
  (> (nth 1 a) (nth 1 b)))

;;; Given an expression find all variables that is a polynomial base
;;; Return list in the form '( (var1 degree1) (var2 degree2) ... )
;;; Note dynamic scope of mpb-total-base.

(defun math-total-polynomial-base (expr)
  (let ((mpb-total-base nil))
    (math-polynomial-base expr 'math-polynomial-p1)
    mpb-total-base))

(defun math-polynomial-p1 (subexpr)
  (if (eq 'var (car subexpr))
      (let ((exponent (math-polynomial-p mpb-top-expr subexpr)))
	(if exponent
	       (if (math-in-list subexpr mpb-total-base)
		   nil
		   (setq mpb-total-base (cons (cons subexpr (list exponent))
					      mpb-total-base))))
	))
  nil)

;;; Check if elt is in lst of the form '( (elt1 foo1) (elt2 foo2) )
;;; Return t if it is otherwise nil.
(defun math-in-list (elt lst)
  (cond ((consp lst)
	 (or (equal elt (car (car lst)))
	     (math-in-list elt (cdr lst))))
	(t nil)))

;;; Check if elt is in lst of the form '( (elt1 foo1) (elt2 foo2) )
;;; Return the element in the list that matched if it is.
;;; nil otherwise.

(defun math-in-list1 (elt lst)
  (cond ((consp lst)
	 (if (equal elt (car (car lst)))
	     (car lst)
	     (math-in-list1 elt (cdr lst))))
	(t nil)))

;;; Check if -1 is a factor of the expression.
;;; The function will only find the trivial case.  (Simplified expr.)
(defun math-neg-factorp (expr)
  (cond	((math-primp expr) 
	 (or (math-lessp expr 0)
	     (math-equal expr 0)))
	((consp expr)
	 (cond ((eq (car expr) 'neg) t)
	       ((eq (car expr) '-)
		(math-neg-factorp (nth 1 expr)))))
	(t nil)))

;;; If -1 is factor of expr return expr times -1
(defun math-cancel-neg (expr)
  (if (math-neg-factorp expr)
      (math-neg expr)
      expr))

;;; Expand and simplify exprs in list
(defun math-exp-simp-list (lst)
  (mapcar (function (lambda (x) (math-simplify (math-expand-tree x)))) lst))


;;; Some routines that aid in debugging

;;; print arg into *scratch* buffer
(defun mpr (arg &optional arg1 arg2 arg3)
  (print (append (and arg (list arg))
		 (and arg1 (list arg1))
		 (and arg2 (list arg2))
		 (and arg3 (list arg3)))
	 (get-buffer "*scratch*")))      


;;; Read expression 
(defun in (expr)
  (math-read-expr expr))

;;; Read expression converting to polynomial in `base'.
(defun inp (expr &optional base)
  (if (or base)
      (let ((a  (math-read-expr expr)))
	(math-is-polynomial a base))
      (let ((a  (math-read-expr expr)))
	(math-is-polynomial a (math-polynomial-base a)))))

;;; Convert expr to readable string.
(defun out (expr)
  (math-format-value expr))

;;; Convert polynomial to readable string.
(defun outp (expr &optional base)
  (if (null base)
      (setq base '(var X var-X)))
  (math-format-value (math-build-polynomial-expr expr base)))


