;;; -*- Emacs-Lisp -*- Emacs/RCS interface
;; Copyright (C) 1987 Free Software Foundation, Inc.

;; This file is part of GNU Emacs.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the GNU Emacs General Public
;; License for full details.

;; Everyone is granted permission to copy, modify and redistribute
;; GNU Emacs, but only under the conditions described in the
;; GNU Emacs General Public License.   A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

;; $Header: rcs.el,v 1.3 88/11/09 10:37:38 GMT cph Exp $

;;; ===============================================
;;; Global Variables (can be redefined in rcs-hook)
;;; ===============================================

(defvar rcs-ci-default-switches "-u"
  "Default switches to pass to the `rcs-ci' command.")

(defvar rcs-co-default-switches "-l"
  "Default switches to pass to the `rcs-co' command.")

(defvar rcs-diff-default-switches ""
  "Default switches to pass to the `rcs-diff' command.")

(defvar rcs-log-default-switches ""
  "Default switches to pass to the `rlog' command.")

(defvar rcs-directory-alist '()
  "An association between working directories and RCS directories.
The car of each entry is a working directory, and the cdr is the
RCS directory corresponding to that working directory.")

(defvar rcs-auto-header-insertion nil
  "If non-nil, insert header at beginning of files when first checked in.")

(defvar rcs-head ">>> RCS info <<<\n"
  "RCS head banner.
This is only used when `rcs-auto-header-insertion' is non-nil.")

(defvar rcs-tail ">>> end RCS info <<<\n"
  "RCS tail banner.
This is only used when `rcs-auto-header-insertion' is non-nil.")

(defvar rcs-keywords '("Log" "Header")
  "List of RCS keywords to be included in the header.
Default is `Header' and `Log'.
This is only used when `rcs-auto-header-insertion' is non-nil.")

;;; ==================================================
;;; Other Variables (shouldn't be touched by the user)
;;; ==================================================

(defconst rcs-log-buffer-name "*RCS-log*")
(defconst rcs-message-buffer-name "*RCS*")
(defconst rcs-scratch-buffer-name " *RCS-scratch*")
(defvar rcs-comment-begin "")
(defvar rcs-comment-body "")
(defvar rcs-comment-end "")

;;;; Check In

(defun rcs-ci (switches rcs-files)
  "Check a file in to RCS.
With prefix argument, prompt for switches to give to `ci'.
Otherwise the value of `rcs-ci-default-switches' is used.
Edit log message at prompt except for the initial revision."
  (interactive
   (list (if current-prefix-arg
	     (read-string "ci switches: " "-")
	     rcs-ci-default-switches)
	 (rcs-read-file-names "RCS check-in file: " t)))
  (let* ((working-file (car rcs-files))
	 (rcs-file (cdr rcs-files))
	 (rcs-log-buffer (get-buffer-create rcs-log-buffer-name)))
    (save-window-excursion
      (if (file-exists-p rcs-file)
	  (rcs-init-existing rcs-log-buffer)
	  (rcs-init-new working-file
			rcs-file
			(file-name-directory working-file))))
    (rcs-ci-internal switches working-file rcs-file rcs-log-buffer)
    (bury-buffer rcs-log-buffer)))

(defun rcs-ci-internal (switches working-file rcs-file rcs-log-buffer)
  (let* ((working-file-name (file-name-nondirectory working-file))
	 (working-file-buffer (get-file-buffer working-file))
	 (source-buffer)
	 (cmd
	  (concat "cd "
		  (file-name-directory working-file)
		  ";ci "
		  (if (string-match "-k" switches)
		      switches
		      (rcs-default-revision-switch switches
						   "rfluq"
						   working-file
						   working-file-buffer
						   t))
		  " "
		  rcs-file))
	 (rcs-message-buffer (get-rcs-message-buffer)))
    (if working-file-buffer
	(progn (find-file working-file)
	       (save-buffer)))
    (message "Checking in %s to RCS..." working-file-name)
    (rcs-display rcs-message-buffer cmd)
    (save-excursion
      (set-buffer rcs-log-buffer)
      (call-process-region (point-min) (point-max) "/bin/sh" nil
			   rcs-message-buffer nil "-c" cmd))
    (bury-buffer rcs-message-buffer)
    (if working-file-buffer
	(rcs-revert-buffer t t))
    (message "Checking in %s to RCS...done" working-file-name)))

(defun rcs-ci-locked-files (switches working-directory)
  "Check in all locked RCS files for a given working directory.
With prefix argument, prompt for switches to give to `ci'.
Otherwise the value of `rcs-ci-default-switches' is used.
Edit log message at prompt."
  (interactive
   (list (if current-prefix-arg
	     (read-string "ci switches: " "-")
	     rcs-ci-default-switches)
	 (read-file-name "Check-in locked files for directory: "
			 default-directory
			 default-directory
			 'lambda)))
  (if (not (file-directory-p working-directory))
      (error "File not a directory: %s" working-directory))
  (if (not (string-match "/$" working-directory))
      (setq working-directory (concat working-directory "/")))
  (let ((rcs-directory
	 (rcs-expand-directory
	  (working->rcs-directory working-directory)))
	(rcs-log-buffer (get-buffer-create rcs-log-buffer-name))
	(rcs-temp-buffer nil))
    (save-window-excursion
      (rcs-init-existing rcs-log-buffer))
    (unwind-protect
	(progn 
	  (setq rcs-temp-buffer (get-buffer-create " *RCS-temp*"))
	  (save-excursion
	    (set-buffer rcs-temp-buffer)
	    (erase-buffer))
	  (call-process "/bin/sh" nil rcs-temp-buffer nil "-c"
			(concat "rlog -L -R " rcs-directory "*,v"))
	  (save-excursion
	    (set-buffer rcs-temp-buffer)
	    (goto-char (point-min))
	    (if (search-forward "rlog error" nil t)
		(error "Unable to read RCS directory" rcs-directory))
	    (while (not (eobp))
	      (let ((start (point)))
		(end-of-line)
		(save-excursion
		  (let ((rcs-file (buffer-substring start (point))))
		    (rcs-ci-internal
		     switches
		     (concat
		      working-directory
		      (let ((nondirectory (file-name-nondirectory rcs-file)))
			(substring nondirectory 0
				   (string-match ",v$" nondirectory))))
		     rcs-file
		     rcs-log-buffer)))
		(forward-line 1)))))
      (kill-buffer rcs-temp-buffer)
      (bury-buffer rcs-log-buffer))))

(defun rcs-init-existing (rcs-log-buffer)
  (pop-to-buffer rcs-log-buffer)
  (text-mode)
  (message
   (substitute-command-keys
    "Edit RCS log message, finish with \\[exit-recursive-edit]."))
  (recursive-edit)
  (goto-char (point-max))
  (if (and (not (bobp))
	   (save-excursion (backward-char 1) (not (looking-at "\n"))))
      (insert ?\n)))

(defun rcs-init-new (working-file rcs-file working-directory)
  (rcs-check-comment-string working-file) ;Determine the current mode
  (if (and rcs-auto-header-insertion
	   (not (save-excursion (search-forward "$Header" nil t))))
      (rcs-insert-header))
  (rcs-command (format "Creating RCS file %s" rcs-file)
	       "cd " working-directory
	       ";rcs -i -c\"" rcs-comment-body "\" " rcs-file))

(defun rcs-insert-header ()
  (goto-char 1)
  (insert rcs-comment-begin ?\n)
  (if rcs-head
    (insert rcs-comment-body rcs-head
	    rcs-comment-body ?\n))
  (let ((keywords (reverse rcs-keywords)))
    (while keywords
      (insert rcs-comment-body ?$ (car keywords) "$\n")
      (setq keywords (cdr keywords))))
  (if rcs-tail
    (insert rcs-comment-body rcs-tail
	    rcs-comment-body ?\n))
  (insert rcs-comment-end "\n\n")
  (goto-char 1))

(defun rcs-check-comment-string (working-file)
  "Determine the right comment symbols for the current mode.
If the mode is not known, prompt the user for the three parts of 
comment symbols."
  (let ((mode-name (rcs-get-file-mode working-file)))
    (cond ((or (eq mode-name 'TeX-mode)
	       (eq mode-name 'LaTeX-mode))
	   (rcs-set-comment-string "% "))
	  ((or (eq mode-name 'scribe-mode)
	       (eq mode-name 'texinfo-mode))
	   (rcs-set-comment-string "@comment "))
	  ((or (eq mode-name 'emacs-lisp-mode)
	       (eq mode-name 'lisp-mode)
	       (eq mode-name 'scheme-mode))
	   (rcs-set-comment-string ";;; "))
	  ((eq mode-name 'c-mode)
	   (setq rcs-comment-begin "/* ")
	   (setq rcs-comment-body " * ")
	   (setq rcs-comment-end " */"))
	  ((eq mode-name 'fortran-mode)
	   (rcs-set-comment-string "C "))
	  ((string-match "[Mm]akefile" working-file)
	   (rcs-set-comment-string "# "))
	  (t
	   (setq rcs-comment-begin
		 (read-string "Opening comment string: ")) 
	   (setq rcs-comment-body
		 (read-string "Intermediate comment string: ")) 
	   (setq rcs-comment-end
		 (read-string "Closing comment string: "))))))

(defun rcs-set-comment-string (string)
  (setq rcs-comment-begin string)
  (setq rcs-comment-body string)
  (setq rcs-comment-end string))

(defun rcs-get-file-mode (working-file)
  (let ((buffer (get-file-buffer working-file)))
    (if buffer
	(save-excursion
	  (set-buffer buffer)
	  major-mode)
	(save-excursion
	  (unwind-protect
	      (progn
		(find-file working-file)
		major-mode)
	    (kill-buffer (current-buffer)))))))

;;;; Check Out

(defun rcs-co (switches rcs-files)
  "Check a file out from RCS and visit it.
With prefix argument, prompt for switches to give to `co'.
Otherwise the value of `rcs-co-default-switches' is used."
  (interactive
   (list (if current-prefix-arg
	     (read-string "co switches: " "-")
	     rcs-co-default-switches)
	 (rcs-read-file-names "RCS check-out file: " nil)))
  (let ((working-file (car rcs-files))
	(rcs-file (cdr rcs-files)))
    (if (and (file-exists-p working-file)
	     (file-writable-p working-file)
	     (or (not (y-or-n-p (concat "Overwrite " working-file "? ")))
		 (progn (delete-file working-file)
			nil)))
	(error "abort RCS checkout"))
    (let ((working-file-name (file-name-nondirectory working-file))
	  (working-file-buffer (get-file-buffer working-file)))
      (if (and working-file-buffer
	       (buffer-modified-p working-file-buffer)
	       (not (y-or-n-p (concat "Buffer "
				      (buffer-name working-file-buffer)
				      " is modified.  Continue anyway? "))))
	  (error "abort RCS checkout"))
      (rcs-command (format "Checking out %s from RCS" working-file-name)
		   "cd "
		   (file-name-directory working-file)
		   ";co "
		   (rcs-default-revision-switch switches
						"rlq"
						working-file
						working-file-buffer
						t)
		   " "
		   rcs-file)
      (if working-file-buffer
	  (progn
	    (switch-to-buffer working-file-buffer)
	    (rcs-revert-buffer t t))
	  (find-file working-file)))))

(defun rcs-unlock (rcs-files)
  "Unlock a file that has been checked out from RCS."
  (interactive
   (list (rcs-read-file-names "RCS unlock file: " nil)))
  (let ((working-file (car rcs-files))
	(rcs-file (cdr rcs-files)))
    (let ((working-file-name (file-name-nondirectory working-file))
	  (working-file-buffer (get-file-buffer working-file)))
      (if (and working-file-buffer
	       (buffer-modified-p working-file-buffer)
	       (not (y-or-n-p (concat "Buffer "
				      (buffer-name working-file-buffer)
				      " is modified.  Continue anyway? "))))
	  (error "abort RCS unlock"))
      (rcs-command (format "Unlocking %s from RCS" working-file-name)
		   "cd " (file-name-directory working-file)
		   ";rm -f " working-file-name
		   ";rcs -u " rcs-file
		   ";co " rcs-file)
      (if working-file-buffer
	  (progn
	    (switch-to-buffer working-file-buffer)
	    (rcs-revert-buffer t t))))))

;;;; RCS Log

(defun rcs-diff (switches rcs-files)
  "Run `rcsdiff' on a file.
With prefix argument, prompt for switches to give to `rcsdiff'.
Otherwise the value of `rcs-diff-default-switches' is used."
  (interactive
   (list (if current-prefix-arg
	     (read-string "rcsdiff switches: " "-")
	     rcs-diff-default-switches)
	 (rcs-read-file-names "Run rcsdiff on file: " nil)))
  (let ((working-file (car rcs-files))
	(rcs-file (cdr rcs-files)))
    (let ((working-file-buffer (get-file-buffer working-file)))
      (rcs-command-nowait
       (format "Running rcsdiff on %s" (file-name-nondirectory working-file))
       "cd " (file-name-directory working-file)
       ";rcsdiff "
       (rcs-default-revision-switch switches
				    "r"
				    working-file
				    working-file-buffer
				    nil)
       " "
       rcs-file))))

(defun rcs-log (switches rcs-files)
  "Read log information for an RCS file.
With prefix argument, prompt for switches to give to `rlog'.
Otherwise the value of `rcs-log-default-switches' is used."
  (interactive
   (list (if current-prefix-arg
	     (read-string "rlog switches: " "-")
	     rcs-log-default-switches)
	 (rcs-read-file-names "Read RCS log for file: " nil)))
  (let ((working-file (car rcs-files))
	(rcs-file (cdr rcs-files)))
    (rcs-command-nowait
     (format "Reading RCS log for %s" (file-name-nondirectory working-file))
     "cd " (file-name-directory working-file)
     ";rlog " switches " " rcs-file)))

(defun rcs-list-locked-files (directory)
  "List locked RCS files for a given working directory."
  (interactive "DShow locked files for directory: ")
  (rcs-command-nowait (format "Listing RCS locked files for %s" directory)
		      "rlog -L -R "
		      (rcs-expand-directory (working->rcs-directory directory))
		      "*,v"))

;;;; Utilities

(defun rcs-read-file-names (prompt expand-p)
  (rcs-files
   (let ((filename (buffer-file-name)))
     (read-file-name
      (concat prompt
	      (if filename
		  (concat "(default " (file-name-nondirectory filename) ") ")
		  ""))
      default-directory
      filename
      nil))
   expand-p))

(defun rcs-files (file-name expand-p)
  "Returns a pair (WORKING-FILE . RCS-FILE)"
  (let ((files
	 (let ((directory (file-name-directory file-name))
	       (name (file-name-nondirectory file-name)))
	   (if (string-match ",v$" name)
	       (cons (concat (rcs->working-directory directory)
			     (substring name 0 -2))
		     file-name)
	       (cons file-name
		     (concat (working->rcs-directory directory) name ",v"))))))
    (if expand-p
	(rplacd files (rcs-expand-file-name (cdr files))))
    files))

(defun working->rcs-directory (working-file-directory)
  (rcs-guarantee-directory
   "RCS"
   (let ((entry (assoc working-file-directory rcs-directory-alist)))
     (if entry
	 (cdr entry)
	 (concat working-file-directory "RCS/")))))

(defun rcs->working-directory (rcs-file-directory)
  (rcs-guarantee-directory
   "working"
   (let ((entry (rassoc rcs-file-directory rcs-directory-alist)))
     (cond (entry (car entry))
	   ((string-match "/RCS/$" rcs-file-directory)
	    (substring rcs-file-directory 0 -4))
	   (t rcs-file-directory)))))

(defun rcs-guarantee-directory (type directory)
  (if (not (file-exists-p directory))
      (rcs-create-directory type directory))
  directory)

(defun rcs-create-directory (type directory)
  (if (y-or-n-p (concat type " directory " directory
			" does not exist.  Create? "))
      (rcs-command (format "Creating %s directory %s" type directory)
		   "mkdir "
		   (if (string-match "/$" directory)
		       (substring directory 0 -1)
		       directory))
      (error "abort RCS operation")))

(defun rcs-expand-file-name (file-name)
  (expand-file-name (file-name-nondirectory file-name)
		    (rcs-expand-directory (file-name-directory file-name))))

(defun rcs-expand-directory (directory)
  (cond ((file-directory-p directory) directory)
	((file-exists-p directory) (rcs-expand-file-name-1 directory))
	(t (error "Nonexistent RCS directory %s" directory))))

(defun rcs-expand-file-name-1 (file-name)
  (let ((rcs-temp-buffer nil))
    (unwind-protect
	(progn
	  (setq rcs-temp-buffer (get-buffer-create " *RCS-temp*"))
	  (save-excursion
	    (set-buffer rcs-temp-buffer)
	    (erase-buffer))
	  (call-process "/bin/sh" nil rcs-temp-buffer nil "-c"
			(concat "rcsxpnd " file-name))
	  (save-excursion
	    (set-buffer rcs-temp-buffer)
	    (goto-char (point-min))
	    (buffer-substring (point) (save-excursion (end-of-line) (point)))))
      (kill-buffer rcs-temp-buffer))))

(defun rassoc (key alist)
  (let ((result nil))
    (while (and alist (not result))
      (if (equal key (cdr (car alist)))
	  (setq result (car alist))
	  (setq alist (cdr alist))))
    result))

(defun rcs-command (string &rest args)
  (message "%s..." string)
  (rcs-command-1 string (apply 'concat args)
		 (function
		  (lambda (output-buffer program &rest args)
		    (apply 'call-process program nil output-buffer nil
			   args))))
  (message "%s...done" string))

(defun rcs-command-nowait (string &rest args)
  (rcs-command-1 string (apply 'concat args)
		 (function
		  (lambda (output-buffer program &rest args)
		    (apply 'start-process "rcs-command-nowait" output-buffer
			   program args)))))

(defun rcs-command-1 (string command process-function)
  (let ((rcs-message-buffer (get-rcs-message-buffer)))
    (rcs-display rcs-message-buffer command)
    (funcall process-function rcs-message-buffer shell-file-name "-c" command)
    (bury-buffer rcs-message-buffer)))

(defun rcs-display (buffer command)
  (pop-to-buffer buffer)
  (goto-char (point-max))
  (recenter 0)
  (insert command ?\n)
  (sit-for 0)
  (other-window 1))

(defun rcs-default-revision-switch
  (switches options working-file working-file-buffer branch-p)
  (let ((regexp (concat "-[" options "]")))
    (if (or (string-match (concat regexp "[0-9]") switches)
	    (not (or working-file-buffer
		     (file-exists-p working-file))))
	switches
	(let ((revision
	       (save-excursion
		 (if working-file-buffer
		     (progn
		       (set-buffer working-file-buffer)
		       (rcs-get-revision))
		     (progn
		       (set-buffer (get-buffer-create rcs-scratch-buffer-name))
		       (erase-buffer)
		       (insert-file working-file)
		       (prog1 (rcs-get-revision)
			 (kill-buffer (current-buffer))))))))
	  (if (or (not revision)
		  (< (length revision) 3))
	      switches
	      (let ((index
		     (and (string-match regexp switches)
			  (match-end 0)))
		    (revision (rcs-revision-to-string revision branch-p)))
		(if index
		    (concat (substring switches 0 index)
			    revision
			    (if (or (= index (length switches))
				    (= (aref switches index) ? ))
				""
				" ")
			    (substring switches index))
		    (concat "-"
			    (char-to-string (aref options 0))
			    revision
			    " "
			    switches))))))))

(defun rcs-get-revision ()
  (save-excursion
    (cond ((rcs-find-keyword "Header")
	   (skip-chars-forward "^ ")
	   (skip-chars-forward " ")
	   (rcs-parse-revision))
	  ((rcs-find-keyword "Revision")
	   (rcs-parse-revision))
	  (t nil))))

(defun rcs-parse-revision ()
  (let ((start (point)))
    (skip-chars-forward "0-9.")
    (let ((end (point))
	  (result '()))
      (goto-char start)
      (while (search-forward "." end 'move)
	(setq result
	      (cons (string-to-int
		     (buffer-substring start (match-beginning 0)))
		    result))
	(setq start (point)))
      (nreverse (cons (string-to-int (buffer-substring start end)) result)))))

(defun rcs-find-keyword (keyword)
  "Find the first occurrence of KEYWORD in the current buffer.
Returns t if found, leaving point at the beginning of the binding.
Otherwise returns nil."
  (goto-char (point-min))
  (and (let ((case-fold-search nil))
	 (search-forward (concat "$" keyword ":") (point-max) t))
       (progn (skip-chars-forward " ")
	      t)))

(defun rcs-revision-to-string (revision branch-p)
  (let ((result (int-to-string (car revision))))
    (setq revision (cdr revision))
    (while (if branch-p (cdr revision) revision)
      (setq result
	    (concat result
		    "."
		    (int-to-string (car revision))))
      (setq revision (cdr revision)))
    result))

(defun get-rcs-message-buffer ()
  (get-buffer-create rcs-message-buffer-name))

(defun rcs-revert-buffer (&optional no-auto-save-p no-confirm-p)
  "Revert the current buffer.
Like \\[revert-buffer], but tries to keep the same `window-start'."
  (interactive "P")
  (let ((start (window-start)))
    (revert-buffer no-auto-save-p no-confirm-p)
    (if (<= start (point-max))
	(save-excursion
	  (goto-char start)
	  (beginning-of-line)
	  (set-window-start (selected-window) (point))))))

(defun make-rcs-directory-alist (working-root rcs-root branch
					      &rest directories)
  ;; useful for writing an RCS hook.
  (mapcar (function
	   (lambda (directory)
	     (cons (concat working-root directory)
		   (concat rcs-root directory "RCS/"))))
	  (mapcar (function
		   (lambda (directory)
		     (concat branch "/" directory "/")))
		  directories)))

(run-hooks 'rcs-hook)