;; remote.el version 3.12
;;
;; Remote editing via rcp.
;; Written by Eric Raible (raible@orville.nas.nasa.gov),
;; based on ideas and code by Nick Tran.
;;
;; This file is distributed under the same terms as GNU Emacs.
;;
;; This file is meant to allow editing of remote files in a natural,
;; intuitive, and non-obtrusive fashion.  I put the following in my
;; .emacs:  
;;	    (global-set-key "\C-xr" 'remote-editing)
;;
;; After this package is loaded, whatever bindings you have for
;; find-file, find-file-read-only, find-alternate-file, save-buffer,
;; write-file, and insert-file will be usurped.  Use describe-key on
;; those bindings for more details.
;;
;; Note: if you have an extensive .rhosts file, check out
;; rhosts-to-alist below. 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; 11/11/89  Darryl Okahata, HP NMD
;; Modified for HP sites.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'remote)

(defvar default-remote-host nil
  "The host to use for remote file operations when none other is appropriate.
If nil: ask the name from user at run-time.")

(defvar remote-rcp-syntax "."
  "Controls the syntax rcp uses to specify files on remote machines.
If the value is \"@\", use names of the form user@machine:filename.
If the value is \".\", use names of the form machine.name:filename.
If the value is nil, then different usernames on different machines is
not handled.

Also see remote-username-alist.")

(defvar remote-username-alist
  '(("machine1" . "ericr")
    ("machine2" . "elr2"))
  "Alist describing mapping from remote machine names to user names.

Each entry of the alist is a pair of strings, where the car is the
 name of a remote machine, and the cdr is your username on that machine.
This is used to transparently access files on remote machines where
 your username might be different.
One special entry, \"default\", optionally gives the username to use
 if none of the others match.  If default is not given, (user-login-name)
 is used instead.
Also see the function rhosts-to-alist, which can convert your .rhost
 file into the proper format.")

(defvar rsh (cond ((eq system-type 'hpux) "remsh")
		  (t "rsh"))
  "Name of the command to invoke a remote shell.")

(defvar rcp "rcp"
  "Name of the command to do a remote copy")

(defvar track-default-remote-host t
  "Controls whether  default-remote-host  changes.
When nil,  default-remote-host  keep its value, otherwise it will always
have the value of the last remote host successfully read.")

(defvar remote-directory-map
  (let ((map (copy-keymap minibuffer-local-map)))
    (define-key map " "  'remote-directory-complete)
    (define-key map "\t" 'remote-directory-complete)
    (define-key map "?"  'remote-directory-complete)
    (define-key map "\e" 'remote-directory-set-host)
    map)
  "Map used by read-remote-file-name-with-completion.")

(make-variable-buffer-local 'buffer-remote-file-name)
(set-default 'buffer-remote-file-name "")
(make-variable-buffer-local 'remote-editing)

(if (not (assoc 'remote-editing minor-mode-alist))
    (setq minor-mode-alist
	  (cons '(remote-editing " Remote") minor-mode-alist)))

;;; Here starts the code...

(defun remote-editing (arg)
  "Toggle remote-editing mode.  Remote-editing mode is described below.

With arg, turn on remote editing mode iff arg is positive, otherwise toggle.
In remote editing mode, the normal bindings for find-file,
 find-file-read-only, find-alternate-file, save-buffer, write-file,
 and insert-file are changed to operate on a remote system by default.
When remote editing, a prefix arg allows local file operations.  When not
 remote editing, a prefix arg allows remote file operations.
It is assumed that .rhosts files are set up properly on both machines.
If you have different usernames on other machines, see the
 documentation for remote-username-alist."
  (interactive "P")
  (setq remote-editing
	(if (null arg) (not remote-editing)
	  (> (prefix-numeric-value arg) 0)))
  (set-buffer-modified-p (buffer-modified-p))) ;No-op, but updates mode line.

(require 'backquote)
;;;
;;; Macro used as front-end to normal file operation key bindings to decide
;;; between local and remote editing modes.
;;;
;;; Automatically constructs doc string and includes prefix arg hacking
;;; to temporarily toggle sense of remote-editing.
;;;
;;; defaultlocal idea by wjc@ho5cad.att.com (Bill Carpenter)

(defmacro def-local-or-remote (binding name remote defaultlocal)
  (if (eq (key-binding binding) name)	; already done - avoid infinite loop
      nil
    (let* ((local (or (key-binding binding) defaultlocal))
	   (r (symbol-name remote))
	   (l (symbol-name local)))
      (` (progn
	   (munge-remote-bindings '(, local) '(, name))
	   (defun (, name) (arg)
	     (, (concat "Call either " r " or " l ".
If remote-editing (which see), call " r ",
 else call " l ".
See also the documentation for " r "
 and " l "."))
	     (interactive "P")
	     (call-interactively
	      (if (xor remote-editing arg)
		  '(, remote)
		'(, local)))))))))

(defun munge-remote-bindings (old new)
  (mapcar (function (lambda (map)
		      (substitute-key-definition old new map)))
	  (list global-map esc-map ctl-x-map)))

(def-local-or-remote "\C-x\C-f"	find-local-or-remote-file
  find-remote-file		find-file)

(def-local-or-remote "\C-x\C-r"	find-local-or-remote-file-read-only
  find-remote-file-read-only	find-file-read-only)

(def-local-or-remote "\C-x\C-v"	find-alternate-local-or-remote-file
  find-alternate-remote-file	find-alternate-file)

(def-local-or-remote "\C-x\C-s"	save-local-or-remote-buffer
  save-remote-buffer		save-buffer)

(def-local-or-remote "\C-x\C-w"	write-local-or-remote-file
  write-remote-file		write-file)

(def-local-or-remote "\C-xi"	insert-local-or-remote-file
  insert-remote-file		insert-file)

(defun find-remote-file (host file)
  "Edit remote file HOST:FILE (using rcp).
This command is similiar to find-file, but uses rcp to read the file from
a remote machine.  Also see remote-editing."
  (interactive (read-remote-file-name "Find remote file"))
  (let ((f-or-b (get-remote-file-or-buffer host file "retrieve"))
	local-file)
    (if f-or-b
	(if (bufferp f-or-b)
	    (switch-to-buffer f-or-b)
	  (setq local-file f-or-b)
	  (let ((buf (generate-new-buffer
		      (concat host (file-name-nondirectory file)))))
	    (switch-to-buffer buf)
	    (setq buffer-file-name (buffer-name))
	    (insert-file-contents local-file)
	    (set-buffer-modified-p nil)
	    (delete-file local-file)
	    (normal-mode)
	    (remote-editing 1)
	    (setq buffer-remote-file-name (concat host file)
		  buffer-offer-save t))))))

(defun find-remote-file-read-only ()
  "Edit remote file FILENAME, but mark buffer as read-only.
Also see find-remote-file and remote-editing."
  (interactive)
  (call-interactively 'find-remote-file)
  (setq buffer-read-only t))

(defun find-alternate-remote-file ()
  "Find alternate file using rcp.
This command is similiar to find-alternate-file, but uses rcp to fetch
the file from a remote machine.  Also see remote-editing."
  (interactive)
  (and (buffer-modified-p)
       (not buffer-read-only)
       (not (yes-or-no-p (format "Buffer %s is modified; kill anyway? "
				 (buffer-name))))
       (error "Aborted"))
  (let ((obuf (current-buffer))
	(oname (buffer-name)))
    (rename-buffer " **lose**")
    (unwind-protect
	(apply 'find-remote-file
	       (read-remote-file-name "Find remote alternate file"))
      (if (eq obuf (current-buffer))
	  (rename-buffer oname)
	(kill-buffer obuf)))))

(defun save-remote-buffer ()
  "Save a file using rcp.
This command is similiar to save-buffer, but uses rcp to write the file back
to a remote machine.  Also see remote-editing."
  (interactive)
  (if (buffer-modified-p)
      (if (zerop (length buffer-remote-file-name))
	  (call-interactively 'write-remote-file)
	(do-write-remote-file buffer-remote-file-name))
    (message "(No changes need to be saved)")))

(defun write-remote-file (host file)
  "Write a file HOST:FILE using rcp.
This command is similiar to write-file, but uses rcp to write the file back
to a remote machine.  Also see remote-editing."
  (interactive (read-remote-file-name "Write remote file"))
  (do-write-remote-file (concat host file)))

(defun insert-remote-file (host file)
  "Insert a remote file HOST:FILE using rcp.
This command is similiar to insert-file, but uses rcp to read the file from
a remote machine.  Also see remote-editing."
  (interactive (read-remote-file-name "Insert remote file"))
  (let ((f-or-b (get-remote-file-or-buffer host file "insert")))
    (if f-or-b
	(if (bufferp f-or-b)
	    (insert-buffer f-or-b)
	  (insert-file f-or-b)
	  (delete-file f-or-b)))))

; This function is modified from one from Roland McGrath
; (mcgrath%paris.Berkeley.EDU@ginger.Berkeley.EDU) sent me.
; It makes it easy to maintain remote-username-alist.
(defun rhosts-to-alist (file)
  "Turns FILE, which is in the format of `.rhosts', into a Lisp alist
\(\"MACHINE\" . \"USER\"\), which is inserted at point."
  (interactive "fConvert file: ")
  (let ((standard-output (current-buffer))
	(temp (prog1 (set-buffer (get-buffer-create " *rhosts-to-alist*"))
		(erase-buffer))))
    (insert-file-contents (expand-file-name file))
    (princ "'(")
    (while (not (eobp))
      (if (re-search-forward "^\\([^ \t\n]+\\)[ \t]*\\([^ \t\n]+\\)?$" nil t)
	  (progn
	    (prin1 (cons (buffer-substring (match-beginning 1) (match-end 1))
			 (if (match-beginning 2)
			     (buffer-substring
			      (match-beginning 2) (match-end 2))
			     (user-login-name))))
	    (princ "\n  ")))
      (forward-line 1))
    (princ ")\n")
    (kill-buffer temp)))

;;;
;;; Internal routines
;;;

(defun get-remote-username (hostname)
  (or (cdr (assoc (strip-trailing-colons hostname) remote-username-alist))
      (cdr (assoc "default" remote-username-alist))
      (user-login-name)))

(defun do-write-remote-file (remote)
  (let* ((temp (concat "/tmp/" (buffer-name)))
	 (output (save-excursion
		   (prog1 (set-buffer (get-buffer-create "*Rcp Output*"))
		     (erase-buffer))))
	 (cursor-in-echo-area t)
	 time
	 full-remote-name)
    ;; write-file doesn't quite do it.
    (save-restriction
      (widen)
      (write-region (point-min) (point-max) temp nil 'no-message))
    (message "Sending %s..." remote)
    (setq full-remote-name (munge-remote-filename remote))
    (if (setq time (remote-process-wait
		    (start-process "rcp"
				   output
				   rcp temp full-remote-name)))
	(progn
	  (if remote-editing
	      (let ((new-name
		     (concat (remote-filename-host-part remote)
			     (file-name-nondirectory
			      (remote-filename-file-part remote)))))
		(or (get-buffer new-name) (rename-buffer new-name))
		(set-buffer-modified-p nil)))
	  (setq buffer-remote-file-name remote)
	  (message "%d bytes in %d seconds" (buffer-size) time)
	  (delete-file temp))
      (remote-rcp-error output full-remote-name "update"))))

(defun get-remote-file-or-buffer (host file message)
  "Return a remote file as either a buffer or a file.
If the file HOST:FILE already has been read in, return the buffer
that contains it; otherwise try and rcp the file to the local machine.
If successful, return the local file name."
  (if (string-match "^~" file)
      (setq file (remote-expand-~ host file)))
  (let ((remote (concat host file))
	(temp (concat "/tmp/" (file-name-nondirectory file))))
    (if (file-directory-p file)
	(progn
	  (message "Remote dired not yet implemented")
	  nil)
      (or (get-remote-buffer remote)	  ;; already exists
	  (let* ((output (save-excursion
			   (prog1 (set-buffer (get-buffer-create
					       "*Rcp Output*"))
			     (erase-buffer))))
		 (cursor-in-echo-area t)
		 time)
	    (message "Retrieving %s..." remote)
 	    (setq remote (munge-remote-filename remote))
	    (if (setq time (remote-process-wait
			    (start-process "rcp"
					   output
					   rcp
					   remote
					   temp)))
		(progn
		  (message "%d bytes in %d seconds"
			   (nth 7 (file-attributes temp)) time)
		  temp)
	      (remote-rcp-error output remote message)))))))

(defun munge-remote-filename (remote-name)
  (let* ((file (remote-filename-file-part remote-name))
	 (host (remote-filename-host-part remote-name))
	 (name (get-remote-username host)))
    (cond ((equal remote-rcp-syntax "@")
	   (concat name "@" host file))
	  ((equal remote-rcp-syntax ".")
	   (concat (strip-trailing-colons host) "." name ":" file))
	  (t remote-name))))

(defun get-remote-buffer (name)
  (save-window-excursion
    (let ((buffers (buffer-list)) found)
      (while (and (not found) buffers)
	(set-buffer (car buffers))
	(if (string= name buffer-remote-file-name)
	    (setq found (car buffers)))
	(setq buffers (cdr buffers)))
      found)))

(defun read-remote-file-name (prompt)
  "Read a remote file specification, and return list (host file).
Prompting with PROMPT, get a file name.  The initial host defaults
to a value is derived from buffer-remote-file-name, or if there is
none, then from the global default (default-remote-host).

The host can be changed by hitting an ESC."
  (let* ((host (or (remote-filename-host-part buffer-remote-file-name)
		   default-remote-host
		   (read-string (format "%s on what host? " prompt))))
	 (file (file-name-directory
		(or (remote-filename-file-part buffer-remote-file-name) ""))))
    (setq host (strip-trailing-colons host))
    (let* ((res (read-remote-file-name-with-completion prompt host file))
	   (host (nth 0 res))
	   (file (nth 1 res)))
      (and track-default-remote-host
	   (setq default-remote-host host))
      (list host
	    (if (or (null file) (string= file (file-name-directory file)))
		(concat file
			(or (if (not (string= buffer-remote-file-name ""))
				(file-name-nondirectory
				 (remote-filename-file-part
				  buffer-remote-file-name)))
			    (remote-filename-file-part (buffer-name))
			    (buffer-name)))
	      file)))))

(defun read-remote-file-name-with-completion (prompt host file)
  (let ((cursor-in-echo-area t))
    (message "Hit esc to change the default host")
    (sit-for 1))
  (while (string= ""
		  (setq file
			(read-from-minibuffer
			 (concat prompt " (on " host "): ")
			 file remote-directory-map))))
  (list (concat host ":")
	(if (last-char-is file ?*)
	    (substring file 0 -1)
	  file)))

(defvar remote-dir-info nil
  "Holds the completion table information for read-remote-file-name.
Don't change this.")

(defun remote-directory-set-host ()
  "Internal routine for read-remote-file-name.  Don't call."
  (interactive)
  (let* ((enable-recursive-minibuffers t)
	 (new (read-string (format "New host (default %s): " host))))
    (if (string= new "")
	()
      (setq host (strip-trailing-colons new)
					; New host =>
	    remote-dir-info ())		;  must invalidate remote-dir-info,
      (erase-buffer)			;  current file name,
      (exit-minibuffer))))		;  and try again

(defun remote-directory-complete ()
  "Internal routine for read-remote-file-name.  Don't call.
Note that this routine uses the dynamic binding of host."
  (interactive)
  (let ((file (buffer-substring (point-min) (point-max))))
    (if (string-match "^~" file)
	(progn
	  (erase-buffer)
	  (setq file (remote-expand-~ host file))
	  (insert file))) ; into the minibuffer
    (if (or (null remote-dir-info)		       ; no table
	    (not (equal (nth 0 remote-dir-info) host)) ; out-of-date host
	    (not (eq 0 (string-match		       ; out-of-date file
			(regexp-quote (nth 1 remote-dir-info)) file)))
	    (string-match			       ; a new directory
	     "/" (substring file (length (nth 1 remote-dir-info))))
	    )
	(setq remote-dir-info
	      (list host file (remote-directory-list host file))))
    (let* ((minibuffer-completion-table
	    (mapcar 'list 
		    (all-completions file (nth 2 remote-dir-info))))
	   (length (length minibuffer-completion-table)))
      (cond ((= length 0)
	     (let ((minibuffer-completion-table '(())))
	       (minibuffer-complete)))
	    ((= length 1)
	     (minibuffer-complete))
	    (t (minibuffer-completion-help)
	       (minibuffer-complete))))))

(defun remote-expand-~ (host file)
  (let* ((host (strip-trailing-colons host))
	 (proc (start-process "expand"
			      ()
			      rsh host
			      "-l" (get-remote-username host)
			      "echo" file))
	 (output)
	 (cursor-in-echo-area t))
    (message (format "Querying %s to expand %s" host file))
    (set-process-filter proc
			'(lambda (proc string)
			   (setq output string)))
    (accept-process-output proc)
    (substring output 0 -1)))		; remove newline

(defun remote-directory-list (host file)
  (let* ((cursor-in-echo-area t)
	 (process)
	 (host (substring host 0 (string-match ":" host)))
	 (remote-directory-output ""))	; dynamic binding for filter
    (message "Remote directory listing of %s:%s (be patient)" host file)
    ;;
    ;; System dependent - we want a listing that gives all completions, but
    ;; leaves a slash ("/") at the end of directories.
    ;;
    (setq file (concat file "*"))
    (setq process (start-process
		   "Remote Directory Listing"
		   nil
		   rsh host
		   "-l" (get-remote-username host)
		   "/bin/ls" "-dF" file))
    (set-process-filter process 'remote-directory-filter)
    (if (remote-process-wait process)
	(or (parse-remote-output remote-directory-output) (list nil))
      (error "Couldn't get a remote listing of %s:%s" host file))))

(defun remote-directory-filter (process output)
  ;; remote-directory-output dynamically bound in remote-directory-list
  (setq remote-directory-output (concat remote-directory-output output)))

(defun parse-remote-output (output)
  (let ((case-fold-search t)
	(result ()))
    (if (or (string-match "no match" output)
	    (string-match "not found" output)
	    (string-match "no such file or directory" output))
	()
      (while (string-match "\012" output)
	;; cons the output into a list -
	(setq result (cons (list (substring output 0 (match-beginning 0)))
			   result)
	      output (substring output (match-end 0)))))
    result))

(defun last-char-is (string char)
  "True if last character of STRING is CHAR."
  (and (> (length string) 0)
       (= (aref string (1- (length string)))
	  char)))

(defun strip-trailing-colons (host)
  (while (last-char-is host ?:)
    (setq host (substring host 0 -1)))
  host)

(defun remote-filename-host-part (name)
  (if (string-match ".+:" name)
      (substring name 0 (match-end 0))))

(defun remote-filename-file-part (name)
  (if (string-match ".+:\\(.+\\)" name)
      (substring name (match-beginning 1) (match-end 1))))

(defun xor (a b)
  (eq (null a) (not (null b))))

(defun remote-process-wait (proc)
  (condition-case ignore
      (let ((time 0))
	(while (eq (process-status proc) 'run)
	  (setq time (1+ time))
	  (sleep-for 1))
	(if (and (eq (process-status proc) 'exit)
		 (eq (process-exit-status proc) 0))
	    time
	  nil))
    (quit (progn (kill-process proc)
		 nil))))

(defun remote-rcp-error (buffer file-name message)
  (save-window-excursion
    (switch-to-buffer buffer)
    (delete-other-windows)
    (goto-char 1)
    (insert (format "Unable to %s %s\n\n" message file-name))
    (goto-char (point-max))
    (message "Hit any character to continue")
    (read-char)
    (bury-buffer buffer)))

(defun remote-this-buffer-is-saved ()
  (if (string= (buffer-name) "remote.el")
      (progn
	(save-excursion
	  (goto-char (point-min))
	  (search-forward "version ")
	  (read (current-buffer))
	  (let ((minor (save-excursion (read (current-buffer)))))
	    (kill-word 1)
	    (prin1 (1+ minor) (current-buffer))))
	(message "Remember to byte-compile")
	(sit-for 1)))
  nil)
