;;; rp-describe-function.el
;;;
;;; This package replaces GNU Emacs' "describe-function" with a version
;;; customized for Emacs LISP programmers.  I like this version better
;;; because:
;;; 
;;; 	1) it provides more information, including:
;;; 		the function type: LISP, macro or builtin C
;;; 		whether the function still needs to be autoloaded
;;; 		whether the function is interactive
;;; 		whether the function is bytecompiled
;;; 	  --->	the function's argument list
;;; 		the function's local and global key bindings
;;;                  
;;; 	2) it can quickly flash the info using momentary-string-display
;;; 
;;; 	3) it's smarter about figuring out what function you mean from
;;; 	   the buffer context.
;;;
;;; Copyright (C) 1991 Free Software Foundation, Inc.
;;;
;;; This file may someday become part of GNU Emacs.
;;;
;;; This software is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility to
;;; anyone for the consequences of using it or for whether it serves any
;;; particular purpose or works at all, unless he says so in writing.  Refer to
;;; the GNU Emacs General Public License for full details.
;;;
;;; Everyone is granted permission to copy, modify and redistribute this file,
;;; but only under the conditions described in the GNU Emacs General Public
;;; License.  A copy of this license is supposed to have been given to you
;;; along with GNU Emacs so you can know your rights and responsibilities.  It
;;; should be in a file named COPYING.  Among other things, the copyright
;;; notice and this notice must be preserved on all copies.
;;;
;;; Hacked from various sources by Robert Potter (rpotter@grip.cis.upenn.edu).
;;; Last edit Thu May  2 23:58:16 1991

(require 'cl)
(provide 'rp-describe-function)


;; some rather general utility functions


(defun looking-back-at (PAT)
  "t when text before point matches regular expression PAT."
  (save-excursion
    (save-restriction
      (narrow-to-region (point-min) (point))
      (re-search-backward (concat "\\(" PAT "\\)\\'") (point-min) t)
      )
    )
  )


(defun symbol-around-point ()
  "Returns a string consisting of the symbol that point is within (or near)."
  (save-excursion
    (while (looking-at "\\sw\\|\\s_")
      (forward-char 1))
    (if (looking-back-at "\\sw\\|\\s_")
      (buffer-substring (point)
			(progn (while (looking-back-at "\\sw\\|\\s_")
				 (forward-char -1))
			       (point)))
      "")))


; derived from momentary-string-display

(defun flash-string (STRING)
  "Momentarily display STRING in the buffer; erase it on the next keystroke.
The string is displayed starting on the line after point.  The window is
recentered if necessary to make the whole string visible.  If the window isn't
large enough, at least you get to read the beginning."
  (let ((buffer-read-only nil)
	(modified (buffer-modified-p))
	(name buffer-file-name)
	insert-start
	insert-end)
    (unwind-protect
	(progn
	  (save-excursion
	    ;; defeat file locking... don't try this at home, kids!
	    (setq buffer-file-name nil)
	    (forward-line 1)
	    (setq insert-start (point))
	    (insert STRING)
	    (setq insert-end (point)))
	  ; make sure the whole string is visible
	  (when (not (pos-visible-in-window-p insert-end))
	    (recenter (max 0
			   (- (window-height)
			      (count-lines insert-start insert-end)
			      2))))
	  (message "Type %s to continue editing..."
		   (single-key-description ?\ ))
	  (let ((char (read-char)))
	    (or (eq char ?\ )
		(setq unread-command-char char))))
      (if insert-end
	  (save-excursion
	    (delete-region insert-start insert-end)))
      (setq buffer-file-name name)
      (set-buffer-modified-p modified))))


;; the more specific functions


(defun function-description (FUNCTION-SYMBOL &optional DO-BINDINGS)
  "Returns a string with the full, formatted description of FUNCTION-SYMBOL.
If optional second argument DO-BINDINGS is non-nil, current global and local
key bindings will be supplied as well (this is slower)."
  (let ((func (symbol-function FUNCTION-SYMBOL)))
    (format "%s -- %s\n%s\n%s\n"
	    FUNCTION-SYMBOL
	    (cond ((subrp func)
		   "a built-in C subroutine")
		  ((and (symbolp func) (fboundp func))
		   (format "same as \"%s\"" func))
		  ((stringp func)
		   "a keyboard macro")
		  ((not (consp func))
		   func)		; does this happen?
		  ((eq (car func) 'autoload)
		   "a yet-to-be-autoloaded lisp function")
		  ((eq (car func) 'lambda)
		   (format "a%s%s lisp function taking %s"
			   (if (eq (car (cadddr func)) 'interactive)
			       "n interactive"
			     "")
			   (if (or (eq (car (cadddr func)) 'byte-code)
				   (eq (caar (cddddr func)) 'byte-code))
			       " bytecompiled"
			     "")
			   (or (cadr func) "()")))
		  ((eq (car func) 'macro)
		   (format "a macro taking %s"
			   (or (caddr func) "()")))
		  (t
		   func))		; I don't think this happens, but WTF
	    (if (and DO-BINDINGS
		     (commandp func))
		(let ((global-bindings (where-is-internal FUNCTION-SYMBOL))
		      (local-bindings (let ((gm (current-global-map)))
					(prog2
					 (use-global-map (make-keymap))
					 (where-is-internal
					  FUNCTION-SYMBOL
					  (current-local-map))
					 (use-global-map gm)))))
		  (if (or global-bindings local-bindings)
		      (format "\nBound %s%s%s.\n"
			      (if global-bindings
				  (format "globally to %s"
					  (mapcar 'key-description
						  global-bindings))
				"")
			      (if (and global-bindings local-bindings)
				  " and "
				"")
			      (if local-bindings
				  (format "locally to %s"
					  (mapcar 'key-description
						  local-bindings))
				""))
		    ""))
	      "")
	    (or (documentation func)
		"(not documented as a function)"))))


(defun funcsym-near-point (INTERACT &optional PROMPT)
  "Extract a function symbol from the text around point.
Depending on argument INTERACT, the user may be prompted to confirm or override
the selection:
   INTERACT = t      always prompt
   INTERACT = nil    never prompt
   otherwise         prompt only when no valid function symbol is found
The optional PROMPT argument may be a string with which to prompt the user."
  (let* ((fcap (function-called-at-point))
	 (sap (intern-soft (symbol-around-point)))
	 (fn (or (if (fboundp fcap)
		     fcap)
		 (if (fboundp sap)
		     sap))))
    (if (or (eq INTERACT t)
	    (and (not (eq INTERACT nil))
		 (null fn)))
	(let* ((enable-recursive-minibuffers t)	     
	       (val (completing-read (format "%s%s: "
					     (or PROMPT
						 "Function")
					     (if fn (format " (default %s)"
							    fn)
					       ""))
				     obarray 'fboundp t)))
	  (if (equal val "")
	      fn
	    (intern val)))
      fn)))


(defun slow-describe-function (FUNCTION-SYMBOL)
  "Display some information about a FUNCTION-SYMBOL.
This is slower than quick-describe-function because it prompts for the symbol
and because it uses the help buffer.

When called interactively, FUNCTION-SYMBOL defaults to the function point is
currently near."
  (interactive (list (funcsym-near-point t "Describe function")))
  (with-output-to-temp-buffer "*Help*"
    (princ (function-description FUNCTION-SYMBOL t))
    (print-help-return-message))
  nil
  )


;;; inspired by Cesar Quiroz

(defun quick-describe-function (FUNCTION-SYMBOL)
  "Momentarily display some information about a FUNCTION-SYMBOL.
This is faster than slow-describe-function because it rarely prompts for the
symbol and because it displays the info right in the current buffer.

When called interactively, FUNCTION-SYMBOL defaults to the function point is
currently near."
  (interactive (list (funcsym-near-point 'maybe "Describe function")))
  (flash-string (concat (make-string (- (window-width) 1) ?#)
			"\n"
			(function-description FUNCTION-SYMBOL t)
			(make-string (- (window-width) 1) ?#)
			"\n")))


; before redefining describe-function, save the old one for emergencies (unless
; we have already done so)
(if (not (fboundp 'old-describe-function))
    (fset 'old-describe-function (symbol-function 'describe-function)))


(defun describe-function (FUNCTION-SYMBOL)
  "Display a FUNCTION-SYMBOL's type, arguments, bindings and documentation.

If called non-interactively or with a interactive argument the description
is displayed in the help buffer.  Otherwise, the description is momentarily
displayed in the current window, disappearing on the next keystroke.

When called interactively, FUNCTION-SYMBOL defaults to the function point is
currently near."
  (interactive "P")
  (cond ((not (interactive-p))
	 (slow-describe-function FUNCTION-SYMBOL))
	(FUNCTION-SYMBOL		; any interactive argument
	 (call-interactively 'slow-describe-function))
	(t
	 (call-interactively 'quick-describe-function))))
