;;; slurp.el - slurp a directory tree of files into a buffer.

;;; Author: Dominic Prior <dominic@smallworld.co.uk>

;;; Commentary:

;; Loads many files into one buffer so that fast, easy-to-use
;; operations like isearch can work on many files at once.

;; There is no attempt to check soft or hard links or to detect
;; infinite circularities.  The running file-count in the message area
;; might help by alerting the user to hit C-g.

;; This runs on GNU Emacs 19.25.

;; An example file pattern could be
;;   \.[ch]$
;; for C programming, or
;;   [^~]$
;; for all files except backup files.


;;; Code:


(require 'cl)

(defun slurp (dir pattern &optional count-data)
  "Insert the files in DIR and the subdirs of DIR that match PATTERN.
Returns a list of the file-count and the total-size.
The optional COUNT-DATA arg is used during the recursion."
  (interactive "*Ddirectory: \nsFile-name pattern: ")
  (if (null count-data)
      (setq count-data '(0 0)))
  (loop for f in (directory-files dir)
        do
        (let
            ((file (concat (file-name-as-directory dir) f)))
          (if (file-directory-p file)
              (if (and (not (equal f "."))
                       (not (equal f "..")))
                  (setq count-data (slurp file pattern count-data)))
            (if (string-match pattern f)
                (progn
                  (insert ":::: " file " ::::\n\n")
                  (let
                      ((size (second (insert-file-contents file))))
                    (forward-char size)
                    (insert "\n\n\n")
                    (incf (second count-data) size))
                  (incf (first count-data))
                  (message "%d files, %d bytes" (first count-data) (second count-data)))))))
  count-data)

(provide 'slurp)

;;; slurp.el ends here
