;; symlink-fix: Remove symbolic links from file pathnames.
;; Copyright (C) 1989, 1990, 1991 Free Software Foundation, Inc.

;; This file is not officially part of GNU Emacs.

;; This file is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
;; to anyone for the consequences of using it or for whether it serves any
;; particular purpose or works at all, unless he says so in writing.
;; Refer to the GNU Emacs General Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute this
;; file, but only under the conditions described in the GNU Emacs General
;; Public License.  A copy of this license is supposed to have been given
;; to you along with GNU Emacs so you can know your rights and
;; responsibilities.  It should be in a file named COPYING.  Among other
;; things, the copyright notice and this notice must be preserved on all
;; copies.

;; Created by: Joe Wells, joew@uswest.com
;; Created on: summer 1988
;; Last modified by: Joe Wells, jbw@bucsd.bu.edu
;; Last modified on: Wed Sep 11 21:56:04 1991
;; Filename: symlink-fix.el
;; Purpose: remove symbolic links from pathnames
;; Change log: 
;; 
;; Fri Aug 23 12:54:22 1991  Joe Wells  (jbw at teton)
;; 
;; 	* Fixed some bugs with the symlink-mapping-alist.
;; 
;; Thu Aug 22 14:58:33 1991  Joe Wells  (jbw at teton)
;; 
;; 	* Added documentation.
;; 

;; LCD Archive Entry:
;; symlink-fix|Joe Wells|jbw@cs.bu.edu|
;; Remove symbolic links from file pathnames
;; |August 23, 1991||packages/symlink-fix.el.Z|

;; Features:
;;
;; 1. Makes Emacs use the real path, which prevents accidentally visiting
;; the same file twice by different paths.  This problem could happen when
;; Emacs used the output of a program that had called getwd, and also in
;; other similar circumstances.
;;
;; Known problems:
;;
;; 1. expand-file-name (if overloaded) now clobbers the match data.  I'll
;; fix this sometime (as well as possible considering the problem with 0
;; indices into strings).
;;
;; 2. symlink-expand-file-name does a lot of string creation.  I intend to
;; fix this.
;;
;; 3. expand-file-name (if overloaded) doesn't work really well with the
;; automounter.  In essence, you have to selectively disable part of its
;; functionality to prevent the automounter from unmounting file systems
;; from under you.
;;
;; 4. expand-file-name (if overloaded) doesn't work if this file was
;; loaded before ange-ftp unless you modify ange-ftp-expand-file-name to
;; not be clever and always call ange-ftp-real-expand-file-name.  Loading
;; after ange-ftp results in extra work being done.

;; TODO: dynamically change whether symlinks are expanded via variable.

(defvar symlink-overload-expand-file-name nil
  "*Whether expand-file-name should be replaced by symlink-expand-file-name
at the time symlink-fix.el is loaded.  To perform this operation later, do
this:

  (fset 'symlink-new-expand-file-name 'symlink-expand-file-name)

To return to the original expand-file-name do this:

  (fset 'symlink-new-expand-file-name 'symlink-original-expand-file-name)")

(defvar symlink-mapping-alist nil
  "*Used to hide certain directories during symlink elimination.
Should be an alist where each element is of the form (REGEXP . REPLACEMENT).
When a symbolic link points to a path matched by REGEXP, the portion of
the path that matches either the entire REGEXP is replaced by REPLACEMENT
\(as described for replace-match).  If REPLACEMENT is nil, then the
original symbolic link is used rather than the path it points to.  This is
only applied when the symbolic link points to an absolute path.

As an example, if /tmp_mnt is an automounter directory, you might want to
make symlink-mapping-alist have this value:

  ((\"\\\\`/tmp_mnt/\" . nil))

Another example is if /u is a directory containing symbolic links to each
person's real home directory, which are located in /home/machine/username,
and it is desired that this be hidden.  Then symlink-mapping-alist can be
given this value:

  ((\"\\\\`/home/[-a-z0-9]+/[a-z]+\\\\'\" . nil))

Note that doing this defeats the whole point of using
symlink-expand-file-name, which is to have Emacs use the same pathname
that is reported by getwd (which is used by many other programs whose
output Emacs uses to find files).

Also note that any transformation specified should yield the original
symbolic link or an equivalent one.  Otherwise all bets are off.")

(defun assoc-string-match (string alist)
  "Like CL (assoc string ALIST :test #'(lambda (x y) (string-match y x))).
If a match is made, then the match data is from the successful match,
otherwise it is clobbered."
  ;; TODO: use a real save-match-data
  (let (item)
    (catch 'found
      (while (consp alist)
	(setq item (car alist))
	(and (consp item)
	     (string-match (car item) string)
	     (throw 'found item))
	(setq alist (cdr alist))))))

;; function not in GNU Emacs
;; TODO: give it arguments like replace-regexp, not replace-match
(defun replace-regexp-string (string regexp to-string
				     &optional fixedcase literal)
  "Replace matches in STRING for REGEXP by TO-STRING.
If fourth arg FIXEDCASE is non-nil, do not alter case of replacement text.
Otherwise convert to all caps or cap initials, like replaced text.
If fifth arg LITERAL is non-nil, use TO-STRING literally.
Otherwise treat \ as special:
  \& in TO-STRING means substitute original matched text,
  \N means substitute match for \(...\) number N,
  \\ means insert one \."
  (save-excursion
    (set-buffer (get-buffer-create " *replace-regexp-string*"))
    (erase-buffer)
    (buffer-flush-undo (current-buffer))
    (insert string)
    (goto-char (point-min))
    (while (re-search-forward regexp nil t)
      (replace-match to-string fixedcase literal))
    (buffer-string)))

(defun symlink-expand-file-name (file-name &optional directory)
  "Convert FILENAME to absolute, and canonicalize it.
Second arg DEFAULT is directory to start with if FILENAME is relative
\(does not start with slash); if DEFAULT is nil or missing,
the current buffer's value of default-directory is used.
Filenames containing . or .. as components are simplified;
initial ~ is expanded.  See also the function  substitute-in-file-name.

This is not the default version of expand-file-name.
All symbolic links are resolved from FILENAME, with exceptions specified
by symlink-mapping-alist.  The original version of expand-file-name is
available on symlink-original-expand-file-name."
  (let (left right split link new)
    (setq right (symlink-original-expand-file-name file-name directory))
    (setq left "")
    (while (not (equal right ""))
      (setq split (symlink-split-file-name right))
      (setq left (symlink-join-file-name left (car split)))
      (setq right (cdr split))
      (setq link (file-symlink-p left))
      (if (null link)
	  nil
	(if (eq 0 (length link)) (setq link "."))
	(if (not (eq (aref link 0) ?/))
	    (setq split (symlink-split-file-name link)
		  left (symlink-join-file-name (file-name-directory left)
					       (car split))
		  right (symlink-join-file-name (cdr split) right))
	  
	  ;; Check for symlink hiding:
	  (setq new (assoc-string-match link symlink-mapping-alist))
	  (if new
	      (if (cdr new)
		  (setq left (replace-regexp-string link (car new) (cdr new))))
	    
	    ;; We have an absolute path now so we're starting over.
	    (setq right (symlink-join-file-name link right))
	    (setq left "")))))
    (setq left (symlink-original-expand-file-name left))
    (if (string-equal left file-name)
	file-name
      left)))

;; Concatenates LEFT and RIGHT, preserving at most one slash between them.
;; This horrible hack is necessary to work around the fact that
;; expand-file-name treats // specially.
(defun symlink-join-file-name (left right)
  (let* ((llen (length left))
	 (rlen (length right))
	 (lend llen)
	 (rstart 0)
	 slash-found)
    (while (and (> lend 0)
		(eq ?/ (aref left (1- lend))))
      (setq slash-found t
	    lend (1- lend)))
    (while (and (< rstart rlen)
		(eq ?/ (aref right rstart)))
      (setq slash-found t
	    rstart (1+ rstart)))
    (concat (if (eq lend llen) left (substring left 0 lend))
	    (if slash-found "/" "")
	    (if (eq rstart 0) right (substring right rstart rlen)))))

;; Splits FILENAME into two strings, and returns a list of the two
;; strings.  The first string will be the first filename component in
;; FILENAME, plus any leading slashes, and the second string will be the
;; rest of FILENAME, possibly a string of length 0.
(defun symlink-split-file-name (file-name)
  (if (string-match "\\`\\(/*[^/]+\\)\\(/.*\\)\\'" file-name)
      (cons (substring file-name (match-beginning 1) (match-end 1))
	    (substring file-name (match-beginning 2) (match-end 2)))
    (cons file-name "")))

;; TODO: change symlink-new-expand-file-name into a real function that
;; checks dynamically what to call.
(or (fboundp 'symlink-original-expand-file-name)
    (progn
      (fset 'symlink-original-expand-file-name
	    (symbol-function 'expand-file-name))
      (fset 'symlink-new-expand-file-name 'symlink-original-expand-file-name)
      (fset 'expand-file-name 'symlink-new-expand-file-name)))

(if symlink-overload-expand-file-name
    (fset 'symlink-new-expand-file-name 'symlink-expand-file-name))

(provide 'symlink-fix)
