;;; This file provides a number of functions to handle your buffers.
;;; 
;;; The uninteresting-buffers variable contains a regexp defining the
;;; names of buffers which are "uninteresting".
;;; 
;;; If there are no buffers which are not "uninteresting", then the
;;; functions will find a buffer which is less uninteresting.  The
;;; totally-uninteresting variable contains a regexp defining the names of
;;; buffers which are totally-uninteresting.
;;; 
;;; The interesting-buffer-p function indicates whether the given (or
;;; current) buffer is interesting or not.
;;; 
;;; The find-interesting-buffer function will find an interesting buffer
;;; and return it.  If there are no interesting buffers, it will try to
;;; find a less uninteresting buffer and return it.  Otherwise it returns
;;; a *scratch* buffer.
;;; 
;;; The close-current-window command will close the current window,
;;; or will bury the buffer and find a new interesting one if there is
;;; only one window.
;;;
;;; The kill-buffer-or-window command will kill the current buffer and
;;; get a new one, or close the current window,  if there are no more
;;; interesting buffers.
;;; 
;;; I bind close-current-window to the PF3 key on my keypad and the
;;; kill-buffer-or-window to the PF4 key on my keypad.  PF1 and PF2 are
;;; bound to functions which go forward and backward through the window
;;; list.
;;; 
;;; One other thing... kill-buffer-or-window doesn't kill a buffer in
;;; cmushell-mode, since I don't want to accidentally kill my shell
;;; buffer.   It _will_ if called with an argument (or prefix argument).
;;; 
;;; You'll probably want to add to the uninteresting-buffers regexp.
;;; Candidates would be buffers like *Help* and *info*.
;;; 

(defvar uninteresting-buffers "^ .*\\|^\\*scratch\\*$"
"*This variable contains a regexp representing the names of buffers
which are uninteresting.  This variable is used by the function
find-interesting-buffer, which selects an interesting buffer and
returns it.")

(defvar totally-uninteresting-buffers "^ .*\\|^\\*scratch\\*$"
  "This regexp indicates _really_ uninteresting buffers.")

(defun fib-check-name (buflist regexp)
  "This recursive function returns a buffer from BUFLIST whose
name does not match REGEXP, or returns nil"
  (cond
   ((eq buflist nil)
    nil)
   ((string-match regexp (buffer-name (car buflist)))
    (fib-check-name (cdr buflist) regexp))
   (t
    (car buflist))))

(defun interesting-buffer-p (&optional buffer) 
  "This predicate indicates whether or not the given BUFFER is
interesting.  Interesting buffers are ones whose name does _not_ match
the regexp in the variable uninteresting-buffers.  If buffer is nil or
not specified, the current buffer is used."
  (if (not buffer) (setq buffer (current-buffer)))
  (and (not (get-buffer-window buffer))
       (not (string-match uninteresting-buffers (buffer-name buffer)))))

(defun find-interesting-buffer ()
  "This function will find an interesting buffer and return it.  An
uninteresting buffer is one whose name matches the regexp in the
variable uninteresting-buffers.  If it cannot find an interesting
buffer, it will return the first buffer it thinks has any interest at
all (i.e. one which doesn't match the regexp in the variable
totally-uninteresting-buffers)."
  (interactive)
  (cond
   ((fib-check-name (buffer-list) uninteresting-buffers))
   ((fib-check-name (buffer-list) totally-uninteresting-buffers))
   (t
    (get-buffer-create "*scratch*"))))

(defun close-current-window ()
  "Close current window; select an interesting buffer if there is only
one window.  Continuous calls to this function will cycle through the
interesting buffers, if there are any.  If there are no interesting
buffers, it will cycle through the uninteresting buffers."
  (interactive)
  (bury-buffer)
  (if (eq (selected-window) (next-window))		; only one window ?
      (switch-to-buffer (find-interesting-buffer))
    (delete-window)))

(defun switch-to-interesting-buffer ()
  "Find the next interesting buffer and display it in the current window."
  (interactive)
  (bury-buffer)
  (switch-to-buffer (find-interesting-buffer)))

(defun kill-buffer-or-window (&optional even-shell)
  "kill the buffer in the current window and fill the window with a another buffer
   unless it is an uninteresting buffer, in which case kill the window as well."
  (interactive "P")
  (if (or (not (equal major-mode 'cmushell-mode)) even-shell)
      (progn
	(kill-buffer (buffer-name))			        ; kill the current buffer & fill the window with another buffer
	(let ((buf (find-interesting-buffer)))
	  (if (interesting-buffer-p buf)			; if the buffer is _really_ an interesting buffer
	      (if (and (get-buffer-window buf)			; if buffer already displayed in a window, and not this window
		       (not (eq (selected-window) (next-window))))
		  (delete-window)				; delete the window
		(switch-to-buffer buf))				; switch to the interesting buffer
	    (if (eq (selected-window) (next-window))            ; if a boring buffers and there is only one window
		(switch-to-buffer buf)				; switch to the buffer
	      (bury-buffer buf)					; otherwise, bury the uninteresting buffer
	      (delete-window)))))))				; and delete the window

